unit ADC_Continuous;

interface
uses
{$IFDEF VER130}  // Delphi 5
  SysUtils, Classes,
{$ELSE}         // anything other than Delphi 5, presumably XE2 or XE3
  System.SysUtils,
  System.Classes,
{$ENDIF}
  Windows,
  theseUtils,
  ContBufThreads;

type
  TADCContAcqWorker = class;
  PContCallback = procedure (pBuf: PWord; BufSize, Flags, Context: LongWord); cdecl;

  TADCContBufWorker = class(TStartableThread)
  public
    BytesPerBuf: LongWord;
    hBufMutex, hBlankBufSem, hDataBufSem, hKillSem: THandle;
    MyAcqThread: TADCContAcqWorker;
    BufBuf: array of PContBuf;
    NextIndexIn, NextIndexOut: Integer;
    Callback: PContCallback;
    CallbackContext: LongWord;

    procedure GetDebugStats(var Blanks, Accessings, GotDatas: LongWord);
  protected
    procedure Execute; override;
    function GetBlankBuf(Flags: LongWord): PContBuf;
    procedure PutDataBuf(pNewBuf: PContBuf);
    function GetDataBufOrKilled: PContBuf;
    procedure PutBlankBuf(pNewBuf: PContBuf);
    function ExtraBuf: PContBuf;
  end;

  TADCContAcqWorker = class(TStartableThread)
  public
    DI: LongWord;
    TargetPipe: LongWord;
    bCounterControl: Boolean;
    DivisorA, DivisorB: Word;
    //BytesPerBC,
    IOStatus: LongWord;
    MyBufThread: TADCContBufWorker;
    hStartEvent: THandle;
  protected
    procedure Execute; override;
  end;

function ADC_BulkContinuousCallbackStart(DeviceIndex: LongWord; BufSize: LongWord; BaseBufCount: LongWord; Context: LongWord; pCallback: Pointer): LongWord; cdecl;
function ADC_BulkContinuousCallbackStartClocked(DeviceIndex: LongWord; BufSize: LongWord; BaseBufCount: LongWord; Context: LongWord; pCallback: Pointer; var Hz: Double): LongWord; cdecl;
function ADC_BulkContinuousEnd(DeviceIndex: LongWord; pIOStatus: PLongWord): LongWord; cdecl;

function ADC_BulkContinuousDebug(DeviceIndex: LongWord; var Blanks, Accessings, GotDatas: LongWord): LongWord; cdecl;

implementation

uses
  Math,
  DeviceLevel,
  FastADC,
  u8254,
  ADC,
  driverLevel
  ;

function ADC_BulkContinuousCallbackStart_Inner(DeviceIndex: LongWord; BufSize: LongWord; BaseBufCount: LongWord; Context: LongWord; pCallback: Pointer; bClocked: Boolean = False): Longword; cdecl;
var
  oADCWorker: TADCWorker;
  oADCContAcqWorker: TADCContAcqWorker;
  oADCContBufWorker: TADCContBufWorker;
  I: Integer;
  hWaits: array[0..1] of THandle;
  Hz: Double;
  
begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[DeviceIndex] do begin
      if not bADCBulk then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      if (BufSize and $1FF) <> 0 then begin
        Result := ERROR_INVALID_USER_BUFFER;
        Exit;
      end;

      oADCWorker := TADCWorker(ADCWorker);
      oADCContAcqWorker := TADCContAcqWorker(ADCContAcqWorker);
      oADCContBufWorker := TADCContBufWorker(ADCContBufWorker);
      if ((oADCWorker <> nil) and (oADCWorker.BytesLeft <> 0))
      or (oADCContAcqWorker <> nil) or (oADCContBufWorker <> nil) then begin
        Result := ERROR_TOO_MANY_OPEN_FILES;
        Exit;
      end;

      if bClocked then
      begin
        CTR_8254Mode(DeviceIndex, 0, 1, 2);
        CTR_8254Mode(DeviceIndex, 0, 2, 3);
      end;

      ADCContAcqWorker := TADCContAcqWorker.Create(True);
      ADCContBufWorker := TADCContBufWorker.Create(True);
      with TADCContBufWorker(ADCContBufWorker) do begin
        FreeOnTerminate := False;
        MyAcqThread := TADCContAcqWorker(ADCContAcqWorker);
        BytesPerBuf := BufSize;
        Callback := pCallback;
        CallbackContext := Context;
        hBufMutex := CreateMutex(nil, False, nil);
        hKillSem := CreateSemaphore(nil, 0, $FFFFFF, nil);
        hBlankBufSem := CreateSemaphore(nil, BaseBufCount, $FFFFFF, nil);
        hDataBufSem := CreateSemaphore(nil, 0, $FFFFFF, nil);
        SetLength(TADCContBufWorker(ADCContBufWorker).BufBuf, BaseBufCount);
        for I := 0 to BaseBufCount-1 do begin
          New(TADCContBufWorker(ADCContBufWorker).BufBuf[I]);
          SetLength(TADCContBufWorker(ADCContBufWorker).BufBuf[I].ADBuf, BufSize);
          TADCContBufWorker(ADCContBufWorker).BufBuf[I].BufState := bsBlank;
        end;
        Start;
      end;
      with TADCContAcqWorker(ADCContAcqWorker) do begin
        FreeOnTerminate := False;
        MyBufThread := TADCContBufWorker(ADCContBufWorker);
        hStartEvent := CreateEvent(nil, True, False, nil);
        DI := DeviceIndex;
        TargetPipe := $86;
        //BytesPerBC := BufSize * BaseBufCount;
        bCounterControl := bClocked;
        if bCounterControl then CalcHzDivisors(@Hz, @DivisorA, @DivisorB, RootClock);
        Start;
      end;
      hWaits[0] := TADCContAcqWorker(ADCContAcqWorker).Handle;
      hWaits[1] := TADCContAcqWorker(ADCContAcqWorker).hStartEvent;
      WaitForMultipleObjects(2, @hWaits[0], False, INFINITE);
      CloseHandle(TADCContAcqWorker(ADCContAcqWorker).hStartEvent);

      if oADCWorker <> nil then oADCWorker.Free;
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function ADC_BulkContinuousCallbackStartClocked(DeviceIndex: LongWord; BufSize: LongWord; BaseBufCount: LongWord; Context: LongWord; pCallback: Pointer; var Hz: Double): LongWord; cdecl;
begin
  ADC_BulkContinuousCallbackStart_Inner(DeviceIndex, BufSize, BaseBufCount, Context, pCallback, True);
end;//ADC_BulkContinuousCallbackStartClocked

function ADC_BulkContinuousCallbackStart(DeviceIndex: LongWord; BufSize: LongWord; BaseBufCount: LongWord; Context: LongWord; pCallback: Pointer): LongWord; cdecl;
begin
  ADC_BulkContinuousCallbackStart_Inner(DeviceIndex, BufSize, BaseBufCount, Context, pCallback {,False});
end;


function ADC_BulkContinuousEnd(DeviceIndex: LongWord; pIOStatus: PLongWord): LongWord; cdecl;
var
  oADCContAcqWorker: TADCContAcqWorker;
  oADCContBufWorker: TADCContBufWorker;
  I: Integer;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[DeviceIndex] do begin
      if not bADCBulk then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      oADCContAcqWorker := TADCContAcqWorker(ADCContAcqWorker);
      oADCContBufWorker := TADCContBufWorker(ADCContBufWorker);
      ADCContAcqWorker := nil;
      ADCContBufWorker := nil;
      if (oADCContAcqWorker = nil) or (oADCContBufWorker = nil) then begin
        Result := ERROR_HANDLE_EOF;
        Exit;
      end;

      if pIOStatus <> nil then pIOStatus^ := oADCContAcqWorker.IOStatus;

      oADCContAcqWorker.Terminate;
      ReleaseSemaphore(oADCContBufWorker.hKillSem, 1, nil);
      ReleaseSemaphore(oADCContBufWorker.hDataBufSem, 1, nil);

      oADCContAcqWorker.WaitFor;
      oADCContBufWorker.WaitFor;
      CloseHandle(oADCContBufWorker.hBufMutex);
      CloseHandle(oADCContBufWorker.hKillSem);
      CloseHandle(oADCContBufWorker.hBlankBufSem);
      CloseHandle(oADCContBufWorker.hDataBufSem);
      for I := 0 to High(oADCContBufWorker.BufBuf) do begin
        Dispose(oADCContBufWorker.BufBuf[I]);
      end;
      SetLength(oADCContBufWorker.BufBuf, 0);

      oADCContAcqWorker.Free;
      oADCContBufWorker.Free;

      if pRingAcqData <> nil then begin
        CloseHandle(pRingAcqData.hSem);
        Dispose(pRingAcqData);
        pRingAcqData := nil;
      end;
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function ADC_BulkContinuousDebug(DeviceIndex: LongWord; var Blanks, Accessings, GotDatas: LongWord): LongWord; cdecl;
var
  oADCContAcqWorker: TADCContAcqWorker;
  oADCContBufWorker: TADCContBufWorker;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[DeviceIndex] do begin
      if not bADCBulk then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      oADCContAcqWorker := TADCContAcqWorker(ADCContAcqWorker);
      oADCContBufWorker := TADCContBufWorker(ADCContBufWorker);
      if (oADCContAcqWorker = nil) or (oADCContBufWorker = nil) then begin
        Result := ERROR_HANDLE_EOF;
        Exit;
      end;

      oADCContBufWorker.GetDebugStats(Blanks, Accessings, GotDatas);
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

procedure TADCContAcqWorker.Execute;
  function DoBCControl(TCSize: LongWord; ControlData: LongWord): LongWord;
  begin
    Result := _GenericVendorWrite(DI, $BC, TCSize shr 16, TCSize, SizeOf(ControlData), @ControlData, Dev[DI].bCyUSB);
  end;

  procedure DoErrOut(Status: LongWord);
  begin
    IOStatus := Status;
    ReturnValue := Status;
    ReleaseSemaphore(MyBufThread.hKillSem, 1, nil);
    ReleaseSemaphore(MyBufThread.hDataBufSem, 1, nil);
  end;

  var
    BytesLeft, DataSize,
    Status: LongWord;
    ThisBuf: PContBuf;
begin
  SetThreadPriority(Handle, THREAD_PRIORITY_TIME_CRITICAL);

  IOStatus := ERROR_SUCCESS;
  //GenericVendorWrite(DI, AUR_SET_GPIF_MODE, 0, $1, 0, nil);
  if bCounterControl then
    _GenericVendorWrite(DI, $C5, DivisorA, DivisorB, 0, nil, Dev[DI].bCyUSB) // debug mode for CNSPoway Boian Mitov not supported in latest firmwares... --J2H
  else
    DoBCControl(0, $01000007)
  ;
  SetEvent(hStartEvent);

  while not Terminated do begin
    ThisBuf := MyBufThread.GetBlankBuf(0);

    ThisBuf.UsedSize := 0;
    Status := _GenericBulkIn(DI, TargetPipe, @ThisBuf.ADBuf[0], Length(ThisBuf.ADBuf), ThisBuf.UsedSize);

    if Status <> ERROR_SUCCESS then begin
      DoErrOut(Status);
      DoBCControl(0, $00020002);
      //D4StartStop($2);
      Exit;
    end;

    if ThisBuf.UsedSize <> 0 then
      MyBufThread.PutDataBuf(ThisBuf)
    else
      MyBufThread.PutBlankBuf(ThisBuf)
    ;
  end;

  DoBCControl(0, $00020002);
  //D4StartStop($2);

  //`@ This whole drain loop was commented out in DebugTemp. Wut?
  repeat
    BytesLeft := 0;
    DataSize := SizeOf(BytesLeft);
    Status := _GenericVendorRead(DI, $BC, 0, 0, DataSize, @BytesLeft, Dev[DI].bCyUSB);
    if Status <> ERROR_SUCCESS then begin
      DoErrOut(Status);
      Exit;
    end;
    BytesLeft := BytesLeft and $0000FFFF;
    if BytesLeft = 0 then Break;

    ThisBuf := MyBufThread.GetBlankBuf(0);

    ThisBuf.UsedSize := 0;
    DataSize := Length(ThisBuf.ADBuf);
    if DataSize < BytesLeft then BytesLeft := DataSize;
    Status := _GenericBulkIn(DI, TargetPipe, @ThisBuf.ADBuf[0], BytesLeft, ThisBuf.UsedSize);

    if Status <> ERROR_SUCCESS then begin
      MyBufThread.PutBlankBuf(ThisBuf);
      DoErrOut(Status);
      Exit;
    end;

    if ThisBuf.UsedSize <> 0 then
      MyBufThread.PutDataBuf(ThisBuf)
    else
      MyBufThread.PutBlankBuf(ThisBuf)
    ;
  until True;//`@ False;
  ReturnValue := ERROR_SUCCESS;
end;

{ TADCContBufWorker }

procedure TADCContBufWorker.Execute;
var
  ThisBuf: PContBuf;

begin
  repeat
    if Terminated then begin
      ReturnValue := ERROR_OPERATION_ABORTED;
      Exit;
    end;

    ThisBuf := GetDataBufOrKilled;
    if ThisBuf = nil then begin
      try
        Callback(nil, 0, CallbackFlag_EndStream, CallbackContext);
      except
      end;
      ReturnValue := ERROR_OPERATION_ABORTED;
      Exit;
    end;

    if (ThisBuf.UsedSize <> 0) or (ThisBuf.Flags <> 0) then try
      Callback(@ThisBuf.ADBuf[0], ThisBuf.UsedSize, ThisBuf.Flags, CallbackContext);
    except
    end;

    PutBlankBuf(ThisBuf);
  until False;
  ReturnValue := ERROR_SUCCESS;
end;

function TADCContBufWorker.ExtraBuf: PContBuf;
var
  I: Integer;

begin
  WaitForSingleObject(hBufMutex, INFINITE);
    I := Length(BufBuf);
    SetLength(BufBuf, I + 1);
    New(BufBuf[I]);
    Result := BufBuf[I];
    SetLength(Result.ADBuf, BytesPerBuf);
    Result.BufState := bsAccessing;
  ReleaseMutex(hBufMutex);
end;

function TADCContBufWorker.GetBlankBuf(Flags: LongWord): PContBuf;
var
  I: Integer;

begin
  if WaitForSingleObject(hBlankBufSem, 0) = WAIT_TIMEOUT then begin
    Result := ExtraBuf;
    Result.Flags := CallbackFlag_Inserted or Flags;
    Exit;
  end;
  Result := nil;
  WaitForSingleObject(hBufMutex, INFINITE);
    for I := 0 to High(BufBuf) do
      if BufBuf[I].BufState = bsBlank then begin
        Result := BufBuf[I];
        Result.Flags := Flags;
        Break;
      end
    ;
  ReleaseMutex(hBufMutex);

  if Result = nil then begin
    Result := ExtraBuf; //We have a semaphore to prevent this, so not likely to matter.
    Result.Flags := CallbackFlag_Inserted or Flags;
  end;
end;

function TADCContBufWorker.GetDataBufOrKilled: PContBuf;
var
  I: Integer;

begin
  repeat
    WaitForSingleObject(hDataBufSem, INFINITE);
    Result := nil;
    WaitForSingleObject(hBufMutex, INFINITE);
      for I := 0 to High(BufBuf) do
        if (BufBuf[I].BufState = bsGotData) and (BufBuf[I].Index = NextIndexOut) then begin
          Result := BufBuf[I];
          Result.BufState := bsAccessing;
          Inc(NextIndexOut);
          Break;
        end
      ;
    ReleaseMutex(hBufMutex);

    if Result = nil then begin
      if WaitForSingleObject(hKillSem, 0) <> WAIT_TIMEOUT then Exit;
    end;
  until Result <> nil; //We have a semaphore to prevent this, so just go around again if it happens.
end;

procedure TADCContBufWorker.GetDebugStats(var Blanks, Accessings, GotDatas: LongWord);
var
  I: Integer;
  iBlanks, iAccessings, iGotDatas: LongWord;

begin
  iBlanks := 0;
  iAccessings := 0;
  iGotDatas := 0;
  WaitForSingleObject(hBufMutex, INFINITE);
    for I := 0 to High(BufBuf) do
      case BufBuf[I].BufState of
        bsBlank: Inc(iBlanks);
        bsAccessing: Inc(iAccessings);
        bsGotData: Inc(iGotDatas);
      end
    ;
  ReleaseMutex(hBufMutex);
  Blanks     := iBlanks;
  Accessings := iAccessings;
  GotDatas   := iGotDatas;
end;

procedure TADCContBufWorker.PutBlankBuf(pNewBuf: PContBuf);
begin
  WaitForSingleObject(hBufMutex, INFINITE);
    pNewBuf.BufState := bsBlank;
  ReleaseMutex(hBufMutex);
  ReleaseSemaphore(hBlankBufSem, 1, nil);
end;

procedure TADCContBufWorker.PutDataBuf(pNewBuf: PContBuf);
begin
  WaitForSingleObject(hBufMutex, INFINITE);
    pNewBuf.BufState := bsGotData;
    pNewBuf.Index := NextIndexIn;
    Inc(NextIndexIn);
  ReleaseMutex(hBufMutex);
  ReleaseSemaphore(hDataBufSem, 1, nil);
end;


end.

