unit CyUSB;
{
  Should contain everything related to talking to the Cypress IOCTL layer

  No other unit should call any Cypress-specific functions
}
interface

uses
  Windows,
  AIOUSB_Constants;

type
  TCySetupPacket = packed record
    bRequestType: Byte;
    bRequest: Byte;
    Value,
    Index,
    Len: Word;
    ulTimeOut: LongWord;
  end;

  TCySingleTransfer = packed record
    SetupPacket: TCySetupPacket;

    reserved: Byte;

    ucEndpointAddress: Byte;
    NtStatus,
    UsbdStatus,
    IsoPacketOffset,
    IsoPacketLength,
    BufferOffset,
    BufferLength: LongWord;
  end;
  PCySingleTransfer = ^TCySingleTransfer;

function _Cy_GenericRead(DeviceIndex: LongWord; RequestType, Request: Byte; Value, Index: Word; var DataSize: LongWord; pData: Pointer): LongWord; register;
function _hCy_GenericVendorRead(hDevice: THandle; Request: Byte; Value, Index: Word; var DataSize: LongWord; pData: Pointer): LongWord; cdecl;
function _hCy_GenericVendorWrite(hDevice: THandle; Request: Byte; Value, Index: Word; DataSize: LongWord; pData: Pointer): LongWord; cdecl;

const
  CyBaseTimeout = 42; //It's an integer, and all the .PDF says is it's the value you'd expect divided by 1000.
  AIOCyUSBClassGUID:  TGUID = (D1: $D52D27E3; D2: $CF76; D3: $25EB; D4: ($72, $EA, $59, $6F, $22, $34, $C0, $E1) );

  IOCTL_ADAPT_INDEX = $0000;
  IOCTL_CY_BASE = LongWord(FILE_DEVICE_UNKNOWN shl 16) or (FILE_ANY_ACCESS shl 14);
  IOCTL_ADAPT_SEND_EP0_CONTROL_TRANSFER = IOCTL_CY_BASE or ((IOCTL_ADAPT_INDEX+8) shl 2) or METHOD_BUFFERED;
  IOCTL_ADAPT_SELECT_INTERFACE          = IOCTL_CY_BASE or ((IOCTL_ADAPT_INDEX+3) shl 2) or METHOD_BUFFERED;
  IOCTL_ADAPT_SEND_NON_EP0_DIRECT       = IOCTL_CY_BASE or ((IOCTL_ADAPT_INDEX+18) shl 2) or METHOD_NEITHER;

implementation

uses
  Math, DeviceLevel;
  

function _Cy_GenericRead(DeviceIndex: LongWord; RequestType, Request: Byte; Value, Index: Word; var DataSize: LongWord; pData: Pointer): LongWord; register;
var
  Buf: AnsiString;
  pCT: PCySingleTransfer;
  BytesReturned: LongWord;
const
  C_EP0_BM_READ_REQUEST = $80;
  
begin
  try
    SetString(Buf, PAnsiChar(pData), DataSize);
  except
    Result := ERROR_NOACCESS;
    Exit;
  end;

  Buf := StringOfChar(AnsiChar(#0), SizeOf(TCySingleTransfer)) + Buf;
  pCT := PCySingleTransfer(PAnsiChar(Buf));
  pCT.SetupPacket.bRequestType := RequestType or C_EP0_BM_READ_REQUEST;
  pCT.SetupPacket.bRequest := Request;
  pCT.SetupPacket.Value := Value;
  pCT.SetupPacket.Index := Index;
  pCT.SetupPacket.Len := DataSize;
  pCT.SetupPacket.ulTimeOut := CyBaseTimeout;
  pCT.ucEndpointAddress := 0; //EP0.
  pCT.IsoPacketLength := 0;
  pCT.BufferOffset := SizeOf(TCySingleTransfer);
  pCT.BufferLength := pCT.SetupPacket.Len;
  BytesReturned := 0;
  if DeviceIoControl(GetThreadHandleForDevice(DeviceIndex), IOCTL_ADAPT_SEND_EP0_CONTROL_TRANSFER, PAnsiChar(Buf), Length(Buf), PAnsiChar(Buf), Length(Buf), BytesReturned, nil) then begin
    Result := ERROR_SUCCESS;
    BytesReturned := Min(BytesReturned, Length(Buf));
    if BytesReturned <= SizeOf(TCySingleTransfer) then
      DataSize := 0
    else
      DataSize := BytesReturned - SizeOf(TCySingleTransfer)
    ;
    try
      Move(Buf[1 + SizeOf(TCySingleTransfer)], pData^, DataSize);
    except
      Result := ERROR_NOACCESS;
    end;
  end
  else begin
    Result := GetLastError;
    if Result = ERROR_SUCCESS then Result := ERROR_GEN_FAILURE;
    Buf := '';
  end;
end;

function _hCy_GenericVendorRead(hDevice: THandle; Request: Byte; Value, Index: Word; var DataSize: LongWord; pData: Pointer): LongWord; cdecl;
var
  Buf: AnsiString;
  pCT: PCySingleTransfer;
  BytesReturned: LongWord;

begin
  try
    SetString(Buf, PAnsiChar(pData), DataSize);
  except
    Result := ERROR_NOACCESS;
    Exit;
  end;

  Buf := StringOfChar(AnsiChar(#0), SizeOf(TCySingleTransfer)) + Buf;
  pCT := PCySingleTransfer(PAnsiChar(Buf));
  pCT.SetupPacket.bRequestType := C_ENDPOINT_ZERO_READ;
  pCT.SetupPacket.bRequest := Request;
  pCT.SetupPacket.Value := Value;
  pCT.SetupPacket.Index := Index;
  pCT.SetupPacket.Len := DataSize;
  pCT.SetupPacket.ulTimeOut := CyBaseTimeout;
  pCT.ucEndpointAddress := 0; //EP0.
  pCT.IsoPacketLength := 0;
  pCT.BufferOffset := SizeOf(TCySingleTransfer);
  pCT.BufferLength := pCT.SetupPacket.Len;
  BytesReturned := 0;
  if DeviceIoControl(hDevice, IOCTL_ADAPT_SEND_EP0_CONTROL_TRANSFER, PAnsiChar(Buf), Length(Buf), PAnsiChar(Buf), Length(Buf), BytesReturned, nil) then begin
    Result := ERROR_SUCCESS;
    BytesReturned := Min(BytesReturned, Length(Buf));
    if BytesReturned <= SizeOf(TCySingleTransfer) then
      DataSize := 0
    else
      DataSize := BytesReturned - SizeOf(TCySingleTransfer)
    ;
    try
      Move(Buf[1 + SizeOf(TCySingleTransfer)], pData^, DataSize);
    except
      Result := ERROR_NOACCESS;
    end;
  end
  else begin
    Result := GetLastError;
    if Result = ERROR_SUCCESS then Result := ERROR_GEN_FAILURE;
    Buf := '';
  end;
end;

function _hCy_GenericVendorWrite(hDevice: THandle; Request: Byte; Value, Index: Word; DataSize: LongWord; pData: Pointer): LongWord; cdecl;
var
  Buf: AnsiString;
  pCT: PCySingleTransfer;
  BytesReturned: LongWord;

begin
  try
    SetString(Buf, PAnsiChar(pData), DataSize);
  except
    Result := ERROR_NOACCESS;
    Exit;
  end;

  Buf := StringOfChar(AnsiChar(#0), SizeOf(TCySingleTransfer)) + Buf;
  pCT := PCySingleTransfer(PAnsiChar(Buf));
  pCT.SetupPacket.bRequestType := C_ENDPOINT_ZERO_WRITE;
  pCT.SetupPacket.bRequest := Request;
  pCT.SetupPacket.Value := Value;
  pCT.SetupPacket.Index := Index;
  pCT.SetupPacket.Len := DataSize;
  pCT.SetupPacket.ulTimeOut := CyBaseTimeout;
  pCT.ucEndpointAddress := 0; //EP0.
  pCT.IsoPacketLength := 0;
  pCT.BufferOffset := SizeOf(TCySingleTransfer);
  pCT.BufferLength := pCT.SetupPacket.Len;
  BytesReturned := 0;
  if DeviceIoControl(hDevice, IOCTL_ADAPT_SEND_EP0_CONTROL_TRANSFER, PAnsiChar(Buf), Length(Buf), PAnsiChar(Buf), Length(Buf), BytesReturned, nil) then
    Result := ERROR_SUCCESS
  else begin
    Result := GetLastError;
    if Result = ERROR_SUCCESS then Result := ERROR_GEN_FAILURE;
  end;
end;

end.

