unit DAC;

interface

uses
  Windows,

  Flash,
  TheseUtils,
  ContBufThreads;


type
  TDACData = AnsiString;
  
  TDACWorker = class(TStartableThread)
    public
      Index: Integer;
    protected
      procedure Execute; override;
  end;
  
  TDACEntry = packed record
    Channel: Word;
    Value: Word;
  end;
  PDACEntry = ^TDACEntry;

function DACDirect(DeviceIndex: LongWord; Channel: Word; Value: Word): LongWord; cdecl;
function DACMultiDirect(DeviceIndex: LongWord; pDACData: PDACEntry; DACDataCount: LongWord): LongWord; cdecl;
function DACSetBoardRange(DeviceIndex: LongWord; RangeCode: LongWord): LongWord; cdecl;
function DACSetChannelCal(DeviceIndex: LongWord; Channel: LongWord; CalFileName: PAnsiChar): LongWord; cdecl;
function DACOutputOpen(DeviceIndex: LongWord; var ClockHz: Double): LongWord; cdecl; // deprecated
function DACOutputAbort(DeviceIndex: LongWord): LongWord; cdecl; // deprecated
function DACOutputClose(DeviceIndex: LongWord; bWait: LongBool): LongWord; cdecl; // deprecated
function DACOutputCloseNoEnd(DeviceIndex: LongWord; bWait: LongBool): LongWord; cdecl; // deprecated
function DACOutputSetCount(DeviceIndex, NewCount: LongWord): LongWord; cdecl; // deprecated
function DACOutputFrame(DeviceIndex, FramePoints: LongWord; FrameData: PWord): LongWord; cdecl; // deprecated
function DACOutputFrameRaw(DeviceIndex, FramePoints: LongWord; FrameData: PWord): LongWord; cdecl; // deprecated
function DACOutputStatus(DeviceIndex: LongWord): LongWord; cdecl; // deprecated lol
function DACOutputStart(DeviceIndex: LongWord): LongWord; cdecl; // deprecated
function DACOutputSetInterlock(DeviceIndex: LongWord; bInterlock: LongBool): LongWord; cdecl; // depre-shrug. eh.
function DACOutputSetClock(DeviceIndex: LongWord; var ClockHz: Double): LongWord; cdecl;
function DACOutputProcess(DeviceIndex: LongWord; var ClockHz: Double; Samples: LongWord; pSampleData: PWord): LongWord; cdecl;

//procedure TDACWorker.Execute; // deprecated
function DACOutputInvade(DeviceIndex, InvasionType: LongWord; InvasionData: PLongWord): LongWord; cdecl; // deprecated
implementation

uses

{$IFDEF VER130}  // Delphi 5
  SysUtils, Classes,
{$ELSE}         // anything other than Delphi 5, presumably XE2 or XE3
  System.SysUtils,
  System.Classes,
{$ENDIF}
  Math,
  DeviceLevel,
  AIOUSB_Constants,
  VRTable;

function DACDirect(DeviceIndex: LongWord; Channel: Word; Value: Word): LongWord; cdecl;
begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin

    if ImmDACs = 0 then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if bDACStream and (bDACOpen or bDACClosing) then begin
      Result := ERROR_TOO_MANY_OPEN_FILES;
      Exit;
    end;

    if Channel >= ImmDACs then begin
      Result := ERROR_INVALID_ADDRESS;
      Exit;
    end;

    Result := GenericVendorWrite(DeviceIndex, AUR_DAC_IMMEDIATE, Value, Channel, 0, nil);
  end;
end;


// this function converts data from one format the customer uses to a different format for the firmware. -- J2H todo document both here
function DACMultiDirect(DeviceIndex: LongWord; pDACData: PDACEntry; DACDataCount: LongWord): LongWord; cdecl;
var
  Content: AnsiString;
  pDE: PDACEntry;
  pDB, pMB: PByte;
  I, J: Integer;
  L: LongWord;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if ImmDACs = 0 then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if bDACStream and (bDACOpen or bDACClosing) then begin
      Result := ERROR_TOO_MANY_OPEN_FILES;
      Exit;
    end;

    if DACDataCount = 0 then Exit; //If length is zero, it's a nop

    pDE := pDACData;
    L := 0;
    for I := 0 to DACDataCount - 1 do begin
      if pDE.Channel > L then L := pDE.Channel;
      Inc(pDE);
    end;
    //L is now highest channel

    if L >= ImmDACs then begin
      Result := ERROR_INVALID_ADDRESS;
      Exit;
    end;

    L := (L div 8) + 1; //Number of blocks of 8 DACs

    Content := StringOfChar(AnsiChar(#0), L * 17);

    pDB := PByte(PAnsiChar(Content));
    pDE := pDACData;
    for I := 0 to DACDataCount - 1 do begin
      J := pDE.Channel;
      pMB := pDB;
      Inc(pMB, J * 2 + (J div 8) + 1);
      PWord(pMB)^ := pDE.Value;

      pMB := pDB;
      Inc(pMB, (J div 8) * 17);
      pMB^ := pMB^ or (1 shl (J and 7));

      Inc(pDE);
    end;

    Result := GenericVendorWrite(DeviceIndex, AUR_DAC_IMMEDIATE, 0, 0, Length(Content), PAnsiChar(Content));
  end;
end;


function DACSetBoardRange(DeviceIndex: LongWord; RangeCode: LongWord): LongWord; cdecl;
begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  if not Dev[DeviceIndex].bDACBoardRange then begin
    //`@ If the board has channel range, then channel range in a loop.
    Result := ERROR_BAD_TOKEN_TYPE;
    Exit;
  end;

  Result := GenericVendorWrite(DeviceIndex, AUR_DAC_RANGE, RangeCode, 0, 0, nil);
end;

function StringReplaceC(const S: AnsiString; const OldChar, NewChar: AnsiChar): AnsiString;
var
  I: Integer;

begin
  Result := S;
  for I := 1 to Length(Result) do
    if Result[I] = OldChar then
      Result[I] := NewChar
  ;
end;

function DACSetChannelCal(DeviceIndex: LongWord; Channel: LongWord; CalFileName: PAnsiChar): LongWord; cdecl;
var
  CalTable: array of Word;
  Suc: Boolean;
  L, L2: LongWord;
  hFil: THandle;
  InterBuf: AnsiString;
  I, J, K: Integer;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if (FlashSectors = 0) or not bDACChannelCal then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if Channel >= ImmDACs then begin
      Result := ERROR_INVALID_ADDRESS;
      Exit;
    end;

    if (CalFileName = ':1TO1:') or (CalFileName = ':NONE:') then begin
      Result := AIOUSB_FlashEraseSector(DeviceIndex, Channel * 2);
      if Result <> ERROR_SUCCESS then Exit;
      Result := AIOUSB_FlashEraseSector(DeviceIndex, Channel * 2 + 1);
      if Result <> ERROR_SUCCESS then Exit;

      Result := GenericVendorWrite(DeviceIndex, AUR_DEBUG_FLASH_1TO1, 0, Channel, 0, nil);
    end
    else begin

      hFil := CreateFileA(CalFileName, GENERIC_READ, FILE_SHARE_READ or FILE_SHARE_WRITE, nil, OPEN_EXISTING, 0, 0);
      if (hFil = 0) or (hFil = INVALID_HANDLE_VALUE) then begin
        Result := GetLastError;
        if Result = ERROR_SUCCESS then Result := ERROR_GEN_FAILURE;
        Exit;
      end;
      L := GetFileSize(hFil, nil);
      Result := GetLastError;
      if Result <> ERROR_SUCCESS then begin
        CloseHandle(hFil);
        Exit;
      end;
      L := Min(L, $08000000); //Cap for RAM reasons. Even a 0x-based file should be $80000 bytes.
      try
        SetString(InterBuf, nil, L + 1);
        InterBuf[L + 1] := #13; //Terminate for the read-out logic.
      except
        CloseHandle(hFil);
        Result := ERROR_OUTOFMEMORY;
        Exit;
      end;
      Suc := ReadFile(hFil, InterBuf[1], L, L2, nil);
      Result := GetLastError;
      if (Result = ERROR_SUCCESS) and not Suc then Result := ERROR_GEN_FAILURE;
      if (Result = ERROR_SUCCESS) and (L2 < L) then begin
        Result := GetLastError;
        if Result = ERROR_SUCCESS then Result := ERROR_GEN_FAILURE;
      end;
      CloseHandle(hFil);
      if Result <> ERROR_SUCCESS then Exit;

      SetLength(CalTable, $10000);
      try
        InterBuf := StringReplaceC(InterBuf, #10, #13);
        J := 1;
        for I := 0 to High(CalTable) do begin
          repeat
            K := InStrC(#13, InterBuf, J);
            if K = 0 then begin
              Result := ERROR_HANDLE_EOF;
              SetLength(CalTable, 0);
              Exit;
            end
            else if K = J then begin
              Inc(J);
            end
            else begin
{$IFDEF VER130}  // Delphi 5
              CalTable[I] := StrToInt(Trim(Copy(InterBuf, J, K - J)));
{$ELSE}
              CalTable[I] := StrToInt(Trim(UnicodeString(Copy(InterBuf, J, K - J))));
{$ENDIF}

              J := K + 1;
              Break;
            end;
          until False;
        end;
      except
        SetLength(CalTable, 0);
        Result := ERROR_INVALID_DATA;
        Exit;
      end;

      Result := AIOUSB_FlashEraseSector(DeviceIndex, Channel * 2);
      if Result <> ERROR_SUCCESS then Exit;
      Result := AIOUSB_FlashEraseSector(DeviceIndex, Channel * 2 + 1);
      if Result <> ERROR_SUCCESS then Exit;

      Result := AIOUSB_FlashWrite(DeviceIndex, Channel * $20000, Length(CalTable) * SizeOf(CalTable[0]), @CalTable[0]);
      SetLength(CalTable, 0);

    end;
  end;
end;



function DACOutputOpen(DeviceIndex: LongWord; var ClockHz: Double): LongWord; cdecl; // deprecated
begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin

    if not bDACStream then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if bDACOpen or bDACClosing then begin
      Result := ERROR_TOO_MANY_OPEN_FILES;
      Exit;
    end;

    //DoSetIFace(Handle, $100);

    Result := GenericVendorWrite(DeviceIndex, AUR_DAC_CONTROL, $80, 0, 0, nil); //Reset
    if Result <> ERROR_SUCCESS then Exit;

    Result := GenericVendorWrite(DeviceIndex, AUR_DAC_DATAPTR, $00, 0, 0, nil); //DataPtr << 0
    if Result <> ERROR_SUCCESS then Exit;

    CtrDivisor := Round(12000000 / ClockHz);
    ClockHz := 12000000 / CtrDivisor;

    Result := GenericVendorWrite(DeviceIndex, AUR_DAC_DIVISOR, 0, CtrDivisor, 0, nil); //Set divisor
    if Result <> ERROR_SUCCESS then Exit;

    hDACDataMutex := CreateMutex(nil, False, nil);
    hDACDataSem := CreateSemaphore(nil, 0, $FFFF, nil);

    DACWorkerThread := TDACWorker.Create(True);
    DACWorkerThread.FreeOnTerminate := False;
    TDACWorker(DACWorkerThread).Index := DeviceIndex;
    TDACWorker(DACWorkerThread).Start;

    bDACOpen := True;
  end;
end;

function DACOutputAbort(DeviceIndex: LongWord): LongWord; cdecl; // deprecated
begin
  Result := ERROR_NOT_SUPPORTED; //`@
end;

function DACOutputClose(DeviceIndex: LongWord; bWait: LongBool): LongWord; cdecl; // deprecated
var
  L, L2, Releases: Integer;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if not bDACStream then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if not bDACOpen then begin
      Result := ERROR_FILE_NOT_FOUND;
      Exit;
    end;
    bDACOpen := False;

    //Flush pending data to worker buffer, adding EOM($8000), and release semaphore one extra time.
    WaitForSingleObject(hDACDataMutex, INFINITE);
      L := Length(DACData);
      if L = 0 then begin
        SetLength(DACData, 1);
        L2 := Length(PendingDACData);
        if L2 <> 0 then begin //No buffered buffers but we have pending data, so we'll buffer and use that.
          PendingDACData[L2] := AnsiChar(Byte(PendingDACData[L2]) or $80);
          DACData[0] := PendingDACData;
          Releases := 2;
        end
        else begin //No buffered buffers and no pending data, so buffer a little buffer so we can add EOM.
          DACData[0] := #$00#$E0; //`@
          Releases := 2;
        end;
      end
      else begin
        L2 := Length(PendingDACData);
        if L2 <> 0 then begin //Some buffered buffers, but we have pending data, so we'll buffer and use the pending data.
          SetLength(DACDAta, L + 1);
          PendingDACData[L2] := AnsiChar(Byte(PendingDACData[L2]) or $80);
          DACData[L] := PendingDACData;
          Releases := 2;
        end
        else begin //Some buffered buffers and no pending data, so we'll use the last buffer.
          Dec(L);
          L2 := Length(DACData[L]);
          DACData[L][L2] := AnsiChar(Byte(DACData[L][L2]) or $80);
          Releases := 1;
        end;
      end;
    ReleaseMutex(hDACDataMutex);
    PendingDACData := '';

    bDACClosing := True;
    ReleaseSemaphore(hDACDataSem, Releases, nil);
    {
    while WaitForSingleObject(hDACDataSem, 0) <> WAIT_TIMEOUT do
      ReleaseSemaphore(hDACDataSem, 1, nil)
    ;
    }

    {
    if bWait then begin
      //Once worker thread has drained its buffer, it checks bDACOpen, and if false it closes and reports back to us.
      //`@
    end;
    }

    DACWorkerThread.Terminate;
    DACWorkerThread.Free;
  end;
end;

function DACOutputCloseNoEnd(DeviceIndex: LongWord; bWait: LongBool): LongWord; cdecl; // deprecated
var
  L, L2, Releases: Integer;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if not bDACStream then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if not bDACOpen then begin
      Result := ERROR_FILE_NOT_FOUND;
      Exit;
    end;
    bDACOpen := False;

    //Flush pending data to worker buffer, and release semaphore one extra time.
    WaitForSingleObject(hDACDataMutex, INFINITE);
      L2 := Length(PendingDACData);
      if L2 <> 0 then begin
        L := Length(DACData);
        SetLength(DACDAta, L + 1);
        DACData[L] := PendingDACData;
        Releases := 2;
      end
      else begin //No pending data, so just release the semaphore to signal end.
        Releases := 1;
      end;
    ReleaseMutex(hDACDataMutex);
    PendingDACData := '';

    bDACClosing := True;
    ReleaseSemaphore(hDACDataSem, Releases, nil);

    {
    if bWait then begin
      //Once worker thread has drained its buffer, it checks bDACOpen, and if false it closes and reports back to us.
      //`@
    end;
    }

    DACWorkerThread.Terminate;
    DACWorkerThread.Free;
  end;
end;

function DACOutputSetCount(DeviceIndex, NewCount: LongWord): LongWord; cdecl; // deprecated
begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[DeviceIndex] do begin
      if not bDACStream then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      if not bDACOpen then begin
        Result := ERROR_FILE_NOT_FOUND;
        Exit;
      end;

      if NewCount = 0 then begin
        Result := ERROR_INVALID_PARAMETER;
        Exit;
      end;

      if NewCount > 8 then begin
        Result := ERROR_OUTOFMEMORY;
        Exit;
      end;

      DACsUsed := NewCount;
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function DACOutputFrame(DeviceIndex, FramePoints: LongWord; FrameData: PWord): LongWord; cdecl; // deprecated
const
  FrameBufferSize = 8192; //Bytes per buffer, bloody ass band-aid.

var
  I, L, Ch: Integer;
  FrameSamples, FrameBytes, FrameBuffers: Integer;
  NewDACData: TDACData;
  pSample: PWord;
  pBuffer: PAnsiChar;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[DeviceIndex] do begin
      if not bDACStream then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      if not bDACOpen then begin
        Result := ERROR_FILE_NOT_FOUND;
        Exit;
      end;

      if FramePoints <= 0 then begin
        Result := ERROR_INVALID_PARAMETER;
        Exit;
      end;

      if FramePoints > 6400 then begin
        Result := ERROR_OUTOFMEMORY;
        Exit;
      end;

      WaitForSingleObject(hDACDataMutex, INFINITE);
        L := Length(DACData);
      ReleaseMutex(hDACDataMutex);

      if L > 32 then begin
        Result := ERROR_NOT_READY;
        Exit;
      end;

      try
        FrameSamples := Integer(FramePoints) * DACsUsed;
        SetString(NewDACData, PAnsiChar(FrameData), FrameSamples * SizeOf(Word));
      except
        Result := ERROR_NOACCESS;
        Exit;
      end;

      pSample := PWord(PAnsiChar(NewDACData));
      Ch := 0;
      for I := 0 to FrameSamples - 1 do begin
        pSample^ := pSample^ and $0FFF;
        Inc(Ch);
        if Ch = DACsUsed then begin
          pSample^ := pSample^ or $2000;
          Ch := 0;
        end;
        Inc(pSample);
      end;
      Dec(pSample);
      pSample^ := pSample^ or $4000;

      NewDACData := PendingDACData + NewDACData;
      FrameBytes := Length(NewDACData);
      FrameBuffers := FrameBytes div FrameBufferSize;
      pBuffer := PAnsiChar(NewDACData);

      WaitForSingleObject(hDACDataMutex, INFINITE);
        L := Length(DACData);
        SetLength(DACData, L + FrameBuffers);
        for I := 0 to FrameBuffers - 1 do begin
          SetString(DACData[L + I], pBuffer, FrameBufferSize);
          Inc(pBuffer, FrameBufferSize);
        end;
      ReleaseMutex(hDACDataMutex);
      ReleaseSemaphore(hDACDataSem, FrameBuffers, nil);

      FrameBytes := FrameBytes mod FrameBufferSize;
      if FrameBytes <> 0 then
        SetString(PendingDACData, pBuffer, FrameBytes)
      else
        PendingDACData := ''
      ;

      NewDACData := '';
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function DACOutputFrameRaw(DeviceIndex, FramePoints: LongWord; FrameData: PWord): LongWord; cdecl; // deprecated
const
  FrameBufferSize = 8192; //Bytes per buffer, bloody ass band-aid.

var
  I, L: Integer;
  FrameSamples, FrameBytes, FrameBuffers: Integer;
  NewDACData: TDACData;
  pBuffer: PAnsiChar;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[DeviceIndex] do begin
      if not bDACStream then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      if not bDACOpen then begin
        Result := ERROR_FILE_NOT_FOUND;
        Exit;
      end;

      if FramePoints <= 0 then begin
        Result := ERROR_INVALID_PARAMETER;
        Exit;
      end;

      if FramePoints > 6400 then begin
        Result := ERROR_OUTOFMEMORY;
        Exit;
      end;

      WaitForSingleObject(hDACDataMutex, INFINITE);
        L := Length(DACData);
      ReleaseMutex(hDACDataMutex);

      if L > 32 then begin
        Result := ERROR_NOT_READY;
        Exit;
      end;

      try
        FrameSamples := Integer(FramePoints) * DACsUsed;
        SetString(NewDACData, PAnsiChar(FrameData), FrameSamples * SizeOf(Word));
      except
        Result := ERROR_NOACCESS;
        Exit;
      end;

      NewDACData := PendingDACData + NewDACData;
      FrameBytes := Length(NewDACData);
      FrameBuffers := FrameBytes div FrameBufferSize;
      pBuffer := PAnsiChar(NewDACData);

      WaitForSingleObject(hDACDataMutex, INFINITE);
        L := Length(DACData);
        SetLength(DACData, L + FrameBuffers);
        for I := 0 to FrameBuffers - 1 do begin
          SetString(DACData[L + I], pBuffer, FrameBufferSize);
          Inc(pBuffer, FrameBufferSize);
        end;
      ReleaseMutex(hDACDataMutex);
      ReleaseSemaphore(hDACDataSem, FrameBuffers, nil);

      FrameBytes := FrameBytes mod FrameBufferSize;
      if FrameBytes <> 0 then
        SetString(PendingDACData, pBuffer, FrameBytes)
      else
        PendingDACData := ''
      ;

      NewDACData := '';
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function DACOutputStatus(DeviceIndex: LongWord): LongWord; cdecl; // deprecated lol
begin
  Result := ERROR_NOT_SUPPORTED; //`@
end;

function DACOutputStart(DeviceIndex: LongWord): LongWord; cdecl; // deprecated
begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[DeviceIndex] do begin
      if not bDACStream then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      if not bDACOpen then begin
        Result := ERROR_FILE_NOT_FOUND;
        Exit;
      end;

      if bDACStarted then begin
        Result := ERROR_ALREADY_EXISTS;
        Exit;
      end;

      Result := GenericVendorWrite(DeviceIndex, AUR_DAC_CONTROL, $01, 0, 0, nil);

      if Result = ERROR_SUCCESS then bDACStarted := True;
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function DACOutputSetInterlock(DeviceIndex: LongWord; bInterlock: LongBool): LongWord; cdecl; // depre-shrug. eh.
var
  Value: Byte;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[DeviceIndex] do begin
      if not bDACStream then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      if not bDACOpen then begin
        Result := ERROR_FILE_NOT_FOUND;
        Exit;
      end;

      if bInterlock then Value := $20 else Value := $00;
      Result := GenericVendorWrite(DeviceIndex, AUR_DAC_CONTROL, Value, 0, 0, nil);
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function DACOutputSetClock(DeviceIndex: LongWord; var ClockHz: Double): LongWord; cdecl;
type
  TDIO16ClockData = packed record
    Disables: Byte;
    WriteOctDac, ReadOctDac: Word;
  end;

var
  DCD: TDIO16ClockData;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin

    if not (bDACStream or bDACDIOClock) then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if bDACDIOClock then begin
      DCD.Disables := $00;
      DCD.WriteOctDac := OctDacFromFreq(ClockHz);
      DCD.ReadOctDac := $0000;
      Result := GenericVendorWrite(DeviceIndex, AUR_DIO_SETCLOCKS, 0, 0, SizeOf(DCD), @DCD);
    end
    else begin
      CtrDivisor := Round(C_DAC_BASE_CLOCK_RATE / ClockHz);
      ClockHz := C_DAC_BASE_CLOCK_RATE / CtrDivisor;

      Result := GenericVendorWrite(DeviceIndex, AUR_DAC_DIVISOR, 0, CtrDivisor, 0, nil); //Set divisor
    end;
  end;
end;

function DACOutputProcess(DeviceIndex: LongWord; var ClockHz: Double; Samples: LongWord; pSampleData: PWord): LongWord; cdecl;
const
  BlockSamples = 31*512;
var
  TargetPipe: LongWord;
  L: LongWord;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if not bDACStream then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if bDACOpen or bDACClosing then begin
      Result := ERROR_TOO_MANY_OPEN_FILES;
      Exit;
    end;

    if Samples >= C_USBDA128A_RAM_SIZE then begin
      Result := ERROR_OUTOFMEMORY;
      Exit;
    end;

    //Result := DoSetIFace(Handle, $100);
    //if Result <> ERROR_SUCCESS then begin Result := GetMappedFailure(DeviceIndex); Exit; end;

    Result := GenericVendorWrite(DeviceIndex, AUR_DAC_CONTROL, $80, 0, 0, nil); //Reset
    if Result <> ERROR_SUCCESS then Exit;

    Result := GenericVendorWrite(DeviceIndex, AUR_DAC_DATAPTR, $00, 0, 0, nil); //DataPtr << 0
    if Result <> ERROR_SUCCESS then Exit;

    CtrDivisor := Round(C_DAC_BASE_CLOCK_RATE / ClockHz);
    ClockHz := C_DAC_BASE_CLOCK_RATE / CtrDivisor;

    Result := GenericVendorWrite(DeviceIndex, AUR_DAC_DIVISOR, 0, CtrDivisor, 0, nil); //Set divisor
    if Result <> ERROR_SUCCESS then Exit;

    //Transfer data
    TargetPipe := $02;
    while Samples > BlockSamples do begin
      L := 0;
      Result := AWU_GenericBulkOut(DeviceIndex, TargetPipe, pSampleData, BlockSamples * 2, L);
      if Result <> ERROR_SUCCESS then Exit;
      Inc(pSampleData, BlockSamples);
      Dec(Samples, BlockSamples);
    end;
    L := 0;
    Result := AWU_GenericBulkOut(DeviceIndex, TargetPipe, pSampleData, Samples * 2, L);
    if Result <> ERROR_SUCCESS then Exit; 

    Result := GenericVendorWrite(DeviceIndex, AUR_DAC_CONTROL, $01, 0, 0, nil); //Start
    //if Result <> ERROR_SUCCESS then Exit;
  end;
end;


{ TDACWorker }


function DACOutputInvade(DeviceIndex, InvasionType: LongWord; InvasionData: PLongWord): LongWord; cdecl; // deprecated
begin
  Result := ERROR_NOT_SUPPORTED;
end;

procedure TDACWorker.Execute; // deprecated
var
  {I,} L, Intro: Integer;
  DACData: TDACData;
  TargetPipe, nBytes: LongWord;
  bExit: Boolean;
  //DataPtr: LongWord;
  //BufferSends: LongInt;
  //hDevice: THandle;

begin
  Intro := (128 + 32)*1024; //After one and a quarter 128KB SRAMs, issue the start, if the user hasn't issued it first.
  //BufferSends := 0;
  TargetPipe := $02;
  //DataPtr := 0;

  //hDevice := Dev[Index].Handle;

  bExit := False;
  repeat
    try
      WaitForSingleObject(Dev[Index].hDACDataSem, INFINITE);

      if Dev[Index].bDACAborting then Break;

      WaitForSingleObject(Dev[Index].hDACDataMutex, INFINITE);
        L := Length(Dev[Index].DACData);
        if L = 0 then begin
          if Dev[Index].bDACClosing then bExit := True;
        end
        else begin
          DACData := Dev[Index].DACData[0];
          Dec(L);
          if L <> 0 then begin
            Dev[Index].DACData[0] := Dev[Index].DACData[L]; //For garbage collection
            Move(Dev[Index].DACData[1], Dev[Index].DACData[0], L * 4);
          end;
          SetLength(Dev[Index].DACData, L);
        end;
      ReleaseMutex(Dev[Index].hDACDataMutex);
      if bExit then Break;

      L := Length(DACData);
      {
      for I := 1 to L do
        if ((I and 1) = 0) and ((Byte(DACData[I]) and $80) <> 0) then
          Beep
      ;
      }
      AWU_GenericBulkOut(Index, TargetPipe, PAnsiChar(DACData), L, nBytes);
      DACData := '';

      //Inc(BufferSends);
      //Inc(DataPtr, L);
      if Intro <> 0 then begin
        Dec(Intro, L);
        if Intro <= 0 then begin
          Intro := 0;

          {
          VRD.Request := AUR_DAC_CONTROL; //Reset
          VRD.Value := $80;
          VRD.Index := 0;
          nBytes := 0;
          DeviceIoControl(hDevice, IOCTL_AIOUSB_WRITE, @VRD, SizeOf(VRD), nil, 0, nBytes, nil);

          VRD.Request := AUR_DAC_DIVISOR; //Set divisor
          VRD.Value := 0;
          VRD.Index := Dev[Index].CtrDivisor;
          nBytes := 0;
          DeviceIoControl(hDevice, IOCTL_AIOUSB_WRITE, @VRD, SizeOf(VRD), nil, 0, nBytes, nil);
          }

          DACOutputStart(Index);
        end;
      end;

      {
      VRD.Request := AUR_DAC_DATAPTR; //Set DataPtr to proper value - debug
      VRD.Value := (DataPtr shr 6) and $0700;
      VRD.Index := DataPtr and $3FFF;
      nBytes := 0;
      DeviceIoControl(hDevice, IOCTL_AIOUSB_WRITE, @VRD, SizeOf(VRD), nil, 0, nBytes, nil);
      }
    except
      Break;
    end;
  until Terminated; 

  WaitForSingleObject(Dev[Index].hDACDataMutex, INFINITE);
    SetLength(Dev[Index].DACData, 0);
  ReleaseMutex(Dev[Index].hDACDataMutex);

  //if BufferSends <> 0 then BufferSends := 0;

  //CloseHandle(hDevice);
  CloseHandle(Dev[Index].hDACDataMutex);
  CloseHandle(Dev[Index].hDACDataSem);
  Dev[Index].bDACClosing := False;
  Dev[Index].bDACAborting := False;
  Dev[Index].bDACStarted := False;
end;

end.
