unit DeviceLevel;

interface

uses Windows,
     WinUSB,
     CyUSB,
     theseUtils,
     DAC,
     ContBufThreads,
     ADC_Continuous,
     ADC_Ring;
type

  TDeviceData = record
    bOpen, bDeviceWasHere, bGotHandle: Boolean;
    hHandleTableMutex: THandle;
    HandleTable: array of record
      ThreadID: LongWord;
      hDevice: THandle;
    end;
    ExpandFromOldHandle, hWinUSBFile: THandle;
    bCyUSB: Boolean; //`@ Should be a load-type enum, to protect if WinUSB not loaded.
    DevPath: AnsiString;
    DIOBytes, DIOConfigBits: LongWord;
    LastDIOData: array of Byte;
    Counters: LongWord;
    Tristates: LongWord;
    bGateSelectable: Boolean;
    RootClock: Double;
    //LastPurpose: Byte;
    LastEventCount: Word;
    PID: LongWord;
    CtrDivisor: Word;
    bGetName: Boolean;
    ConfigBytes: LongWord;

    ImmDACs, ImmADCs: LongWord;

    bDACBoardRange, bDACChannelCal, bDACStream, bDACDIOClock: Boolean;
    DACsUsed: Integer;
    bDACOpen, bDACClosing, bDACAborting, bDACStarted: Boolean;
    DACData: array of TDACData;
    hDACDataMutex, hDACDataSem: THandle;
    PendingDACData: TDACData;
//    WorkerThread: TDACWorker;
    DACWorkerThread: TStartableThread;

//  CtrMonitorThread: TCtrMonitorThread;
    CtrMonitorThread: TStartableThread;
    bADCBulk: Boolean;
    ADCChannels, ADCMUXChannels: LongWord;
//    ADCWorker: TADCWorker;
    ADCWorker: TStartableThread;
//    ADCContAcqWorker: TADCContAcqWorker;
//    ADCContBufWorker: TADCContBufWorker;
    ADCContAcqWorker: TStartableThread;
    ADCContBufWorker: TStartableThread;
    pRingAcqData: TPRingAcqData;
    RangeShift: Byte;
    FastITConfig, FastITBakConfig: array of Byte;
    CSA_FastConfig, CSA_FastBakConfig: array of Byte;

    bDIOStream, bDIOOpen, bDIORead: Boolean;

    StreamingBlockSize: LongWord;

    bDIODebounce, bDIOSPI: Boolean;

    bSetCustomClocks: Boolean;

    WDGBytes: LongWord;

    bClearFIFO: Boolean;

    FlashSectors: Integer;

    bFirmware20: Boolean;
  end;
  PDeviceData = ^TDeviceData;

function GetThreadHandleForDevice(DeviceIndex: LongWord): THandle;
  
function CheckFirmware20(DeviceIndex: LongWord): LongWord;
function GetDevices: LongWord; cdecl;
function QueryDeviceInfo(DeviceIndex: LongWord; pPID: PLongWord; pNameSize: PLongWord; pName: PAnsiChar; pDIOBytes, pCounters: PLongWord): LongWord; cdecl;
function ClearDevices: LongWord; cdecl;
function AIOUSB_ReloadDeviceLinks: LongWord; cdecl;
function GetDeviceUniqueStr(DeviceIndex: LongWord; pIIDSize: PLongWord; pIID: PAnsiChar): LongWord; cdecl;
function GetDeviceSerialNumber(DeviceIndex: LongWord; var pSerialNumber: Int64): LongWord; cdecl;
function GetDeviceByEEPROMByte(Data: Byte): LongWord; cdecl;
function GetDeviceByEEPROMData(StartAddress, DataSize: LongWord; pData: PByte): LongWord; cdecl;
function ResolveDeviceIndex(DeviceIndex: LongWord): LongWord; cdecl;
function GetDeviceBySerialNumber(const pSerialNumber: Int64): LongWord; cdecl;
function AIOUSB_CloseDevice(DeviceIndex: LongWord): LongWord; cdecl;
function EnsureOpen(const DeviceIndex: LongWord): LongWord;
function ValidateAndEnsureOpen(var DeviceIndex: LongWord): LongWord;
function Validate(var DeviceIndex: LongWord): LongWord;
function GetMappedFailure(DeviceIndex: LongWord): LongWord;
function GenericVendorRead(DeviceIndex: LongWord; Request: Byte; Value, Index: Word; var DataSize: LongWord; pData: Pointer): LongWord; cdecl;
function GenericVendorWrite(DeviceIndex: LongWord; Request: Byte; Value, Index: Word; DataSize: LongWord; pData: Pointer): LongWord; cdecl;
function GetDevicePaths(const ClassGUID: TGUID): TStringArray;
procedure LoadDriverLinks;
function DEB_SetInterface(DeviceIndex: LongWord; TargetInterface: Word): LongWord; cdecl;
function DEB_BulkIn(DeviceIndex: LongWord): LongWord; cdecl;
function DEB_BulkOut(DeviceIndex, TargetPipe: LongWord; pBuf: Pointer; BufLen: LongWord): LongWord; cdecl;
function OctDacFromFreq(var ClockHz: Double): Word;
function AIOUSB_ResetChip(DeviceIndex: LongWord): LongWord; cdecl;
function AIOUSB_UploadD15LoFirmwaresByPID(pFirmScript: PAnsiChar): LongWord; cdecl;
function AWU_GenericBulkIn(DeviceIndex: LongWord; PipeID: LongWord; pData: Pointer; DataSize: LongWord; var BytesRead: LongWord): LongWord; cdecl;
function AWU_GenericBulkOut(DeviceIndex: LongWord; PipeID: LongWord; pData: Pointer; DataSize: LongWord; var BytesWritten: LongWord): LongWord; cdecl;


function GetPipes(DeviceIndex: LongWord; var PipeCount: LongWord; pPipeData: PLongWord): LongWord; cdecl;
function GetDeviceHandle(DeviceIndex: LongWord; Data: PLongWord): LongWord; cdecl;
function AIOUSB_UploadFirmware(DeviceIndex: LongWord; FirmBuf: PAnsiChar; BufLen: LongWord): LongWord; cdecl;

var
  Dev: array[0..31] of TDeviceData;

implementation

uses
  AIOUSB_Constants,
  VRTable,
  DriverLevel,
  EEPROM,
  SysUtils,
  SetupDi,
  CoreExports,
  Classes,
  math;
  
// creates load value for the oscillator chip used on USB-DIO-16H family, including USB-AI-ARB1
// documented by the chipspec
function OctDacFromFreq(var ClockHz: Double): Word;
var
  Octave, Offset: Integer;

begin
  if ClockHz = 0 then begin
    Result := 0;
    Exit;
  end;
  if ClockHz > 40000000 then ClockHz := 40000000;
  //if ClockHz > 66000000 then ClockHz := 66000000;
  Octave := Floor(3.322 * Log10(ClockHz / 1039));
  if Octave < 0 then begin //Don't bother with upper limit, we already covered it and then some above.
    Octave := 0;
    Offset := 0;
    Result := $0000;
  end
  else begin
    Offset := Round(2048 - (LDExp(2078, 10 + Octave)/ClockHz)); //Don't bother with limits, we covered that when calculating the octave.
    Result := (Octave shl 12) or (Offset shl 2);
    Result := MotorolaWord(Result);
  end;
  ClockHz := (2078 shl Octave) / (2 - Offset / 1024);
end;


function Validate(var DeviceIndex: LongWord): LongWord;
var
  I: Integer;
  bGotOne, bOnly: Boolean;

begin
  if (DeviceIndex = diOnly) or (DeviceIndex = diFirst) then begin
    bOnly := DeviceIndex = diOnly;
    bGotOne := False;
    for I := 0 to High(Dev) do begin
      if Dev[I].bGotHandle then begin
        if bGotOne then begin
          Result := ERROR_DUP_NAME;
          Exit;
        end;
        DeviceIndex := I;
        if not bOnly then begin
          Result := ERROR_SUCCESS;
          Exit;
        end;
        bGotOne := True;
      end;
    end;
    if DeviceIndex < LongWord(Length(Dev)) then
      Result := ERROR_SUCCESS
    else
      Result := ERROR_FILE_NOT_FOUND
    ;
  end
  else if DeviceIndex >= LongWord(Length(Dev)) then begin
    Result := ERROR_INVALID_INDEX;
  end
  else
    Result := ERROR_SUCCESS;
  ;
end;

function AIOUSB_CloseDevice(DeviceIndex: LongWord): LongWord; cdecl;
var
  I: Integer;

begin
  Result := Validate(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if not bGotHandle then begin
      if bDeviceWasHere then begin
        bDeviceWasHere := False;
        Result := ERROR_SUCCESS;
      end
      else
        Result := ERROR_INVALID_HANDLE
      ;
      Exit;
    end;

    try
      bDeviceWasHere := False;
      if bCyUSB then begin
        WaitForSingleObject(hHandleTableMutex, INFINITE);
        for I := 0 to High(HandleTable) do
          CloseHandle(HandleTable[I].hDevice)
        ;
        SetLength(HandleTable, 0);
        ReleaseMutex(hHandleTableMutex);
      end
      else begin
        WaitForSingleObject(hHandleTableMutex, INFINITE);
        for I := 0 to High(HandleTable) do
          WinUSB_Free(HandleTable[I].hDevice)
        ;
        SetLength(HandleTable, 0);
        ReleaseMutex(hHandleTableMutex);

        CloseHandle(hWinUSBFile);
      end;
      Result := ERROR_SUCCESS;
    except
      Result := GetLastError;
      if Result = ERROR_SUCCESS then Result := ERROR_GEN_FAILURE;
    end;
    bGotHandle := False;
    hWinUSBFile := 0;
  end;
end;

function GetDevices: LongWord; cdecl;
var
  I: Integer;

begin
  LoadDriverLinks;

  Result := 0;
  for I := 0 to High(Dev) do
    if EnsureOpen(I) = ERROR_SUCCESS then
      Result := Result or (1 shl I)
  ;
end;

function QueryDeviceInfo(DeviceIndex: LongWord; pPID: PLongWord; pNameSize: PLongWord; pName: PAnsiChar; pDIOBytes, pCounters: PLongWord): LongWord; cdecl;
var
  Content: AnsiString;
  wContent: WideString;
  L, L2: LongWord;

begin
  try
    Result := Validate(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    if not Dev[DeviceIndex].bGotHandle then LoadDriverLinks;

    Result := EnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    if pPID      <> nil then pPID     ^ := Dev[DeviceIndex].PID     ;
    if pDIOBytes <> nil then pDIOBytes^ := Dev[DeviceIndex].DIOBytes;
    if pCounters <> nil then pCounters^ := Dev[DeviceIndex].Counters;
    if pNameSize <> nil then begin
      if not Dev[DeviceIndex].bGetName then begin //Old ASM firmwares do not support get-string
        pNameSize^ := 0;
      end
      else begin
        SetString(wContent, nil, pNameSize^ + 1);
        L := Length(wContent) * SizeOf(wContent[1]);
        Result := GenericVendorRead(DeviceIndex, $06, $0302, 0, L, PWChar(wContent)); // Cypress get-string
        if Result <> ERROR_SUCCESS then Exit;

        L2 := Byte(wContent[1]) div 2;
        if L2 <> 0 then begin
          Dec(L2);
          if pName <> nil then begin
            L := Min(pNameSize^, L2+1);
            wContent := Copy(wContent + #0, 2, L);
            Content := AnsiString(wContent);
            Move(Content[1], pName^, L);
          end;
        end;
        Result := ERROR_SUCCESS;
        pNameSize^ := L2;
      end;
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function ResolveDeviceIndex(DeviceIndex: LongWord): LongWord; cdecl;
begin
  Result := Validate(DeviceIndex);
  SetLastError(Result);
  if Result = ERROR_SUCCESS then begin
    Result := DeviceIndex;
  end
  else begin
    Result := diNone;
  end;
end;

function ClearDevices: LongWord; cdecl;
var
  I: Integer;
  wContent: WideString;
  L: LongWord;
  ErrorCode: LongWord;

begin
  for I := High(Dev) downto 0 do begin
    if Dev[I].bGotHandle then begin
      if not Dev[I].bGetName then begin
        //Old ASM firmwares do not support get-string
      end
      else begin
        ErrorCode := EnsureOpen(I);
        if ErrorCode = ERROR_SUCCESS then begin
          SetString(wContent, nil, 257);
          L := Length(wContent) * SizeOf(wContent[1]);
          ErrorCode := GenericVendorRead(I, $06, $0302, 0, L, PWChar(wContent));// Cypress GetString, used to force access to hardware (?)

          if ErrorCode <> ERROR_SUCCESS then
            GetMappedFailure(I)
          ;
        end;
      end;
    end
    else begin
      Dev[I].bDeviceWasHere := False;
    end;
  end;
  Result := ERROR_SUCCESS;
end;

function GetDeviceBySerialNumber(const pSerialNumber: Int64): LongWord; cdecl; // deprecated
var
  I: Integer;
  ThisSerialNumber: Int64;
  Status: LongWord;

begin
  for I := 0 to High(Dev) do begin
    Status := GetDeviceSerialNumber(I, ThisSerialNumber);
    if Status <> ERROR_SUCCESS then Continue;
    if ThisSerialNumber <> pSerialNumber then Continue;

    Result := I;
    Exit;
  end;

  Result := diNone;
end;

function GetDeviceByEEPROMByte(Data: Byte): LongWord; cdecl;
var
  ThisData: Byte;
  Status, ThisDataSize, MyLastError: LongWord;
  aDeviceIndex: Integer;

begin
  Result := diNone;
  MyLastError := ERROR_FILE_NOT_FOUND;

  for aDeviceIndex := 0 to High(Dev) do begin
    ThisDataSize := 1;
    Status := CustomEEPROMRead(aDeviceIndex, 0, ThisDataSize, @ThisData);
    if Status <> ERROR_SUCCESS then Continue;
    if ThisDataSize <> 1 then Continue;
    if ThisData <> Data then Continue;

    if Result = diNone then begin
      Result := aDeviceIndex;
      MyLastError := ERROR_SUCCESS;
    end
    else begin
      MyLastError := ERROR_DUP_NAME;
    end;
  end;
  SetLastError(MyLastError);
end;

function GetDeviceByEEPROMData(StartAddress, DataSize: LongWord; pData: PByte): LongWord; cdecl;
var
  ipData: PByte;
  ThisData: array of Byte;
  Status, ThisDataSize, MyLastError: LongWord;
  aDeviceIndex, J: Integer;
  bMatch: Boolean;

begin
  Result := diNone;
  MyLastError := ERROR_FILE_NOT_FOUND;

  if DataSize > C_CUSTOM_EEPROM_SIZE then begin
    SetLastError(ERROR_BAD_LENGTH);
    Exit;
  end;

  if DataSize = 0 then begin
    Result := ResolveDeviceIndex(diFirst);
    Exit;
  end;

  SetLength(ThisData, DataSize);
  for aDeviceIndex := 0 to High(Dev) do begin
    ThisDataSize := DataSize;
    Status := CustomEEPROMRead(aDeviceIndex, StartAddress, ThisDataSize, @ThisData[0]);
    if Status <> ERROR_SUCCESS then Continue;
    if ThisDataSize <> DataSize then Continue;

    bMatch := True;
    ipData := pData;
    for J := 0 to DataSize - 1 do begin
      if ThisData[J] <> ipData^ then begin
        bMatch := False;
        Break;
      end;
      Inc(ipData);
    end;
    if not bMatch then Continue;

    if Result = diNone then begin
      Result := aDeviceIndex;
      MyLastError := ERROR_SUCCESS;
    end
    else begin
      MyLastError := ERROR_DUP_NAME;
    end;
  end;
  SetLastError(MyLastError);
end;

function GetDeviceSerialNumber(DeviceIndex: LongWord; var pSerialNumber: Int64): LongWord; cdecl;
begin
  Result := GetEEPROMSerialNumber(DeviceIndex, pSerialNumber);
end;

function AIOUSB_ReloadDeviceLinks: LongWord; cdecl;
var
  aDeviceIndex: Integer;

begin
  Result := ERROR_SUCCESS;
  for aDeviceIndex := 0 to High(Dev) do
    AIOUSB_CloseDevice(aDeviceIndex)
  ;
  LoadDriverLinks;
end;

function GetPipes(DeviceIndex: LongWord; var PipeCount: LongWord; pPipeData: PLongWord): LongWord; cdecl;
var
  bSuc: Boolean;
  InterfaceDescriptor: USB_INTERFACE_DESCRIPTOR;
  I, L: Integer;
  PipeInfo: WINUSB_PIPE_INFORMATION;
  pIPipeData: PLongWord;
  hDevice: THandle;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  if Dev[DeviceIndex].bCyUSB then begin
    //`@ CyUSB.
    Result := ERROR_NOT_SUPPORTED; //To generate the hint for the above.
    Result := ERROR_NOT_SUPPORTED;
    Exit;
  end
  else begin //WinUSB.
    hDevice := GetThreadHandleForDevice(DeviceIndex);
    SetLastError(ERROR_GEN_FAILURE);
    bSuc := WinUsb_QueryInterfaceSettings(hDevice, 0, @InterfaceDescriptor);
    if not bSuc then begin
      Result := GetMappedFailure(DeviceIndex);
      PipeCount := 0;
      Exit;
    end;

    if PipeCount < InterfaceDescriptor.bNumEndpoints then
      L := PipeCount
    else
      L := InterfaceDescriptor.bNumEndpoints
    ;
    PipeCount := InterfaceDescriptor.bNumEndpoints;

    pIPipeData := pPipeData;
    for I := 0 to L - 1 do begin
      pIPipeData^ := $FFFFFFFF;
      Inc(pIPipeData);
    end;

    pIPipeData := pPipeData;
    for I := 0 to L - 1 do begin
      bSuc := WinUsb_QueryPipe(hDevice, 0, I, @PipeInfo);
      if not bSuc then begin
        Result := GetMappedFailure(DeviceIndex);
        Exit;
      end;

      pIPipeData^ := PipeInfo.PipeId or (PipeInfo.PipeType shl 8);

      Inc(pIPipeData);
    end;
  end;

  Result := ERROR_SUCCESS;
end;

function GetDeviceHandle(DeviceIndex: LongWord; Data: PLongWord): LongWord; cdecl;
begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  Data^ := GetThreadHandleForDevice(DeviceIndex);
end;

function AIOUSB_UploadFirmware(DeviceIndex: LongWord; FirmBuf: PAnsiChar; BufLen: LongWord): LongWord; cdecl;
begin
  Result := ERROR_NOT_SUPPORTED;
end;

function StrToLines(S: AnsiString): TStringArray;
var
  Count, Capacity, iC, oC,
  L: Integer;

begin
  Count := 0;
  Capacity := 8;
  SetLength(Result, Capacity);
  S := S + #13;
  L := Length(S);
  iC := 1;
  oC := 1;
  while iC <= L do begin
    if (S[iC] = #13) or (S[iC] = #10) then begin
      if Count >= Capacity then begin
        Capacity := Capacity * 2;
        SetLength(Result, Capacity);
      end;
      Result[Count] := Copy(S, oC, iC - oC);
      Inc(Count);

      if (iC < L) and (S[iC] = #13) and (S[iC+1] = #10) then Inc(iC);
      oC := iC+1;
    end;
    Inc(iC);
  end;
  SetLength(Result, Count);
end;

function AIOUSB_UploadD15LoFirmwaresByPID(pFirmScript: PAnsiChar): LongWord; cdecl;
type
  THexChunk = record
    Addr: Word;
    Data: AnsiString;
  end;
  THexChunks = array of THexChunk;

  TPIDFirmData = record
    PID: Word;
    FirmContent: THexChunks;
  end;

  function ParseHex(HexStrs: TStringArray): THexChunks;
  var
    iLine, iChunk: Integer;
    ThisAddr, ThisData: AnsiString;

  begin
    //Validate hex file.
    SetLength(Result, 0);
    for iLine := 0 to High(HexStrs) do begin
      ThisData := HexStrs[iLine];
      if ThisData = '' then Continue;
      if ThisData[1] <> ':' then Exit;
      if not IsHex(Copy(ThisData, 2, MAXINT)) then Exit;
      if (Length(ThisData) and 1) = 0 then Exit; //If source line is even(has colon plus odd nybbles), reject as invalid.
    end;

    SetLength(Result, Length(HexStrs));
    iChunk := 0;
    for iLine := 0 to High(HexStrs) do begin
      ThisData := HexStrs[iLine];
      if ThisData = '' then Continue;
      if Copy(ThisData, 8, 2) <> '00' then Continue; //We only care about type 00 records.

      ThisAddr := Copy(ThisData, 4, 4);
      ThisData := Copy(ThisData, 10, Length(ThisData) - 11);

      Result[iChunk].Addr := HexToInt(ThisAddr);
      Result[iChunk].Data := HexToStr(ThisData);
      Inc(iChunk);
    end;
    SetLength(Result, iChunk);
  end;

var
  PFD: array of TPIDFirmData; //Loading the script would be a bit faster with pointers, but that's not much of an improvement, so safety is more important.
  iPF, iLine, iPath, iFirmware,
  I: Integer;
  ThisLine, PIDStr, FirmStr: AnsiString;
  DevicePath, FirmScript: TStringArray;
  FirmHex: THexChunks;
  hCyFile: THandle;
  Status: LongWord;
  CPUCSByte: Byte;
  ThisPID: Word;

begin
  Result := 0;
  try
    FirmScript := StrToLines(pFirmScript);
    iPF := -1;
    for iLine := 0 to High(FirmScript) do begin
      ThisLine := FirmScript[iLine];
      if ThisLine = '' then Continue;
      if ThisLine[1] = ';' then Continue;

      SplitStr(ThisLine, '=', PIDStr, FirmStr);
      if PIDStr = '' then Continue;
      if FirmStr = '' then Continue;
      if not IsHex(PIDStr) then Continue;
      if not FileExists(String(FirmStr)) then Continue;

      try
        FirmStr := DumpFileToString(FirmStr);
        FirmStr := AnsiString(UpperCase(Trim(String(FirmStr))));
        FirmHex := ParseHex(StrToLines(FirmStr));
        if Length(FirmHex) = 0 then Continue;
      except
        Continue;
      end;

      Inc(iPF);
      SetLength(PFD, iPF+1);
      with PFD[iPF] do begin
        PID := HexToInt(PIDStr) and $7FFF;
        FirmContent := FirmHex;
      end;
    end;

    DevicePath := GetDevicePaths(D15LoClassGUID);
    for iPath := 0 to High(DevicePath) do begin
      ThisLine := DevicePath[iPath];

      I := Pos('PID_', UpperCase(String(ThisLine)));
      if I = 0 then Continue;
      PIDStr := Copy(ThisLine, I + 4, 4);
      if not IsHex(PIDStr) then Continue;
      ThisPID := HexToInt(PIDStr) and $7FFF;
      iFirmware := High(PFD);
      while (iFirmware >= 0) and (PFD[iFirmware].PID <> ThisPID) do Dec(iFirmware);
      if iFirmware < 0 then Continue; //PID not recognized, don't bother.

      hCyFile := CreateFileA(PAnsiChar(ThisLine), GENERIC_WRITE or GENERIC_READ, FILE_SHARE_WRITE or FILE_SHARE_READ, nil, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, 0);
      ThisLine := '';
      if hCyFile = INVALID_HANDLE_VALUE then hCyFile := 0;
      if hCyFile = 0 then Continue;

      //Transfer the hex file.

      //Put CPU into reset.
      CPUCSByte := $01;
      Status := _hCy_GenericVendorWrite(hCyFile, CUR_RAM_ACCESS, C_CPUCS_ADDRESS, 0, 1, @CPUCSByte);
      if Status = ERROR_SUCCESS then begin
        //Transfer each line.
        for iLine := 0 to High(PFD[iFirmware].FirmContent) do
          with PFD[iFirmware].FirmContent[iLine] do
            _hCy_GenericVendorWrite(hCyFile, CUR_RAM_ACCESS, Addr, 0, Length(Data), PAnsiChar(Data))
        ;
        //take CPU out of reset.
        CPUCSByte := $00;
        _hCy_GenericVendorWrite(hCyFile, CUR_RAM_ACCESS, C_CPUCS_ADDRESS, 0, 1, @CPUCSByte);
        Inc(Result);
      end;

      CloseHandle(hCyFile);
    end;
    SetLastError(ERROR_SUCCESS);
  except
    SetLastError(ERROR_INTERNAL_ERROR);
  end;
end;

function GetDeviceUniqueStr(DeviceIndex: LongWord; pIIDSize: PLongWord; pIID: PAnsiChar): LongWord; cdecl; // deprecated ?
var
  Content: Int64;

begin
  Result := GetDeviceSerialNumber(DeviceIndex, Content);
  if Result <> ERROR_SUCCESS then Exit;

  Move(Content, pIID^, Min(SizeOf(Content), pIIDSize^));
end;

function GetMappedFailure(DeviceIndex: LongWord): LongWord;
begin
  Result := GetLastError;
  case Result of
    ERROR_INVALID_HANDLE, ERROR_DEVICE_REMOVED, ERROR_DEVICE_NOT_CONNECTED, ERROR_FILE_NOT_FOUND: begin
      AIOUSB_CloseDevice(DeviceIndex);
      Dev[DeviceIndex].bDeviceWasHere := True;

      Dev[DeviceIndex].bOpen := False;
      Dev[DeviceIndex].bDACOpen := False;
      Dev[DeviceIndex].bDACClosing := False;
      Result := ERROR_DEVICE_REMOVED;
    end;
    ERROR_SUCCESS: Result := ERROR_GEN_FAILURE;
  end;
end;


function EnsureOpen(const DeviceIndex: LongWord): LongWord;
var
  Device: PDeviceData;
  DeviceInfo: USB_DEV_DSC;
  dwBytesReturned: LongWord;

begin
  Device := @Dev[DeviceIndex];

  if not Device.bGotHandle then begin
    if Device.bDeviceWasHere then
      Result := ERROR_DEVICE_REMOVED
    else
      Result := ERROR_FILE_NOT_FOUND
    ;
    Exit;
  end;

  if Device.bOpen then begin
    Result := ERROR_SUCCESS;
    Exit;
  end;

  Device.PID := $FFFFFFFF;
  if Device.bCyUSB then begin
    dwBytesReturned := SizeOf(DeviceInfo);

    Result := _Cy_GenericRead(DeviceIndex, $00, $06, $0100, $0000, dwBytesReturned, @DeviceInfo); // replace with call to CyUSB_GetDescriptor

    if dwBytesReturned < SizeOf(DeviceInfo) then Result := ERROR_READ_FAULT;
    if Result <> ERROR_SUCCESS then Exit;
    Device.PID := DeviceInfo.idProduct;
  end
  else begin
    dwBytesReturned := 0;
    if WinUSB_GetDescriptor(GetThreadHandleForDevice(DeviceIndex), USB_DEVICE_DESCRIPTOR_TYPE, 0, $0409, @DeviceInfo, SizeOf(DeviceInfo), dwBytesReturned) then
      Device.PID := DeviceInfo.idProduct
    else begin
      Result := GetMappedFailure(DeviceIndex);
      Exit;
    end;
  end;

  Device.DIOBytes := 0;
  Device.DIOConfigBits := 0;
  Device.Counters := 0;
  Device.RootClock := 0;
  Device.Tristates := 0;
  Device.bGetName := False;
  Device.ConfigBytes := 0;
  Device.bGateSelectable := False;
  Device.bDACBoardRange := False;
  Device.bDACChannelCal := False;
  Device.ImmDACs := 0;
  Device.ImmADCs := 0;
  Device.ADCChannels := 0;
  Device.ADCMUXChannels := 0;
  Device.ADCWorker := nil;
  Device.ADCContAcqWorker := nil;
  Device.ADCContBufWorker := nil;
  Device.pRingAcqData := nil;
  Device.bDACStream := False;
  Device.bDACDIOClock := False;
  Device.bADCBulk := False;
  Device.RangeShift := 0;
  Device.bDIOStream := False;
  Device.StreamingBlockSize := 31*1024;
  Device.bDIODebounce := False;
  Device.bDIOSPI := False;
  Device.bClearFIFO := False;
  Device.FlashSectors := 0;
  Device.WDGBytes := 0;
  Device.bSetCustomClocks := False;
  case Device.PID of
    $8001: begin //USB-DIO-32
      Device.DIOBytes := 4;
      Device.Tristates := 1;
      Device.Counters := 3;
      Device.RootClock := 3000000;
      Device.bGetName := True;
      Device.bSetCustomClocks := True;
      Device.bDIODebounce := True;
    end;
    $8004: begin //USB-DIO-32I
      Device.DIOBytes := 4;
      Device.DIOConfigBits := 32;
      Device.Tristates := 1;
      Device.bGetName := True;
      Device.bSetCustomClocks := True;
    end;
    $8005: begin //USB-DIO-24
      Device.DIOBytes := 3;
      Device.DIOConfigBits := 4;
      Device.Tristates := 1;
      Device.bGetName := True;
      Device.bDIODebounce := True;
    end;
    $8006: begin //USB-DIO-24-CTR6
      Device.DIOBytes := 3;
      Device.DIOConfigBits := 4;
      Device.Tristates := 1;
      Device.Counters := 2;
      Device.RootClock := 10000000;
      Device.bGetName := True;
      Device.bSetCustomClocks := True;
      Device.bDIODebounce := True;
    end;
    $8002: begin //USB-DIO-48
      Device.DIOBytes := 6;
      Device.Tristates := 1;
      Device.bGetName := True;
    end;
    $8003: begin //USB-DIO-96
      Device.DIOBytes := 12;
      Device.Tristates := 1;
      Device.bGetName := True;
    end;
    $8008, $8009, $800A: begin //USB-DIO16A family, old revs
      Device.DIOBytes := 1;
      Device.bGetName := True;
      Device.bDIOStream := True;
      Device.bDIOSPI := True;
      Device.bClearFIFO := True;
    end;
    $800C, $800D, $800E, $800F: begin //USB-DIO16A family, current revs
      Device.DIOBytes := 4;
      Device.Tristates := 2;
      Device.bGetName := True;
      Device.bDIOStream := True;
      Device.bDIOSPI := True;
      Device.bClearFIFO := True;
    end;
    $8068: begin //USB-AO-ARB1
      Device.DIOBytes := 4;
      Device.Tristates := 1;
      Device.bGetName := True;
      Device.bDIOStream := True;
      Device.bClearFIFO := True;
      Device.bDACDIOClock := True;
    end;
    $8010, $8011, $8012, $8014, $8015, $8016, //USB-IIRO-16 family
    $8018,        $801A, $801C,        $801E, //USB-IDIO-16 family
    $8019, $801D, $801F: begin //USB-II-16-OEM family
      Device.DIOBytes := 4;
      Device.Tristates := 1;
      Device.bGetName := True;
      Device.WDGBytes := 2;
    end;
    $4001, $4002: begin //USB-DA12-8A
      Device.bGetName := False;
      Device.bDACStream := True;
      Device.ImmDACs := 8;
      Device.DACsUsed := 5;
      Device.bGetName := True;
    end;
    $4003: begin //USB-DA12-8E (no streaming)
      Device.bGetName := False;
      Device.ImmDACs := 8;
      Device.bGetName := True;
    end;
    $8020: begin //USB-CTR-15
      Device.Counters := 5;
      Device.bGateSelectable := True;
      Device.RootClock := 10000000;
      Device.bGetName := True;
    end;
    $8030, $8031: begin //USB-IIRO4-2SM, USB-IIRO4-COM
      Device.DIOBytes := 2;
      Device.Tristates := 1;
      Device.bGetName := True;
    end;
    $8032: begin //USBP-DIO16RO8
      Device.DIOBytes := 3;
      Device.Tristates := 1;
      Device.bGetName := True;
    end;
    $8033: begin //PICO-DIO16RO8
      Device.DIOBytes := 3;
      Device.Tristates := 1;
      Device.bGetName := True;
    end;
    $8036: begin //USBP-II8IDO4
      Device.DIOBytes := 2;
      Device.Tristates := 1;
      Device.bGetName := True;
      Device.ImmADCs := 2;
    end;
    $8037: begin //PICO-II8IDO4
      Device.DIOBytes := 2;
      Device.Tristates := 1;
      Device.bGetName := True;
      Device.ImmADCs := 2;
    end;
    $803C, $803D, $803E: begin //USB-DIO48DO24, USB-DIO24DO12, ACCES USB-DO24
      Device.DIOBytes := 9;
      Device.Tristates := 1;
      Device.bGetName := True;
    end;
    $8040..$8044, $8140..$8144: begin //USB-AI(O)16-16 family
      Device.DIOBytes := 2;
      Device.Tristates := 1;
      Device.Counters := 1;
      Device.RootClock := 10000000;
      Device.bGetName := True;
      Device.bADCBulk := True;
      Device.ADCChannels := 16;
      Device.ADCMUXChannels := 16;
      Device.ConfigBytes := 20;
      Device.RangeShift := 0;
      Device.bClearFIFO := True;
      if (Device.PID and $0100) <> 0 then begin
        Device.bDACBoardRange := True;
        Device.ImmDACs := 2;
      end;
    end;
    $8045..$8049, $8145..$8149: begin //USB-AI(O)16-64 family
      Device.DIOBytes := 2;
      Device.Tristates := 1;
      Device.Counters := 1;
      Device.RootClock := 10000000;
      Device.bGetName := True;
      Device.bADCBulk := True;
      Device.ADCChannels := 16;
      Device.ADCMUXChannels := 64;
      Device.ConfigBytes := 21;
      Device.RangeShift := 2;
      Device.bClearFIFO := True;
      if (Device.PID and $0100) <> 0 then begin
        Device.bDACBoardRange := True;
        Device.ImmDACs := 2;
      end;
    end;
    $804A..$805F, $814A..$815F: begin //USB-AI(O)16-32+ family
      Device.DIOBytes := 2;
      Device.Tristates := 1;
      Device.Counters := 1;
      Device.RootClock := 10000000;
      Device.bGetName := True;
      Device.bADCBulk := True;
      Device.ADCChannels := 16;
      Device.ADCMUXChannels := 32 * ((((Device.PID - $804A) and not $0100) div 5) + 1);
      Device.ConfigBytes := 21;
      Device.RangeShift := 3;
      Device.bClearFIFO := True;
      if (Device.PID and $0100) <> 0 then begin
        Device.bDACBoardRange := True;
        Device.ImmDACs := 2;
      end;
    end;
    $8060, $8070..$807F: begin //USB-AO16-16A family
      Device.DIOBytes := 2;
      Device.Tristates := 1;
      Device.bGetName := True;
      Device.FlashSectors := 32;
      Device.bDACBoardRange := True;
      Device.bDACChannelCal := True;
      //Device.bClearFIFO := True;
      //Add a new-style DAC streaming
      case Device.PID and $6 of
        $0: Device.ImmDACs := 16;
        $2: Device.ImmDACs := 12;
        $4: Device.ImmDACs := 8;
        $6: Device.ImmDACs := 4;
      end;
      if (Device.PID and 1) = 0 then Device.ImmADCs := 2;
    end;
    else begin
      //Device.bADCStream := True; Result := ERROR_SUCCESS;
      //Device.bDIOStream := True; Result := ERROR_SUCCESS;
      //Device.bDIOSPI := True; Result := ERROR_SUCCESS;
    end;
  end;
  if Device.DIOConfigBits = 0 then Device.DIOConfigBits := Device.DIOBytes;

  SetLength(Device.LastDIOData, Device.DIOBytes);

  Device.bOpen := True;
  Result := ERROR_SUCCESS;
end;


function ValidateAndEnsureOpen(var DeviceIndex: LongWord): LongWord;
begin
  Result := Validate(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  Result := EnsureOpen(DeviceIndex);
end;

{
  Modern firmwares special-case a RAM READ operation from address $80xx, IFF Index == $0001;
  it can be used to check various firmware version-specific stuff as follows:
  read 3 bytes: 	byte 0 returns "3", the revision of this structure
  	byte 1 returns "1" on the USB-AI family to indicate Rev E PCB (AIO-capable, bipolar calibration capable)
  	byte 2 returns "3"
  in revision "3", byte 2 bit 0 is "supports GenRegister features"
                          bit 1 is "supports Firmware 2.0 features"
}
function CheckFirmware20(DeviceIndex: LongWord): LongWord;

var
  L: LongWord;
  MemFlags: array[0..2] of Byte; // rev 3 struct has 3 bytes total
const
  bmFWE20 = $02;

begin
  if Dev[DeviceIndex].bFirmware20 then begin // don't bother checking more than once.
    Result := ERROR_SUCCESS;
    Exit;
  end;

  //Read out the first chunk of "memory" flags.
  L := Length(MemFlags);
  Result := GenericVendorRead(DeviceIndex, CUR_RAM_READ, C_RAM_REG_ADDRESS, C_RAM_REG_CHECK_INDEX, L, @MemFlags[0]);
  if Result <> ERROR_SUCCESS then Exit;
  Result := ERROR_NOT_SUPPORTED;
  if LongInt(L) < Length(MemFlags) then Exit;
  if MemFlags[0] = C_RAM_REG_CHECK_UNSUPPORTED then Exit;
  if MemFlags[0] < C_RAM_REG_CHECK_REVISION then Exit;

  if (MemFlags[2] and bmFWE20) = 0 then Exit;

  Result := ERROR_SUCCESS;
  Dev[DeviceIndex].bFirmware20 := True;
end;


function DEB_BulkIn(DeviceIndex: LongWord): LongWord; cdecl;
begin
  Result := ERROR_NOT_SUPPORTED;
end;

function DEB_BulkOut(DeviceIndex, TargetPipe: LongWord; pBuf: Pointer; BufLen: LongWord): LongWord; cdecl;
var
  L: LongWord;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    L := 0;
    Result := AWU_GenericBulkOut(DeviceIndex, TargetPipe, pBuf, BufLen, L);
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function GetThreadHandleForDevice(DeviceIndex: LongWord): THandle;
var
  I: Integer;
  ThreadID: LongWord;

begin
  with Dev[DeviceIndex] do
    if bCyUSB then begin
      ThreadID := GetCurrentThreadId;

      WaitForSingleObject(hHandleTableMutex, INFINITE);
        Result := INVALID_HANDLE_VALUE;
        for I := 0 to High(HandleTable) do
          if HandleTable[I].ThreadID = ThreadID then begin
            Result := HandleTable[I].hDevice;
            Break;
          end
        ;
        if Result = INVALID_HANDLE_VALUE then begin
          I := Length(HandleTable);
          SetLength(HandleTable, I+1);
          HandleTable[I].ThreadID := ThreadID;
          HandleTable[I].hDevice := CreateFileA(PAnsiChar(DevPath), GENERIC_WRITE or GENERIC_READ, FILE_SHARE_WRITE or FILE_SHARE_READ, nil, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, 0);
          Result := HandleTable[I].hDevice;
        end;
      ReleaseMutex(hHandleTableMutex);
    end
    else begin
      WaitForSingleObject(hHandleTableMutex, INFINITE);
        if Length(HandleTable) <> 0 then
          Result := HandleTable[0].hDevice
        else begin
          SetLength(HandleTable, 1);
          HandleTable[0].ThreadID := 0;
          WinUSB_Initialize(hWinUSBFile, HandleTable[0].hDevice);
          Result := HandleTable[0].hDevice;
        end;
      ReleaseMutex(hHandleTableMutex);
    end
  ;
end;

procedure AddThreadHandleForDevice(DeviceIndex: LongWord; hDevice: THandle);
var
  I: Integer;

begin
  with Dev[DeviceIndex] do begin
    WaitForSingleObject(hHandleTableMutex, INFINITE);
      I := Length(HandleTable);
      SetLength(HandleTable, I+1);
      HandleTable[I].ThreadID := GetCurrentThreadId;
      HandleTable[I].hDevice := hDevice;
    ReleaseMutex(hHandleTableMutex);
  end;
end;

function GetDevicePaths(const ClassGUID: TGUID): TStringArray;
var
  DevInterfaceData: SP_DEVICE_INTERFACE_DATA;
  MemberIndex: Integer;
  hDevInfo: THandle;

  L: LongWord;
  pDetailData: PSP_DEVICE_INTERFACE_DETAIL_DATA;

  Buf: AnsiString;
  I, RL: Integer;

begin
  RL := 0;
  SetLength(Result, RL);

  hDevInfo := SetupDiGetClassDevs(@ClassGUID, nil, 0, DIGCF_PRESENT or DIGCF_DEVICEINTERFACE);
  if hDevInfo = 0 then Exit;

  DevInterfaceData.cbSize := sizeof(DevInterfaceData);
  MemberIndex := 0;
  while SetupDiEnumDeviceInterfaces(hDevInfo, nil, ClassGUID, MemberIndex, DevInterfaceData) do begin
    L := 0;
    SetupDiGetDeviceInterfaceDetail(hDevInfo, @DevInterfaceData, nil, 0, @L, nil);

    GetMem(pDetailData, L);
    if SizeOf(Pointer) = 8 then
      pDetailData.cbSize := 8 //The fixed portion is 4. Dunno where 8 comes from.
    else
      pDetailData.cbSize := 5 //The fixed portion is 4. Dunno where 5 comes from.
    ;

    if not SetupDiGetDeviceInterfaceDetail(hDevInfo, @DevInterfaceData, pDetailData, L, nil, nil) then begin
      FreeMem(pDetailData);
      Inc(MemberIndex);
      Continue;
    end;

    SetString(Buf, PAnsiChar(@pDetailData.DevicePath[0]), L - 4);
    FreeMem(pDetailData);
    I := Pos(AnsiChar(#0), Buf);
    if I <> 0 then SetLength(Buf, I - 1);

    SetLength(Result, RL + 1);
    Result[RL] := Buf;
    Inc(RL);
    Buf := '';

    Inc(MemberIndex);
  end;
  SetupDiDestroyDeviceInfoList(hDevInfo);
end;

procedure LoadDriverLinks;
var
  DevicePath: TStringArray;
  ThisPath: AnsiString;
  PI, DI: Integer;
  hFile, hWinUSB: THandle;
  //SetupData: TUSBSetupPacket;
  //ThisSN: Int64;
  //BytesRead: LongWord;
  bFull: Boolean;

begin
  //May want to alter these to not open handles until GetThreadHandleForDevice.

  //CyUSB links.
  DevicePath := GetDevicePaths(AIOCyUSBClassGUID);
  for PI := 0 to High(DevicePath) do begin
    ThisPath := DevicePath[PI];
    for DI := 0 to High(Dev) do
      if Dev[DI].bGotHandle then
        if Dev[DI].DevPath = ThisPath then begin
          ThisPath := '';
          Break;
        end
    ;
    if Length(ThisPath) = 0 then Continue;

    hFile := CreateFileA(PAnsiChar(ThisPath), GENERIC_WRITE or GENERIC_READ, FILE_SHARE_WRITE or FILE_SHARE_READ, nil, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, 0);
    ThisPath := '';
    if hFile = INVALID_HANDLE_VALUE then hFile := 0;
    if hFile = 0 then Continue;

    bFull := True;
    for DI := 0 to High(Dev) do with Dev[DI] do
      if (not bGotHandle) and not bDeviceWasHere then begin
        AddThreadHandleForDevice(DI, hFile);
        bGotHandle := True;
        hWinUSBFile := 0;
        bCyUSB := True;
        DevPath := DevicePath[PI];

        bOpen := False;
        bDACOpen := False;
        bDACClosing := False;

        bFull := False;
        Break;
      end
    ;
    if bFull then begin
      CloseHandle(hFile);
      Break;
    end;
  end;
  SetLength(DevicePath, 0);

  //WinUSB links.
  if bWinUSBLoaded then begin
    DevicePath := GetDevicePaths(AIOWinUSBClassGUID);
    for PI := 0 to High(DevicePath) do begin
      hFile := CreateFileA(PAnsiChar(DevicePath[PI]), GENERIC_WRITE or GENERIC_READ, 0, nil, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, 0);
      DevicePath[PI] := '';
      if hFile = INVALID_HANDLE_VALUE then hFile := 0;
      if hFile = 0 then Continue;

      if not WinUSB_Initialize(hFile, hWinUSB) then begin
        CloseHandle(hFile);
        Continue;
      end;

      {
      if not CheckVID(hWinUSB) then begin
        WinUSB_Free(hWinUSB);
        CloseHandle(hFile);
        Continue;
      end;

      SetupData.RequestType := $C0;
      SetupData.Request := $A2;
      SetupData.Value := $1DF8;
      SetupData.Index := 0;
      SetupData.DataLength := SizeOf(ThisSN);
      if (not WinUSB_ControlTransfer(hWinUSB, SetupData, @ThisSN, SizeOf(ThisSN), BytesRead, nil)) or (BytesRead <> SizeOf(ThisSN)) then begin
        WinUSB_Free(hWinUSB);
        CloseHandle(hFile);
        Continue;
      end;
      }

      bFull := True;
      DI := 0;
      while DI < Length(Dev) do with Dev[DI] do
        if (not bGotHandle) and not bDeviceWasHere then begin
          AddThreadHandleForDevice(DI, hWinUSB);
          bGotHandle := True;
          hWinUSBFile := hFile;
          bCyUSB := False;

          bOpen := False;
          bDACOpen := False;
          bDACClosing := False;

          bFull := False;
          Break;
        end
        else
          Inc(DI)
      ;
      if bFull then begin
        WinUSB_Free(hWinUSB);
        CloseHandle(hFile);
        Break;
      end;
    end;
    SetLength(DevicePath, 0);
  end;
end;


function GenericVendorRead(DeviceIndex: LongWord; Request: Byte; Value, Index: Word; var DataSize: LongWord; pData: Pointer): LongWord; cdecl;
begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  Result := _GenericVendorRead(DeviceIndex, Request, Value, Index, DataSize, pData, Dev[DeviceIndex].bCyUSB);
end;

function GenericVendorWrite(DeviceIndex: LongWord; Request: Byte; Value, Index: Word; DataSize: LongWord; pData: Pointer): LongWord; cdecl;
begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  Result := _GenericVendorWrite(DeviceIndex, Request, Value, Index, DataSize, pData, Dev[DeviceIndex].bCyUSB);
end;


function DEB_SetInterface(DeviceIndex: LongWord; TargetInterface: Word): LongWord; cdecl;
begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    //Result := DoSetIFace(Dev[DeviceIndex].Handle, TargetInterface);
    Result := ERROR_NOT_SUPPORTED;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function AIOUSB_ResetChip(DeviceIndex: LongWord): LongWord; cdecl;
var
  CPUCSByte: Byte;

begin
  try
    //Put CPU into reset
    CPUCSByte := $01;
    Result := GenericVendorWrite(DeviceIndex, $A0, $E600, 0, 1, @CPUCSByte);
    if Result <> ERROR_SUCCESS then Exit;

    //take CPU out of reset
    CPUCSByte := $00;
    Result := GenericVendorWrite(DeviceIndex, $A0, $E600, 0, 1, @CPUCSByte);
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function AWU_GenericBulkOut(DeviceIndex: LongWord; PipeID: LongWord; pData: Pointer; DataSize: LongWord; var BytesWritten: LongWord): LongWord; cdecl;
begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  Result := _GenericBulkOut(DeviceIndex, PipeID, pData, DataSize, BytesWritten);
end;

function AWU_GenericBulkIn(DeviceIndex: LongWord; PipeID: LongWord; pData: Pointer; DataSize: LongWord; var BytesRead: LongWord): LongWord; cdecl;
begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  Result := _GenericBulkIn(DeviceIndex, PipeID, pData, DataSize, BytesRead);
end;

end.
