unit DriverLevel;

interface

function _GenericVendorRead(DeviceIndex: LongWord; Request: Byte; Value, Index: Word; var DataSize: LongWord; pData: Pointer; bCyUSB: Boolean): LongWord; cdecl;
function _GenericVendorWrite(DeviceIndex: LongWord; Request: Byte; Value, Index: Word; DataSize: LongWord; pData: Pointer; bCyUSB: Boolean): LongWord; cdecl;
function _GenericBulkIn(DeviceIndex: LongWord; PipeID: LongWord; pData: Pointer; DataSize: LongWord; var BytesRead: LongWord): LongWord;
function _GenericBulkOut(DeviceIndex: LongWord; PipeID: LongWord; pData: Pointer; DataSize: LongWord; var BytesWritten: LongWord): LongWord; cdecl;


implementation

uses
  Windows, Math,
  AIOUSB_Constants,
  CyUSB,
  WinUSB,
  DeviceLevel;
                            
function _GenericVendorRead(DeviceIndex: LongWord; Request: Byte; Value, Index: Word; var DataSize: LongWord; pData: Pointer; bCyUSB: Boolean): LongWord; cdecl;
var
  SetupData: TWinUSBSetupPacket;
  Buf: AnsiString;
  pCT: PCySingleTransfer;
  BytesReturned: LongWord;

begin
  try
    SetString(Buf, PAnsiChar(pData), DataSize);
  except
    Result := ERROR_NOACCESS;
    Exit;
  end;

  if bCyUSB then begin //CyUSB.
    Result := _Cy_GenericRead(DeviceIndex, C_ENDPOINT_ZERO_READ, Request, Value, Index, DataSize, pData);
  end
  else begin //WinUSB.  // replace with _WinUSB_GenericRead() -- J2H
    SetupData.RequestType := C_ENDPOINT_ZERO_READ;
    SetupData.Request := Request;
    SetupData.Value := Value;
    SetupData.Index := Index;
    SetupData.DataLength := DataSize;
    if WinUSB_ControlTransfer(GetThreadHandleForDevice(DeviceIndex), SetupData, PByte(PAnsiChar(Buf)), DataSize, BytesReturned, nil) then begin
      Result := ERROR_SUCCESS;
      DataSize := Min(BytesReturned, Length(Buf));
      try
        Move(Buf[1], pData^, DataSize);
      except
        Result := ERROR_NOACCESS;
      end;
    end
    else
      Result := GetMappedFailure(DeviceIndex)
    ;
  end;
end;

function _GenericVendorWrite(DeviceIndex: LongWord; Request: Byte; Value, Index: Word; DataSize: LongWord; pData: Pointer; bCyUSB: Boolean): LongWord; cdecl;
var
  SetupData: TWinUSBSetupPacket;
  Buf: AnsiString;
  pCT: PCySingleTransfer;
  BytesReturned: LongWord;

begin
  try
    SetString(Buf, PAnsiChar(pData), DataSize);
  except
    Result := ERROR_NOACCESS;
    Exit;
  end;

  if bCyUSB then begin //CyUSB.
    Result := _hCy_GenericVendorWrite(GetThreadHandleForDevice(DeviceIndex), Request, Value, Index, DataSize, pData);
  end
  else begin //WinUSB.
    SetupData.RequestType := C_ENDPOINT_ZERO_WRITE;
    SetupData.Request := Request;
    SetupData.Value := Value;
    SetupData.Index := Index;
    SetupData.DataLength := DataSize;
    if WinUSB_ControlTransfer(GetThreadHandleForDevice(DeviceIndex), SetupData, PByte(PAnsiChar(Buf)), DataSize, DataSize, nil) then
      Result := ERROR_SUCCESS
    else
      Result := GetMappedFailure(DeviceIndex)
    ;
  end;
end;

function _GenericBulkIn(DeviceIndex: LongWord; PipeID: LongWord; pData: Pointer; DataSize: LongWord; var BytesRead: LongWord): LongWord;
var
  CT: TCySingleTransfer;

begin
  if Dev[DeviceIndex].bCyUSB then begin
    FillChar(CT, SizeOf(CT), 0);
    CT.ucEndpointAddress := PipeID;
    if DeviceIoControl(GetThreadHandleForDevice(DeviceIndex), IOCTL_ADAPT_SEND_NON_EP0_DIRECT, @CT, SizeOf(CT), pData, DataSize, BytesRead, nil) then begin
      Result := ERROR_SUCCESS;
      if BytesRead > DataSize then BytesRead := DataSize;
    end
    else
      Result := GetMappedFailure(DeviceIndex)
    ;
  end
  else begin
    SetLastError(ERROR_GEN_FAILURE);
    if WinUsb_ReadPipe(GetThreadHandleForDevice(DeviceIndex), PipeID, pData, DataSize, BytesRead, nil) then
      Result := ERROR_SUCCESS
    else
      Result := GetMappedFailure(DeviceIndex)
    ;
  end;
end;


function _GenericBulkOut(DeviceIndex: LongWord; PipeID: LongWord; pData: Pointer; DataSize: LongWord; var BytesWritten: LongWord): LongWord; cdecl;
var
  CT: TCySingleTransfer;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  if Dev[DeviceIndex].bCyUSB then begin
    FillChar(CT, SizeOf(CT), 0);
    CT.ucEndpointAddress := PipeID;
    if DeviceIoControl(GetThreadHandleForDevice(DeviceIndex), IOCTL_ADAPT_SEND_NON_EP0_DIRECT, @CT, SizeOf(CT), pData, DataSize, BytesWritten, nil) then begin
      if BytesWritten > DataSize then BytesWritten := DataSize;
    end
    else
      Result := GetMappedFailure(DeviceIndex)
    ;
  end
  else begin
    SetLastError(ERROR_GEN_FAILURE);
    if WinUsb_WritePipe(GetThreadHandleForDevice(DeviceIndex), PipeID, pData, DataSize, BytesWritten, nil) then begin
      Result := ERROR_SUCCESS;
    end
    else
      Result := GetMappedFailure(DeviceIndex)
    ;
  end;
end;

end.

