unit EEPROM;

interface

const
  C_CUSTOM_EEPROM_SIZE = $100;


function CustomEEPROMRead(DeviceIndex: LongWord; StartAddress: LongWord; var DataSize: LongWord; Data: Pointer): LongWord; cdecl;
function CustomEEPROMWrite(DeviceIndex: LongWord; StartAddress: LongWord; DataSize: LongWord; Data: Pointer): LongWord; cdecl;
function GetEEPROMSerialNumber(DeviceIndex: LongWord; var pSerialNumber: Int64): LongWord; cdecl;

implementation

Uses
  Windows,
  AIOUSB_Constants,
  VRTable,
  DeviceLevel;

const
  C_CUSTOM_EEPROM_BLOCK_SIZE = 32;
  C_CUSTOM_EERPOM_START_ADDRESS = $1E00;
  C_SERIALNUMBER_ADDRESS = $1DF8;
  
function CustomEEPROMRead(DeviceIndex: LongWord; StartAddress: LongWord; var DataSize: LongWord; Data: Pointer): LongWord; cdecl;
begin
  if StartAddress >= C_CUSTOM_EEPROM_SIZE then begin
    Result := ERROR_SEEK;
    Exit;
  end;
  if DataSize > (LongWord(C_CUSTOM_EEPROM_SIZE) - StartAddress) then begin
    Result := ERROR_BAD_LENGTH;
    Exit;
  end;

  Result := GenericVendorRead(DeviceIndex, CUR_EEPROM_ACCESS, C_CUSTOM_EERPOM_START_ADDRESS + StartAddress, 0, DataSize, Data);
end;

function CustomEEPROMWrite(DeviceIndex: LongWord; StartAddress: LongWord; DataSize: LongWord; Data: Pointer): LongWord; cdecl;
begin
  if StartAddress >= C_CUSTOM_EEPROM_SIZE then begin
    Result := ERROR_SEEK;
    Exit;
  end;
  if DataSize > (LongWord(C_CUSTOM_EEPROM_SIZE) - StartAddress) then begin
    Result := ERROR_BAD_LENGTH;
    Exit;
  end;

  while DataSize > C_CUSTOM_EEPROM_BLOCK_SIZE do begin
    Result := GenericVendorWrite(DeviceIndex, CUR_EEPROM_ACCESS, C_CUSTOM_EERPOM_START_ADDRESS + StartAddress, 0, C_CUSTOM_EEPROM_BLOCK_SIZE, Data);
    if Result <> ERROR_SUCCESS then Exit;
    Inc(StartAddress, C_CUSTOM_EEPROM_BLOCK_SIZE);
    Dec(DataSize, C_CUSTOM_EEPROM_BLOCK_SIZE);
    Inc(PByte(Data), C_CUSTOM_EEPROM_BLOCK_SIZE);
  end;
  Result := GenericVendorWrite(DeviceIndex, CUR_EEPROM_ACCESS, C_CUSTOM_EERPOM_START_ADDRESS + StartAddress, 0, DataSize, Data);
end;

function GetEEPROMSerialNumber(DeviceIndex: LongWord; var pSerialNumber: Int64): LongWord; cdecl;
var
  L: LongWord;
  Content: Int64;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  L := SizeOf(Content);
  Result := GenericVendorRead(DeviceIndex, CUR_EEPROM_ACCESS, C_SERIALNUMBER_ADDRESS, 0, L, @Content);

  if (Result = ERROR_SUCCESS) and (L = SizeOf(Content)) then begin
    if Content = 0 then Content := -1;
    pSerialNumber := Content;
    if Content = -1 then Result := ERROR_NOT_SUPPORTED;
  end
  else begin
    Result := GetMappedFailure(DeviceIndex);
  end;
end;

end.

