unit FWE20;
interface

uses
  Windows;

function AIOUSB_SaveStateAsDefaults(DeviceIndex: LongWord): LongWord; cdecl;
function AIOUSB_RestoreFactoryDefaults(DeviceIndex: LongWord): LongWord; cdecl;
function AIOUSB_SetDefaultsTable(DeviceIndex: LongWord; pDefTable: Pointer; pDefTableBytes: PLongWord): LongWord; cdecl;
function AIOUSB_GetDefaultsTable(DeviceIndex: LongWord; pDefTable: Pointer; pDefTableBytes: PLongWord): LongWord; cdecl;

function AIOUSB_SetGlobalTickRate(DeviceIndex: LongWord; pHz: PDouble): LongWord; cdecl;

function DIO_Automap_ClearTable(DeviceIndex: LongWord): LongWord; cdecl;
//function DIO_Automap_GetTable(DeviceIndex: LongWord; pMapTable: Pointer; pMapTableBytes: PLongWord): LongWord; cdecl;
function DIO_Automap_SetTable(DeviceIndex: LongWord; pMapTable: Pointer; pMapTableBytes: PLongWord): LongWord; cdecl;
function DIO_Automap_AddEntry(DeviceIndex: LongWord; pMapEntry: PWord): LongWord; cdecl;
function DIO_Automap_SetTickDivisor(DeviceIndex: LongWord; pDivisor: PByte): LongWord; cdecl;

//function DIO_PWM_GetAll(DeviceIndex: LongWord; pDutyCycles: Pointer; pDutyCyclesBytes: PLongWord): LongWord; cdecl;
function DIO_PWM_SetAll(DeviceIndex: LongWord; pDivisor: PByte; pPWMTable: Pointer; pPWMTableBytes: PLongWord): LongWord; cdecl;
//function DIO_PWM_Clear(DeviceIndex: LongWord): LongWord; cdecl;
//function DIO_PWM_Add(DeviceIndex: LongWord; BitIndex: LongWord; DutyCycle: LongWord): LongWord; cdecl;
//function PWM_SetGlobalTicksPerPWMPeriod(DeviceIndex: LongWord; ): LongWord; cdecl;

function DIO_Latch_SetDivisor(DeviceIndex: LongWord; pDivisor: PByte): LongWord; cdecl;
function DIO_Latch_Reset(DeviceIndex: LongWord): LongWord; cdecl;
function DIO_Latch_Read(DeviceIndex: LongWord; pLowLatches, pHighLatches, pNoLatches: PByte; pLatchesBytes: PLongWord): LongWord; cdecl;

function DIO_Deb_SetConfig(DeviceIndex: LongWord; pDebTable: Pointer; pDebTableBytes: PLongWord): LongWord; cdecl;
function DIO_Deb_ReadAll(DeviceIndex: LongWord; pDebData: PByte): LongWord; cdecl;
function WDG_SetConfig(DeviceIndex: LongWord; pTimeoutSeconds: PDouble; pWDGTable: Pointer; pWDGTableBytes: PLongWord): LongWord; cdecl;
function WDG_GetStatus(DeviceIndex: LongWord; pStatus: Pointer; pStatusBytes: PLongWord): LongWord; cdecl;
function WDG_Pet(DeviceIndex: LongWord; PetFlag: LongWord): LongWord; cdecl;

implementation
Uses
  Math,
  DeviceLevel,
  AIOUSB_Constants,
  VRTable;


function SetWatchdog(DeviceIndex: LongWord; var Delay: Double; pData: Pointer): LongWord; cdecl;
var
  DV: Word;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if WDGBytes = 0 then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    DV := Max(1, Round(Delay));
    Delay := DV;
    Result := GenericVendorWrite(DeviceIndex, AUR_WDG, 0, 0, WDGBytes, pData);
  end;
end;


//Load-time defaults.
function AIOUSB_SaveStateAsDefaults(DeviceIndex: LongWord): LongWord; cdecl;
begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    Result := CheckFirmware20 (DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    Result := GenericVendorWrite(DeviceIndex, AUR_GEN_REGISTER, $ACCE, $0010, 0, nil);
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function AIOUSB_RestoreFactoryDefaults(DeviceIndex: LongWord): LongWord; cdecl;
begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    Result := CheckFirmware20(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    Result := GenericVendorWrite(DeviceIndex, AUR_GEN_REGISTER, $ACCE, $0011, 0, nil);
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function AIOUSB_SetDefaultsTable(DeviceIndex: LongWord; pDefTable: Pointer; pDefTableBytes: PLongWord): LongWord; cdecl;
var
  DefBuf: array of Byte;
  L: LongWord;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    Result := CheckFirmware20(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    L := pDefTableBytes^;
    if L > 1024 then begin
      Result := ERROR_BAD_LENGTH;
      Exit;
    end;

    SetLength(DefBuf, L);
    try
      Move(pDefTable^, DefBuf[0], L);
    except
      Result := ERROR_NOACCESS;
      Exit;
    end;

    Result := GenericVendorWrite(DeviceIndex, AUR_GEN_REGISTER, $ACCE, $0022, L, @DefBuf[0]);
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function AIOUSB_GetDefaultsTable(DeviceIndex: LongWord; pDefTable: Pointer; pDefTableBytes: PLongWord): LongWord; cdecl;
var
  DefBuf: array of Byte;
  L: LongWord;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    Result := CheckFirmware20(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    L := pDefTableBytes^;
    if L > 1024 then begin
      Result := ERROR_BAD_LENGTH;
      Exit;
    end;

    SetLength(DefBuf, L);
    try
      Move(pDefTable^, DefBuf[0], L);
    except
      Result := ERROR_NOACCESS;
      Exit;
    end;

    Result := GenericVendorRead(DeviceIndex, AUR_GEN_REGISTER, $0000, $0020, L, @DefBuf[0]);
    if L > LongWord(Length(DefBuf)) then L := Length(DefBuf);
    Move(DefBuf[0], pDefTable^, L);
    pDefTableBytes^ := L;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;


//Global tick.
function AIOUSB_SetGlobalTickRate(DeviceIndex: LongWord; pHz: PDouble): LongWord; cdecl;
var
  Hz, fDivisor: Double;
  Divisor: Word;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    Result := CheckFirmware20(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    Hz := pHz^;
    if Hz = 0 then
      Divisor := $FFFF //Will load 0 into the counter.
    else begin
      fDivisor := 12000000 / Hz; //12MHz is the CPU clock.
      if fDivisor <= 0 then
        Divisor := 0
      else if fDivisor >= $FFFF then
        Divisor := $FFFF
      else
        Divisor := Round(fDivisor)
      ;
    end;

    Result := GenericVendorWrite(DeviceIndex, AUR_GEN_REGISTER, $FFFF - Divisor, $8001, 0, nil);
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;


//Watchdog.
function WDG_SetConfig(DeviceIndex: LongWord; pTimeoutSeconds: PDouble; pWDGTable: Pointer; pWDGTableBytes: PLongWord): LongWord; cdecl;
var
  WDGBuf: array of Byte;
  fTimeout: Double;
  Timeout: Word;
  L: LongWord;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    Result := CheckFirmware20(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    fTimeout := pTimeoutSeconds^;
    if fTimeout <= 0 then
      Timeout := 0
    else if fTimeout >= $FFFF then
      Timeout := $FFFF
    else
      Timeout := Round(fTimeout)
    ;

    L := pWDGTableBytes^;
    if L > 1024 then begin
      Result := ERROR_BAD_LENGTH;
      Exit;
    end;

    SetLength(WDGBuf, 2 + L);
    PWord(@WDGBuf[0])^ := Timeout;
    try
      Move(pWDGTable^, WDGBuf[2], L);
    except
      Result := ERROR_NOACCESS;
      Exit;
    end;

    Result := GenericVendorWrite(DeviceIndex, AUR_GEN_REGISTER, $0000, $0041, 2 + L, @WDGBuf[0]);
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function WDG_GetStatus(DeviceIndex: LongWord; pStatus: Pointer; pStatusBytes: PLongWord): LongWord; cdecl;
var
  WDGBuf: array of Byte;
  L: LongWord;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    Result := CheckFirmware20(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    L := pStatusBytes^;
    if L > 1024 then begin
      Result := ERROR_BAD_LENGTH;
      Exit;
    end;

    SetLength(WDGBuf, L);
    try
      Move(pStatus^, WDGBuf[0], L);
    except
      Result := ERROR_NOACCESS;
      Exit;
    end;

    Result := GenericVendorRead(DeviceIndex, AUR_WDG_STATUS, $0000, $0041, L, @WDGBuf[0]);
    if L > LongWord(Length(WDGBuf)) then L := Length(WDGBuf);
    Move(WDGBuf[0], pStatus^, L);
    pStatusBytes^ := L;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function WDG_Pet(DeviceIndex: LongWord; PetFlag: LongWord): LongWord; cdecl;
var
  TimeoutLeft: Word;
  L: LongWord;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    Result := CheckFirmware20(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    case PetFlag of
      0: begin
        //Read a little status and throw it away, because we can't read zero bytes.
        L := SizeOf(TimeoutLeft);
        Result := GenericVendorRead(DeviceIndex, AUR_WDG_STATUS, $0000, $0041, L, @TimeoutLeft);
      end;
      else begin
        Result := ERROR_INVALID_FLAG_NUMBER;
      end;
    end;

  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;


//DIO automapping.
function DIO_Automap_ClearTable(DeviceIndex: LongWord): LongWord; cdecl;
begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    Result := CheckFirmware20(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    Result := GenericVendorWrite(DeviceIndex, AUR_GEN_REGISTER, $0000, $0082, 0, nil);
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function DIO_Automap_SetTable(DeviceIndex: LongWord; pMapTable: Pointer; pMapTableBytes: PLongWord): LongWord; cdecl;
const
  MapCount = 32;

var
  AutomapBuf: array of Byte;
  L: LongWord;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    Result := CheckFirmware20(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    L := pMapTableBytes^;
    if L > 2*MapCount then begin
      Result := ERROR_BAD_LENGTH;
      Exit;
    end;

    if (L and 1) <> 0 then begin
      L := L and not 1;
      pMapTableBytes^ := L;
    end;

    SetLength(AutomapBuf, 2*MapCount + 1);
    FillChar(AutomapBuf[0], 2*MapCount, 0);
    try
      Move(pMapTable^, AutomapBuf[0], L);
    except
      Result := ERROR_NOACCESS;
      Exit;
    end;
    AutomapBuf[2*MapCount] := L div 2;

    Result := GenericVendorWrite(DeviceIndex, AUR_GEN_REGISTER, $0000, $0081, L, @AutomapBuf[0]);
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function DIO_Automap_AddEntry(DeviceIndex: LongWord; pMapEntry: PWord): LongWord; cdecl;
var
  MapEntry: Word;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    Result := CheckFirmware20(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    try
      MapEntry := pMapEntry^;
    except
      Result := ERROR_NOACCESS;
      Exit;
    end;

    Result := GenericVendorWrite(DeviceIndex, AUR_GEN_REGISTER, $0000, $0084, SizeOf(MapEntry), @MapEntry);
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function DIO_Automap_SetTickDivisor(DeviceIndex: LongWord; pDivisor: PByte): LongWord; cdecl;
var
  Divisor: Byte;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    Result := CheckFirmware20(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    try
      Divisor := pDivisor^;
    except
      Result := ERROR_NOACCESS;
      Exit;
    end;

    Result := GenericVendorWrite(DeviceIndex, AUR_GEN_REGISTER, $0000, $0083, SizeOf(Divisor), @Divisor);
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;


//DIO PWM.
function DIO_PWM_SetAll(DeviceIndex: LongWord; pDivisor: PByte; pPWMTable: Pointer; pPWMTableBytes: PLongWord): LongWord; cdecl;
const
  PWMCount = 8;

var
  PWMBuf: array of Byte;
  L: LongWord;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    Result := CheckFirmware20(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    L := pPWMTableBytes^;
    if L > 2*PWMCount then begin
      Result := ERROR_BAD_LENGTH;
      Exit;
    end;

    if (L and 1) <> 0 then begin
      L := L and not 1;
      pPWMTableBytes^ := L;
    end;

    SetLength(PWMBuf, 2*PWMCount + 2);
    FillChar(PWMBuf[1], 2*PWMCount, 0);
    try
      Move(pPWMTable^, PWMBuf[1], L);
    except
      Result := ERROR_NOACCESS;
      Exit;
    end;
    PWMBuf[2*PWMCount + 1] := L div 2;

    try
      PWMBuf[0] := pDivisor^;
    except
      Result := ERROR_NOACCESS;
      Exit;
    end;

    Result := GenericVendorWrite(DeviceIndex, AUR_GEN_REGISTER, $0000, $0071, L, @PWMBuf[0]);
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

//DIO latching.
function DIO_Latch_SetDivisor(DeviceIndex: LongWord; pDivisor: PByte): LongWord; cdecl;
var
  Divisor: Byte;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    Result := CheckFirmware20(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    try
      Divisor := pDivisor^;
    except
      Result := ERROR_NOACCESS;
      Exit;
    end;

    Result := GenericVendorWrite(DeviceIndex, AUR_GEN_REGISTER, $0000, $0063, SizeOf(Divisor), @Divisor);
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function DIO_Latch_Reset(DeviceIndex: LongWord): LongWord; cdecl;
var
  L: LongWord;
  LatchBuf: array of Byte;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    Result := CheckFirmware20(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    SetLength(LatchBuf, Dev[DeviceIndex].DIOBytes * 3);

    L := Length(LatchBuf);
    Result := GenericVendorRead(DeviceIndex, AUR_DIO_LATCH_READ, $0000, $0000, L, @LatchBuf[0]);
    if Result <> ERROR_SUCCESS then Exit;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function DIO_Latch_Read(DeviceIndex: LongWord; pLowLatches, pHighLatches, pNoLatches: PByte; pLatchesBytes: PLongWord): LongWord; cdecl;
var
  L, oL: LongWord;
  LatchBuf: array of Byte;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    Result := CheckFirmware20(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    oL := Min(pLatchesBytes^, Dev[DeviceIndex].DIOBytes);
    SetLength(LatchBuf, Dev[DeviceIndex].DIOBytes * 3);
    try
      Move(pLowLatches^, LatchBuf[0], oL);
      Move(pHighLatches^, LatchBuf[0], oL);
      Move(pNoLatches^, LatchBuf[0], oL);
    except
      Result := ERROR_NOACCESS;
      Exit;
    end;

    L := Length(LatchBuf);
    Result := GenericVendorRead(DeviceIndex, AUR_DIO_LATCH_READ, $0000, $0000, L, @LatchBuf[0]);
    if Result <> ERROR_SUCCESS then Exit;
    if Integer(L) < Length(LatchBuf) then begin
      Result := ERROR_INVALID_DATA;
      Exit;
    end;

    L := Dev[DeviceIndex].DIOBytes;
    try
      Move(LatchBuf[  L], pLowLatches^, oL);
      Move(LatchBuf[2*L], pHighLatches^, oL);
      Move(LatchBuf[0  ], pNoLatches^, oL);
    except
      Result := ERROR_NOACCESS;
      Exit;
    end;
    pLatchesBytes^ := L;

  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;


//DIO debounce.
function DIO_Deb_SetConfig(DeviceIndex: LongWord; pDebTable: Pointer; pDebTableBytes: PLongWord): LongWord; cdecl;
var
  DebBuf: array of Byte;
  L: LongWord;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    Result := CheckFirmware20(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    L := pDebTableBytes^;
    if L > 1024 then begin
      Result := ERROR_BAD_LENGTH;
      Exit;
    end;

    SetLength(DebBuf, L);
    try
      Move(pDebTable^, DebBuf[0], L);
    except
      Result := ERROR_NOACCESS;
      Exit;
    end;

    Result := GenericVendorWrite(DeviceIndex, AUR_GEN_REGISTER, $0000, $0051, L, @DebBuf[0]);
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function DIO_Deb_ReadAll(DeviceIndex: LongWord; pDebData: PByte): LongWord; cdecl;
var
  DebBuf: array of Byte;
  L: LongWord;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    Result := CheckFirmware20(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    L := Dev[DeviceIndex].DIOBytes;
    SetLength(DebBuf, L);
    try
      Move(pDebData^, DebBuf[0], L);
    except
      Result := ERROR_NOACCESS;
      Exit;
    end;

    Result := GenericVendorRead(DeviceIndex, AUR_DIO_DEB_READ, $0000, $0000, L, @DebBuf[0]);
    if Result <> ERROR_SUCCESS then Exit;

    try
      Move(DebBuf[0], pDebData^, L);
    except
      Result := ERROR_NOACCESS;
      Exit;
    end;

  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;



end.
