unit FastADC;

interface
uses
  Windows,
  theseUtils,
  ContBufThreads;

type
  TADCWorker = class(TStartableThread)
  public
    DI: LongWord;
    TargetPipe: LongWord;
    BytesLeft: LongInt;
    pTar: PWord;
    bAbort: Boolean;
  protected
    ADBuf: array of Byte;

    procedure Execute; override;
  end;

function ADC_BulkAcquire(DeviceIndex: LongWord; BufSize: LongWord; pBuf: Pointer): LongWord; cdecl;
function ADC_BulkPoll(DeviceIndex: LongWord; var BytesLeft: LongWord): LongWord; cdecl;
function ADC_BulkAbort(DeviceIndex: LongWord): LongWord; cdecl;
function ADC_BulkMode(DeviceIndex, BulkMode: LongWord): LongWord; cdecl;

implementation

uses
  Math,
{$IFDEF VER130}  // Delphi 5
  SysUtils, Classes,
{$ELSE}         // anything other than Delphi 5, presumably XE2 or XE3
  System.SysUtils,
  System.Classes,
{$ENDIF}
  DeviceLevel,
  u8254,
  ADC,
  ADC_Continuous
  ;

function ADC_BulkAcquire(DeviceIndex: LongWord; BufSize: LongWord; pBuf: Pointer): LongWord; cdecl;
var
  oADCWorker: TADCWorker;
  oADCContAcqWorker: TADCContAcqWorker;
  oADCContBufWorker: TADCContBufWorker;
  BCData: LongWord;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if not bADCBulk then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    {
    if not bADCOpen then begin
      Result := ERROR_FILE_NOT_FOUND;
      Exit;
    end;
    }

    oADCWorker := TADCWorker(ADCWorker);
    oADCContAcqWorker := TADCContAcqWorker(ADCContAcqWorker);
    oADCContBufWorker := TADCContBufWorker(ADCContBufWorker);
    if ((oADCWorker <> nil) and (oADCWorker.BytesLeft <> 0))
    or (oADCContAcqWorker <> nil) or (oADCContBufWorker <> nil) then begin
      Result := ERROR_TOO_MANY_OPEN_FILES;
      Exit;
    end;

    //GenericVendorWrite(DeviceIndex, AUR_SET_GPIF_MODE, 0, $0, 0, nil);
    BCData := $00000005;
    GenericVendorWrite(DeviceIndex, $BC, BufSize shr 17, BufSize shr 1, SizeOf(BCData), @BCData);

    ADCWorker := TADCWorker.Create(True);
    with TADCWorker(ADCWorker) do begin
      FreeOnTerminate := False;
      DI := DeviceIndex;
      TargetPipe := $86;
      BytesLeft := BufSize;
      pTar := pBuf;
      bAbort := False;
      Start;

    end;

    if oADCWorker <> nil then oADCWorker.Free;
  end;
end;


function ADC_BulkPoll(DeviceIndex: LongWord; var BytesLeft: LongWord): LongWord; cdecl;
begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if not bADCBulk then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if ADCWorker = nil then begin
      Result := ERROR_FILE_NOT_FOUND;
      Exit;
    end;

    if (WaitForSingleObject(ADCWorker.Handle, 0) = WAIT_OBJECT_0) and (ADCWorker.ReturnValue <> ERROR_SUCCESS) then begin
      Result := ADCWorker.ReturnValue;
      Exit;
    end;

    BytesLeft := TADCWorker(ADCWorker).BytesLeft;
  end;
end;

function ADC_BulkAbort(DeviceIndex: LongWord): LongWord; cdecl;
begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if not bADCBulk then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if ADCWorker = nil then begin
      Result := ERROR_FILE_NOT_FOUND;
      Exit;
    end;

    if WaitForSingleObject(ADCWorker.Handle, 0) = WAIT_OBJECT_0 then
      Result := ADCWorker.ReturnValue
    else begin
      TADCWorker(ADCWorker).bAbort := True;
      Result := ERROR_SUCCESS;
    end;
  end;
end;

 procedure TADCWorker.Execute;
var
  L, Status: LongWord;

begin
  SetThreadPriority(Handle, THREAD_PRIORITY_TIME_CRITICAL);

  SetLength(ADBuf, Min(BytesLeft, Dev[DI].StreamingBlockSize));
  repeat
    if bAbort then begin
      ReturnValue := ERROR_OPERATION_ABORTED;
      BytesLeft := 0;
      Exit;
    end;

    L := 0;
    Status := AWU_GenericBulkIn(DI, TargetPipe, @ADBuf[0], Length(ADBuf), L);

    if Status <> ERROR_SUCCESS then begin
      ReturnValue := Status;
      BytesLeft := 0;
      Exit;
    end;

    Move(ADBuf[0], pTar^, L);
    Inc(pTar, L div 2);
    Dec(BytesLeft, L);

    if BytesLeft < Length(ADBuf) then SetLength(ADBuf, BytesLeft);
  until BytesLeft = 0;
  ReturnValue := ERROR_SUCCESS;
end;


function ADC_BulkMode(DeviceIndex, BulkMode: LongWord): LongWord; cdecl;
begin
  Result := ERROR_NOT_SUPPORTED;
end;
end.

