unit ADC_Clocked;

interface
  function ADC_BulkContinuousCallbackStartClocked(DeviceIndex: LongWord; BufSize: LongWord; BaseBufCount: LongWord; Context: LongWord; pCallback: Pointer; var Hz: Double): LongWord; cdecl;

implementation

uses
  Windows,
  DeviceLevel,
  ContBufThreads,
  FastADC,
  ADC_Continuous,
  u8254;

function ADC_BulkContinuousCallbackStartClocked(DeviceIndex: LongWord; BufSize: LongWord; BaseBufCount: LongWord; Context: LongWord; pCallback: Pointer; var Hz: Double): LongWord; cdecl;
var
  oADCWorker: TADCWorker;
  oADCContAcqWorker: TADCContAcqWorker;
  oADCContBufWorker: TADCContBufWorker;
  I: Integer;
  hWaits: array[0..1] of THandle;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[DeviceIndex] do begin
      if not bADCBulk then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      if (BufSize and $1FF) <> 0 then begin
        Result := ERROR_INVALID_USER_BUFFER;
        Exit;
      end;

      oADCWorker := TADCWorker(ADCWorker);
      oADCContAcqWorker := TADCContAcqWorker(ADCContAcqWorker);
      oADCContBufWorker := TADCContBufWorker(ADCContBufWorker);
      if ((oADCWorker <> nil) and (oADCWorker.BytesLeft <> 0))
      or (oADCContAcqWorker <> nil) or (oADCContBufWorker <> nil) then begin
        Result := ERROR_TOO_MANY_OPEN_FILES;
        Exit;
      end;

      CTR_8254Mode(DeviceIndex, 0, 1, 2);
      CTR_8254Mode(DeviceIndex, 0, 2, 3);

      ADCContAcqWorker := TADCContAcqWorker(TADCContAcqWorker).Create(True);
      ADCContBufWorker := TADCContBufWorker(TADCContBufWorker).Create(True);
      with TADCContBufWorker(ADCContBufWorker) do begin
        FreeOnTerminate := False;
        MyAcqThread := TADCContAcqWorker(ADCContAcqWorker);
        BytesPerBuf := BufSize;
        Callback := pCallback;
        CallbackContext := Context;
        hBufMutex := CreateMutex(nil, False, nil);
        hKillSem := CreateSemaphore(nil, 0, $FFFFFF, nil);
        hBlankBufSem := CreateSemaphore(nil, BaseBufCount, $FFFFFF, nil);
        hDataBufSem := CreateSemaphore(nil, 0, $FFFFFF, nil);
        SetLength(TADCContBufWorker(ADCContBufWorker).BufBuf, BaseBufCount);
        for I := 0 to BaseBufCount-1 do begin
          New(TADCContBufWorker(ADCContBufWorker).BufBuf[I]);
          SetLength(TADCContBufWorker(ADCContBufWorker).BufBuf[I].ADBuf, BufSize);
          TADCContBufWorker(ADCContBufWorker).BufBuf[I].BufState := bsBlank;
        end;
        Start;
      end;
      with TADCContAcqWorker(ADCContAcqWorker) do begin
        FreeOnTerminate := False;
        MyBufThread := TADCContBufWorker(ADCContBufWorker);
        hStartEvent := CreateEvent(nil, True, False, nil);
        DI := DeviceIndex;
        TargetPipe := $86;
        //BytesPerBC := BufSize * BaseBufCount;
        bCounterControl := True;//Hz > 1/2;
        if bCounterControl then CalcHzDivisors(@Hz, @DivisorA, @DivisorB, RootClock);
        Start;
      end;
      hWaits[0] := TADCContAcqWorker(ADCContAcqWorker).Handle;
      hWaits[1] := TADCContAcqWorker(ADCContAcqWorker).hStartEvent;
      WaitForMultipleObjects(2, @hWaits[0], False, INFINITE);
      CloseHandle(TADCContAcqWorker(ADCContAcqWorker).hStartEvent);

      if oADCWorker <> nil then oADCWorker.Free;
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;//ADC_BulkContinuousCallbackStartClocked

end.

