unit ADC_Continuous;

interface
uses
{$IFDEF VER130}  // Delphi 5
  SysUtils, Classes,
{$ELSE}         // anything other than Delphi 5, presumably XE2 or XE3
  System.SysUtils,
  System.Classes,
{$ENDIF}
  Windows,
  ContBufThreads;
type
  TADCContAcqWorker = class;
  PContCallback = procedure (pBuf: PWord; BufSize, Flags, Context: LongWord); cdecl;

  TADCContBufWorker = class(TStartableThread)
  public
    BytesPerBuf: LongWord;
    hBufMutex, hBlankBufSem, hDataBufSem, hKillSem: THandle;
    MyAcqThread: TADCContAcqWorker;
    BufBuf: array of PContBuf;
    NextIndexIn, NextIndexOut: Integer;
    Callback: PContCallback;
    CallbackContext: LongWord;

    procedure GetDebugStats(var Blanks, Accessings, GotDatas: LongWord);
  protected
    procedure Execute; override;
    function GetBlankBuf(Flags: LongWord): PContBuf;
    procedure PutDataBuf(pNewBuf: PContBuf);
    function GetDataBufOrKilled: PContBuf;
    procedure PutBlankBuf(pNewBuf: PContBuf);
    function ExtraBuf: PContBuf;
  end;

  TADCContAcqWorker = class(TStartableThread)
  public
    DI: LongWord;
    TargetPipe: LongWord;
    bCounterControl: Boolean;
    DivisorA, DivisorB: Word;
    //BytesPerBC,
    IOStatus: LongWord;
    MyBufThread: TADCContBufWorker;
    hStartEvent: THandle;
  protected
    procedure Execute; override;
  end;

function ADC_BulkContinuousCallbackStart(DeviceIndex: LongWord; BufSize: LongWord; BaseBufCount: LongWord; Context: LongWord; pCallback: Pointer): LongWord; cdecl;
function ADC_BulkContinuousRingStart(DeviceIndex: LongWord; RingBufferSize: LongWord; PacketsPerBlock: LongWord): LongWord; cdecl;
function ADC_BulkContinuousEnd(DeviceIndex: LongWord; pIOStatus: PLongWord): LongWord; cdecl;
function ADC_BulkContinuousDebug(DeviceIndex: LongWord; var Blanks, Accessings, GotDatas: LongWord): LongWord; cdecl;
function ADC_ReadData(DeviceIndex: LongWord; pConfigBuf: Pointer; var ScansToRead: LongWord; pData: PDouble; Timeout: Double): LongWord; cdecl;
function ADC_FullStartRing(DeviceIndex: LongWord; pConfigBuf: Pointer; var ConfigBufSize: LongWord; CalFileName: PAnsiChar; pCounterHz: PDouble; RingBufferSize: LongWord; PacketsPerBlock: LongWord): LongWord; cdecl;

implementation

uses
  DeviceLevel,
  FastADC,
  u8254,
  Math,
  AIOUSB_Types,
  TheseUtils,
  ADC
  ;

function ADC_ReadData(DeviceIndex: LongWord; pConfigBuf: Pointer; var ScansToRead: LongWord; pData: PDouble; Timeout: Double): LongWord; cdecl;
var
  oADCContAcqWorker: TADCContAcqWorker;
  oADCContBufWorker: TADCContBufWorker;
  oRingAcqData: TPRingAcqData;
  mTimeoutEnd: LongWord;
  StartCh, StopCh,
  Oversample, ScanChannels, ScanSamples, SamplesToRead, ScansInRing, ScansRead,
  Transfer: LongWord;
  ADBuf: array of Word;

  function ConfigByte(Offset: Integer): Byte;
  var
    pConfigByte: PByte;

  begin
    pConfigByte := pConfigBuf;
    Inc(pConfigByte, Offset);
    Result := pConfigByte^;
  end;

  function RingSamplesLeft(pRing: TPRingAcqData): LongWord;
  begin
    if pRing = nil then
      Result := 0
    else begin
      Result := pRing.iNextWrite - pRing.iNextRead;
      if LongInt(Result) < 0 then Inc(Result, Length(pRing.RingBuffer));
    end;
  end;

  procedure ReadRingSamples(ToIndex, Count: LongWord);
  var
    nNextRead, RingBufLen: LongWord;

  begin
    with Dev[DeviceIndex].pRingAcqData^ do begin
      //This is the write authority for iNextRead.
      RingBufLen := Length(RingBuffer);
      nNextRead := iNextRead + Count;
      if nNextRead < RingBufLen then begin
        Move(RingBuffer[iNextRead], ADBuf[ToIndex], 2*Count);
      end
      else if nNextRead = RingBufLen then begin
        nNextRead := 0;
        Move(RingBuffer[iNextRead], ADBuf[ToIndex], 2*Count);
      end
      else begin
        Dec(nNextRead, RingBufLen);
        Transfer := RingBufLen - iNextRead;
        Move(RingBuffer[iNextRead], ADBuf[ToIndex], 2*Transfer);
        Move(RingBuffer[0], ADBuf[ToIndex + Transfer], 2*nNextRead);
      end;
      iNextRead := nNextRead;
    end;
  end;

  function WaitForData: Boolean;
  var
    mTimeoutLeft: LongWord;

  begin
    if Timeout = Infinity then
      mTimeoutLeft := INFINITE
    else begin
      mTimeoutLeft := mTimeoutEnd - GetTickCount;
      if LongInt(mTimeoutLeft) < 0 then begin
        Result := False;
        Exit;
      end;
    end;
    Result := WaitForSingleObject(Dev[DeviceIndex].pRingAcqData.hSem, mTimeoutLeft) <> ERROR_TIMEOUT;
  end;

  procedure PrepareData;
  var
    ThisCh: LongWord;
    iSample, iOversample, iADBuf, Ct: Integer;
    OvScale: Double;
    Range: Byte;
    pIData: PDouble;

  begin
    pIData := pData;
    ThisCh := StartCh;
    iADBuf := 0;
    OvScale := 1/65536 / Oversample;
    for iSample := 0 to ScansToRead*ScanChannels - 1 do begin
      Range := ConfigByte(ThisCh shr Dev[DeviceIndex].RangeShift);
      {
      V := 0;
      for iOversample := 0 to Oversample-1 do
        V := V + ADBuf[iADBuf + iOversample]
      ;
      Inc(iADBuf, Oversample);
      V := 1/65536 * V / Oversample;

      if (Range and 1) <> 0 then V := V * 2 - 1;
      if (Range and 2) = 0 then V := V * 2;
      if (Range and 4) = 0 then V := V * 5;
      }
      Ct := 0;
      for iOversample := 0 to Oversample-1 do
        Ct := Ct + ADBuf[iADBuf + iOversample]
      ;
      Inc(iADBuf, Oversample);

      if (Range and 1) <> 0 then Ct := Ct * 2 - ($10000 * LongInt(Oversample));
      if (Range and 2) = 0 then Ct := Ct * 2;
      if (Range and 4) = 0 then Ct := Ct * 5;
      pIData^ := OvScale * Ct;
      Inc(pIData);

      if ThisCh = StopCh then
        ThisCh := StartCh
      else
        ThisCh := (ThisCh + 1) mod Dev[DeviceIndex].ADCMUXChannels
      ;
    end;
  end;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[DeviceIndex] do begin
      if not bADCBulk then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      //Currently only supports cont ring mode. In the future, should support bulk block mode, and hypothetically whatever the hell it would mean to support cont callback mode.
      oADCContAcqWorker := TADCContAcqWorker(ADCContAcqWorker);
      oADCContBufWorker := TADCContBufWorker(ADCContBufWorker);
      oRingAcqData := pRingAcqData;
      if (oADCContAcqWorker = nil) or (oADCContBufWorker = nil) or (oRingAcqData = nil) then begin
        Result := ERROR_NOT_SUPPORTED;
        Exit;
      end;

      if ScansToRead = 0 then Exit; //Requested nop, fast return success.

      StartCh := (ConfigByte($12) and $F) or ((ConfigByte($14) shl 4) and $F0);
      StopCh := ((ConfigByte($12) shr 4) and $F) or (ConfigByte($14) and $F0);
      if StopCh < StartCh then
        ScanChannels := ADCMUXChannels + StartCh - StopCh + 1
      else
        ScanChannels := StopCh - StartCh + 1
      ;
      Oversample := 1 + ConfigByte($13);
      ScanSamples := ScanChannels * Oversample;
      SamplesToRead := ScanSamples * ScansToRead;
      SetLength(ADBuf, SamplesToRead);

      if Timeout = 0 then begin //They want whatever we already have.
        ScansRead := 0;
        ScansInRing := RingSamplesLeft(pRingAcqData) div ScanSamples;
        if ScansInRing > 0 then begin
          if ScansInRing > ScansToRead then
            ScansInRing := ScansToRead
          ;
          ReadRingSamples(0, ScanSamples * ScansInRing);
          ScansToRead := ScansInRing;
        end
        else begin
          ScansToRead := 0;
          Result := ERROR_TIMEOUT;
        end;
        if ScansRead > 0 then
          PrepareData
        else
          Result := ERROR_TIMEOUT
        ;
      end
      else if Timeout < 0 then begin //They want all or nothing.
        Timeout := -Timeout;
        if Timeout <> Infinity then
          mTimeoutEnd := GetTickCount + Round(1000 * Timeout)
        ;
        while RingSamplesLeft(pRingAcqData) < SamplesToRead do
          if not WaitForData then begin
            ScansToRead := 0;
            Result := ERROR_TIMEOUT;
            Exit;
          end
        ;

        ReadRingSamples(0, SamplesToRead);
        PrepareData;
      end
      else begin //They want everything we can give them.
        ScansRead := 0;
        if Timeout <> Infinity then
          mTimeoutEnd := GetTickCount + Round(1000 * Timeout)
        ;
        repeat
          ScansInRing := RingSamplesLeft(pRingAcqData) div ScanSamples;
          if ScansInRing > 0 then begin
            if ScansInRing >= ScansToRead - ScansRead then
              ScansInRing := ScansToRead - ScansRead
            ;
            ReadRingSamples(ScanSamples * ScansRead, ScanSamples * ScansInRing);
            Inc(ScansRead, ScansInRing);
            if ScansRead >= ScansToRead then
              Break
            ;
          end;

        until not WaitForData;
        ScansToRead := ScansRead;
        if ScansRead > 0 then
          PrepareData
        else
          Result := ERROR_TIMEOUT
        ;
      end;
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function ADC_FullStartRing(DeviceIndex: LongWord; pConfigBuf: Pointer;
                           var ConfigBufSize: LongWord; CalFileName: PAnsiChar; pCounterHz: PDouble;
                           RingBufferSize: LongWord; PacketsPerBlock: LongWord
                          ): LongWord; cdecl;
var
  bCounterTriggered: Boolean;
  Content: AnsiString;
  Hz: Double;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[DeviceIndex] do begin
      if ConfigBytes = 0 then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      if ConfigBufSize < ConfigBytes then begin
        Result := ERROR_INSUFFICIENT_BUFFER;
        ConfigBufSize := ConfigBytes;
        Exit;
      end;

      if pConfigBuf = nil then begin
        Result := ERROR_INVALID_PARAMETER;
        Exit;
      end;

      SetString(Content, PAnsiChar(pConfigBuf), ConfigBytes); //Probe buffer

      bCounterTriggered := (PByteArray(pConfigBuf)[$11] and $01) <> 0;

      if bCounterTriggered then begin
        Hz := 0;
        Result := CTR_StartOutputFreq(DeviceIndex, 0, @Hz);
        if Result <> ERROR_SUCCESS then Exit;
      end;

      Result := ADC_SetConfig(DeviceIndex, pConfigBuf, ConfigBufSize);
      if Result <> ERROR_SUCCESS then Exit;

      if CalFileName <> nil then begin
        Result := ADC_SetCal(DeviceIndex, CalFileName);
        if (Result <> ERROR_SUCCESS) and (Result <> ERROR_NOT_SUPPORTED) then Exit;
      end;

      Result := ADC_BulkContinuousRingStart(DeviceIndex, RingBufferSize, PacketsPerBlock);
      if Result <> ERROR_SUCCESS then Exit;

      if bCounterTriggered then
        CTR_StartOutputFreq(DeviceIndex, 0, pCounterHz)
      ;
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function ADC_BulkContinuousCallbackStart(DeviceIndex: LongWord; BufSize: LongWord; BaseBufCount: LongWord; Context: LongWord; pCallback: Pointer): LongWord; cdecl;
var
  oADCWorker: TADCWorker;
  oADCContAcqWorker: TADCContAcqWorker;
  oADCContBufWorker: TADCContBufWorker;
  I: Integer;
  hWaits: array[0..1] of THandle;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[DeviceIndex] do begin
      if not bADCBulk then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      if (BufSize and $1FF) <> 0 then begin
        Result := ERROR_INVALID_USER_BUFFER;
        Exit;
      end;

      oADCWorker := TADCWorker(ADCWorker);
      oADCContAcqWorker := TADCContAcqWorker(ADCContAcqWorker);
      oADCContBufWorker := TADCContBufWorker(ADCContBufWorker);
      if ((oADCWorker <> nil) and (oADCWorker.BytesLeft <> 0))
      or (oADCContAcqWorker <> nil) or (oADCContBufWorker <> nil) then begin
        Result := ERROR_TOO_MANY_OPEN_FILES;
        Exit;
      end;

      ADCContAcqWorker := TADCContAcqWorker.Create(True);
      ADCContBufWorker := TADCContBufWorker.Create(True);
      with TADCContBufWorker(ADCContBufWorker) do begin
        FreeOnTerminate := False;
        MyAcqThread := TADCContAcqWorker(ADCContAcqWorker);
        BytesPerBuf := BufSize;
        Callback := pCallback;
        CallbackContext := Context;
        hBufMutex := CreateMutex(nil, False, nil);
        hKillSem := CreateSemaphore(nil, 0, $FFFFFF, nil);
        hBlankBufSem := CreateSemaphore(nil, BaseBufCount, $FFFFFF, nil);
        hDataBufSem := CreateSemaphore(nil, 0, $FFFFFF, nil);
        SetLength(TADCContBufWorker(ADCContBufWorker).BufBuf, BaseBufCount);
        for I := 0 to BaseBufCount-1 do begin
          New(TADCContBufWorker(ADCContBufWorker).BufBuf[I]);
          SetLength(TADCContBufWorker(ADCContBufWorker).BufBuf[I].ADBuf, BufSize);
          TADCContBufWorker(ADCContBufWorker).BufBuf[I].BufState := bsBlank;
        end;
        Start;
      end;
      with TADCContAcqWorker(ADCContAcqWorker) do begin
        FreeOnTerminate := False;
        MyBufThread := TADCContBufWorker(ADCContBufWorker);
        hStartEvent := CreateEvent(nil, True, False, nil);
        DI := DeviceIndex;
        TargetPipe := $86;
        //BytesPerBC := BufSize * BaseBufCount;
        bCounterControl := False;
        Start;
      end;
      hWaits[0] := TADCContAcqWorker(ADCContAcqWorker).Handle;
      hWaits[1] := TADCContAcqWorker(ADCContAcqWorker).hStartEvent;
      WaitForMultipleObjects(2, @hWaits[0], False, INFINITE);
      CloseHandle(TADCContAcqWorker(ADCContAcqWorker).hStartEvent);

      if oADCWorker <> nil then oADCWorker.Free;
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;


procedure BGRingContCallback(pBuf: PWord; BufSize, Flags, Context: LongWord); cdecl;
var
  SamplesUntilWrap, L: LongWord;

begin
  BufSize := BufSize shr 1; //BufSize is henceforth in samples.
  if BufSize = 0 then Exit;
  if Context >= 32 then Exit;
  if Dev[Context].pRingAcqData = nil then Exit;

  with Dev[Context].pRingAcqData^ do begin
    L := Length(RingBuffer);
    SamplesUntilWrap := L - iNextWrite;

    //This is the write authority for iNextWrite.
    if BufSize < SamplesUntilWrap then begin //Transfer in the beginning or middle.
      Move(pBuf^, RingBuffer[iNextWrite], 2*BufSize);
      Inc(iNextWrite, BufSize);
    end
    else if BufSize = SamplesUntilWrap then begin //Transfer right at the end.
      Move(pBuf^, RingBuffer[iNextWrite], 2*BufSize);
      iNextWrite := 0;
    end
    else begin //Transfer crossing the end, must wrap to beginning as well.
      Move(pBuf^, RingBuffer[iNextWrite], 2*SamplesUntilWrap);
      Dec(BufSize, SamplesUntilWrap);
      Inc(pBuf, SamplesUntilWrap);
      //This assumes the remainder isn't bigger than the entire ring buffer, since ADC_BulkContinuousRingStart already made sure of that.
      Move(pBuf^, RingBuffer[0], 2*BufSize);
      iNextWrite := BufSize;
    end;

    ReleaseSemaphore(hSem, 1, nil);
  end;
end;

function ADC_BulkContinuousRingStart(DeviceIndex: LongWord; RingBufferSize: LongWord; PacketsPerBlock: LongWord): LongWord; cdecl;
var
  oADCWorker: TADCWorker;
  oADCContAcqWorker: TADCContAcqWorker;
  oADCContBufWorker: TADCContBufWorker;
  I: Integer;
  hWaits: array[0..1] of THandle;
  BufSize, BaseBufCount: LongWord;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[DeviceIndex] do begin
      if not bADCBulk then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      RingBufferSize := RingBufferSize and not $1FF;
      if RingBufferSize < 4*1024 then begin
        Result := ERROR_INVALID_USER_BUFFER;
        Exit;
      end;

      if PacketsPerBlock > (RingBufferSize shr 9) then begin
        Result := ERROR_INVALID_USER_BUFFER;
        Exit;
      end;
      if PacketsPerBlock <> 0 then
        BufSize := $200 * PacketsPerBlock
      else
        BufSize := Ceil(RingBufferSize / (8*512)) * 512
      ;
      BaseBufCount := 3;

      oADCWorker := TADCWorker(ADCWorker);
      oADCContAcqWorker := TADCContAcqWorker(ADCContAcqWorker);
      oADCContBufWorker := TADCContBufWorker(ADCContBufWorker);
      if ((oADCWorker <> nil) and (oADCWorker.BytesLeft <> 0))
      or (oADCContAcqWorker <> nil) or (oADCContBufWorker <> nil) then begin
        Result := ERROR_TOO_MANY_OPEN_FILES;
        Exit;
      end;

      if pRingAcqData = nil then begin
        New(pRingAcqData);
        pRingAcqData.hSem := CreateSemaphore(nil, 0, 1, nil);
      end;
      ADCContAcqWorker := TADCContAcqWorker.Create(True);
      ADCContBufWorker := TADCContBufWorker.Create(True);
      with pRingAcqData^ do begin
        SetLength(RingBuffer, RingBufferSize shr 1);
        iNextRead := 0;
        iNextWrite := 0;
      end;
      with TADCContBufWorker(ADCContBufWorker) do begin
        FreeOnTerminate := False;
        MyAcqThread := TADCContAcqWorker(ADCContAcqWorker);
        BytesPerBuf := BufSize;
        Callback := BGRingContCallback;
        CallbackContext := DeviceIndex;
        hBufMutex := CreateMutex(nil, False, nil);
        hKillSem := CreateSemaphore(nil, 0, $FFFFFF, nil);
        hBlankBufSem := CreateSemaphore(nil, BaseBufCount, $FFFFFF, nil);
        hDataBufSem := CreateSemaphore(nil, 0, $FFFFFF, nil);
        SetLength(TADCContBufWorker(ADCContBufWorker).BufBuf, BaseBufCount);
        for I := 0 to BaseBufCount-1 do begin
          New(TADCContBufWorker(ADCContBufWorker).BufBuf[I]);
          SetLength(TADCContBufWorker(ADCContBufWorker).BufBuf[I].ADBuf, BufSize);
          TADCContBufWorker(ADCContBufWorker).BufBuf[I].BufState := bsBlank;
        end;
        Start;
      end;
      with TADCContAcqWorker(ADCContAcqWorker) do begin
        FreeOnTerminate := False;
        MyBufThread := TADCContBufWorker(ADCContBufWorker);
        hStartEvent := CreateEvent(nil, True, False, nil);
        DI := DeviceIndex;
        TargetPipe := $86;
        bCounterControl := False;
        Start;
      end;
      hWaits[0] := TADCContAcqWorker(ADCContAcqWorker).Handle;
      hWaits[1] := TADCContAcqWorker(ADCContAcqWorker).hStartEvent;
      WaitForMultipleObjects(2, @hWaits[0], False, INFINITE);
      CloseHandle(TADCContAcqWorker(ADCContAcqWorker).hStartEvent);

      if oADCWorker <> nil then oADCWorker.Free;
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function ADC_BulkContinuousEnd(DeviceIndex: LongWord; pIOStatus: PLongWord): LongWord; cdecl;
var
  oADCContAcqWorker: TADCContAcqWorker;
  oADCContBufWorker: TADCContBufWorker;
  I: Integer;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[DeviceIndex] do begin
      if not bADCBulk then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      oADCContAcqWorker := TADCContAcqWorker(ADCContAcqWorker);
      oADCContBufWorker := TADCContBufWorker(ADCContBufWorker);
      ADCContAcqWorker := nil;
      ADCContBufWorker := nil;
      if (oADCContAcqWorker = nil) or (oADCContBufWorker = nil) then begin
        Result := ERROR_HANDLE_EOF;
        Exit;
      end;

      if pIOStatus <> nil then pIOStatus^ := oADCContAcqWorker.IOStatus;

      oADCContAcqWorker.Terminate;
      ReleaseSemaphore(oADCContBufWorker.hKillSem, 1, nil);
      ReleaseSemaphore(oADCContBufWorker.hDataBufSem, 1, nil);

      oADCContAcqWorker.WaitFor;
      oADCContBufWorker.WaitFor;
      CloseHandle(oADCContBufWorker.hBufMutex);
      CloseHandle(oADCContBufWorker.hKillSem);
      CloseHandle(oADCContBufWorker.hBlankBufSem);
      CloseHandle(oADCContBufWorker.hDataBufSem);
      for I := 0 to High(oADCContBufWorker.BufBuf) do begin
        Dispose(oADCContBufWorker.BufBuf[I]);
      end;
      SetLength(oADCContBufWorker.BufBuf, 0);

      oADCContAcqWorker.Free;
      oADCContBufWorker.Free;

      if pRingAcqData <> nil then begin
        CloseHandle(pRingAcqData.hSem);
        Dispose(pRingAcqData);
        pRingAcqData := nil;
      end;
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function ADC_BulkContinuousDebug(DeviceIndex: LongWord; var Blanks, Accessings, GotDatas: LongWord): LongWord; cdecl;
var
  oADCContAcqWorker: TADCContAcqWorker;
  oADCContBufWorker: TADCContBufWorker;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[DeviceIndex] do begin
      if not bADCBulk then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      oADCContAcqWorker := TADCContAcqWorker(ADCContAcqWorker);
      oADCContBufWorker := TADCContBufWorker(ADCContBufWorker);
      if (oADCContAcqWorker = nil) or (oADCContBufWorker = nil) then begin
        Result := ERROR_HANDLE_EOF;
        Exit;
      end;

      oADCContBufWorker.GetDebugStats(Blanks, Accessings, GotDatas);
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

procedure TADCContAcqWorker.Execute;
  function DoBCControl(TCSize: LongWord; ControlData: LongWord): LongWord;
  begin
    Result := _GenericVendorWrite(DI, $BC, TCSize shr 16, TCSize, SizeOf(ControlData), @ControlData);
  end;

  procedure DoErrOut(Status: LongWord);
  begin
    IOStatus := Status;
    ReturnValue := Status;
    ReleaseSemaphore(MyBufThread.hKillSem, 1, nil);
    ReleaseSemaphore(MyBufThread.hDataBufSem, 1, nil);
  end;

  var
    BytesLeft, DataSize,
    Status: LongWord;
    ThisBuf: PContBuf;
begin
  SetThreadPriority(Handle, THREAD_PRIORITY_TIME_CRITICAL);

  IOStatus := ERROR_SUCCESS;
  //GenericVendorWrite(DI, AUR_SET_GPIF_MODE, 0, $1, 0, nil);
  if bCounterControl then
    _GenericVendorWrite(DI, $C5, DivisorA, DivisorB, 0, nil) // debug mode for CNSPoway Boian Mitov not supported in latest firmwares... --J2H
  else
    DoBCControl(0, $01000007)
  ;
  SetEvent(hStartEvent);

  while not Terminated do begin
    ThisBuf := MyBufThread.GetBlankBuf(0);

    ThisBuf.UsedSize := 0;
    Status := _GenericBulkIn(DI, TargetPipe, @ThisBuf.ADBuf[0], Length(ThisBuf.ADBuf), ThisBuf.UsedSize);

    if Status <> ERROR_SUCCESS then begin
      DoErrOut(Status);
      DoBCControl(0, $00020002);
      //D4StartStop($2);
      Exit;
    end;

    if ThisBuf.UsedSize <> 0 then
      MyBufThread.PutDataBuf(ThisBuf)
    else
      MyBufThread.PutBlankBuf(ThisBuf)
    ;
  end;

  DoBCControl(0, $00020002);
  //D4StartStop($2);

  //`@ This whole drain loop was commented out in DebugTemp. Wut?
  repeat
    BytesLeft := 0;
    DataSize := SizeOf(BytesLeft);
    Status := _GenericVendorRead(DI, $BC, 0, 0, DataSize, @BytesLeft);
    if Status <> ERROR_SUCCESS then begin
      DoErrOut(Status);
      Exit;
    end;
    BytesLeft := BytesLeft and $0000FFFF;
    if BytesLeft = 0 then Break;

    ThisBuf := MyBufThread.GetBlankBuf(0);

    ThisBuf.UsedSize := 0;
    DataSize := Length(ThisBuf.ADBuf);
    if DataSize < BytesLeft then BytesLeft := DataSize;
    Status := _GenericBulkIn(DI, TargetPipe, @ThisBuf.ADBuf[0], BytesLeft, ThisBuf.UsedSize);

    if Status <> ERROR_SUCCESS then begin
      MyBufThread.PutBlankBuf(ThisBuf);
      DoErrOut(Status);
      Exit;
    end;

    if ThisBuf.UsedSize <> 0 then
      MyBufThread.PutDataBuf(ThisBuf)
    else
      MyBufThread.PutBlankBuf(ThisBuf)
    ;
  until True;//`@ False;
  ReturnValue := ERROR_SUCCESS;
end;

{ TADCContBufWorker }

procedure TADCContBufWorker.Execute;
var
  ThisBuf: PContBuf;

begin
  repeat
    if Terminated then begin
      ReturnValue := ERROR_OPERATION_ABORTED;
      Exit;
    end;

    ThisBuf := GetDataBufOrKilled;
    if ThisBuf = nil then begin
      try
        Callback(nil, 0, CallbackFlag_EndStream, CallbackContext);
      except
      end;
      ReturnValue := ERROR_OPERATION_ABORTED;
      Exit;
    end;

    if (ThisBuf.UsedSize <> 0) or (ThisBuf.Flags <> 0) then try
      Callback(@ThisBuf.ADBuf[0], ThisBuf.UsedSize, ThisBuf.Flags, CallbackContext);
    except
    end;

    PutBlankBuf(ThisBuf);
  until False;
  ReturnValue := ERROR_SUCCESS;
end;

function TADCContBufWorker.ExtraBuf: PContBuf;
var
  I: Integer;

begin
  WaitForSingleObject(hBufMutex, INFINITE);
    I := Length(BufBuf);
    SetLength(BufBuf, I + 1);
    New(BufBuf[I]);
    Result := BufBuf[I];
    SetLength(Result.ADBuf, BytesPerBuf);
    Result.BufState := bsAccessing;
  ReleaseMutex(hBufMutex);
end;

function TADCContBufWorker.GetBlankBuf(Flags: LongWord): PContBuf;
var
  I: Integer;

begin
  if WaitForSingleObject(hBlankBufSem, 0) = WAIT_TIMEOUT then begin
    Result := ExtraBuf;
    Result.Flags := CallbackFlag_Inserted or Flags;
    Exit;
  end;
  Result := nil;
  WaitForSingleObject(hBufMutex, INFINITE);
    for I := 0 to High(BufBuf) do
      if BufBuf[I].BufState = bsBlank then begin
        Result := BufBuf[I];
        Result.Flags := Flags;
        Break;
      end
    ;
  ReleaseMutex(hBufMutex);

  if Result = nil then begin
    Result := ExtraBuf; //We have a semaphore to prevent this, so not likely to matter.
    Result.Flags := CallbackFlag_Inserted or Flags;
  end;
end;

function TADCContBufWorker.GetDataBufOrKilled: PContBuf;
var
  I: Integer;

begin
  repeat
    WaitForSingleObject(hDataBufSem, INFINITE);
    Result := nil;
    WaitForSingleObject(hBufMutex, INFINITE);
      for I := 0 to High(BufBuf) do
        if (BufBuf[I].BufState = bsGotData) and (BufBuf[I].Index = NextIndexOut) then begin
          Result := BufBuf[I];
          Result.BufState := bsAccessing;
          Inc(NextIndexOut);
          Break;
        end
      ;
    ReleaseMutex(hBufMutex);

    if Result = nil then begin
      if WaitForSingleObject(hKillSem, 0) <> WAIT_TIMEOUT then Exit;
    end;
  until Result <> nil; //We have a semaphore to prevent this, so just go around again if it happens.
end;

procedure TADCContBufWorker.GetDebugStats(var Blanks, Accessings, GotDatas: LongWord);
var
  I: Integer;
  iBlanks, iAccessings, iGotDatas: LongWord;

begin
  iBlanks := 0;
  iAccessings := 0;
  iGotDatas := 0;
  WaitForSingleObject(hBufMutex, INFINITE);
    for I := 0 to High(BufBuf) do
      case BufBuf[I].BufState of
        bsBlank: Inc(iBlanks);
        bsAccessing: Inc(iAccessings);
        bsGotData: Inc(iGotDatas);
      end
    ;
  ReleaseMutex(hBufMutex);
  Blanks     := iBlanks;
  Accessings := iAccessings;
  GotDatas   := iGotDatas;
end;

procedure TADCContBufWorker.PutBlankBuf(pNewBuf: PContBuf);
begin
  WaitForSingleObject(hBufMutex, INFINITE);
    pNewBuf.BufState := bsBlank;
  ReleaseMutex(hBufMutex);
  ReleaseSemaphore(hBlankBufSem, 1, nil);
end;

procedure TADCContBufWorker.PutDataBuf(pNewBuf: PContBuf);
begin
  WaitForSingleObject(hBufMutex, INFINITE);
    pNewBuf.BufState := bsGotData;
    pNewBuf.Index := NextIndexIn;
    Inc(NextIndexIn);
  ReleaseMutex(hBufMutex);
  ReleaseSemaphore(hDataBufSem, 1, nil);
end;


end.

