unit DIO;
interface
uses
  Windows;

function DIO_Configure(DeviceIndex: LongWord; Tristate: ByteBool; pOutMask: Pointer; pData: Pointer): LongWord; cdecl;
function DIO_ConfigureEx(DeviceIndex: LongWord; pOutMask: Pointer; pData: Pointer; pTristateMask: Pointer): LongWord; cdecl;
function DIO_ConfigurationQuery(DeviceIndex: LongWord; pOutMask: Pointer; pTristateMask: Pointer): LongWord; cdecl;
function DIO_ConfigureMasked(DeviceIndex: LongWord; pOuts: Pointer; pOutsMask: Pointer; pData: Pointer; pDataMask: Pointer; pTristates: Pointer; pTristatesMask: Pointer): LongWord; cdecl;
function DIO_Write1(DeviceIndex, BitIndex: LongWord; Data: ByteBool): LongWord; cdecl;
function DIO_Write8(DeviceIndex, ByteIndex: LongWord; Data: Byte): LongWord; cdecl;
function DIO_WriteAll(DeviceIndex: LongWord; pData: Pointer): LongWord; cdecl;
function DIO_Read1(DeviceIndex, BitIndex: LongWord; Buffer: PByte): LongWord; cdecl;
function DIO_Read8(DeviceIndex, ByteIndex: LongWord; Buffer: PByte): LongWord; cdecl;
function DIO_ReadAll(DeviceIndex: LongWord; pData: Pointer): LongWord; cdecl;

implementation
uses

  DeviceLevel,
  AIOUSB_Constants,VRTable;

function DIO_Configure(DeviceIndex: LongWord; Tristate: ByteBool; pOutMask: Pointer; pData: Pointer): LongWord; cdecl;
var
  Content: AnsiString;
  Data, OutMask, COSMask: AnsiString;
  L: LongWord;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if DIOBytes = 0 then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if (pOutMask = nil) or (pData = nil) then begin
      Result := ERROR_INVALID_PARAMETER;
      Exit;
    end;

    SetString(Data, PAnsiChar(pData), DIOBytes);
    Move(Data[1], LastDIOData[0], DIOBytes);
    L := (DIOConfigBits + 7) div 8;
    SetString(OutMask, PAnsiChar(pOutMask), L);
    COSMask := StringOfChar(AnsiChar(#$00), L);

    Content := Data + OutMask + COSMask;
    Result := GenericVendorWrite(DeviceIndex, AUR_DIO_CONFIG, Ord(Tristate), 0, Length(Content), PAnsiChar(Content));
  end;
end;

function DIO_ConfigureEx(DeviceIndex: LongWord; pOutMask: Pointer; pData: Pointer; pTristateMask: Pointer): LongWord; cdecl;
var
  Content: AnsiString;
  Data, OutMask, TriMask: AnsiString;
  L: LongWord;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if (DIOBytes = 0) or (Tristates = 0) then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if (pOutMask = nil) or (pData = nil) or (pTristateMask = nil) then begin
      Result := ERROR_INVALID_PARAMETER;
      Exit;
    end;

    SetString(Data, PAnsiChar(pData), DIOBytes);
    Move(Data[1], LastDIOData[0], DIOBytes);
    L := (DIOConfigBits + 7) div 8;
    SetString(OutMask, PAnsiChar(pOutMask), L);
    L := (Tristates + 7) div 8;
    SetString(TriMask, PAnsiChar(pTristateMask), L);

    Content := Data + OutMask + TriMask;
    Result := GenericVendorWrite(DeviceIndex, AUR_DIO_CONFIG, 0, DIOBytes, Length(Content), PAnsiChar(Content));
  end;
end;

function DIO_ConfigurationQuery(DeviceIndex: LongWord; pOutMask: Pointer; pTristateMask: Pointer): LongWord; cdecl;
var
  Content: AnsiString;
  OutMask, TriMask: AnsiString;
  L, OutL, TriL: LongWord;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if (Tristates = 0) then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if (pOutMask = nil) or (pTristateMask = nil) then begin
      Result := ERROR_INVALID_PARAMETER;
      Exit;
    end;

    OutL := (DIOConfigBits + 7) div 8;
    TriL := (Tristates + 7) div 8;

    L := OutL + TriL;
    SetString(Content, nil, L);
    Result := GenericVendorRead(DeviceIndex, AUR_DIO_CONFIG_QUERY, 0, DIOBytes, L, PAnsiChar(Content));
    //OutMask + TriMask := Content;
    OutMask := Copy(Content, 1, OutL);
    TriMask := Copy(Content, 1 + OutL, TriL);
    Move(OutMask[1], pOutMask^, OutL);
    Move(TriMask[1], pTristateMask^, TriL);
  end;
end;

function DIO_ConfigureMasked(DeviceIndex: LongWord; pOuts: Pointer; pOutsMask: Pointer; pData: Pointer; pDataMask: Pointer; pTristates: Pointer; pTristatesMask: Pointer): LongWord; cdecl;

  function StringOfByte(B: Byte; Count: Integer): AnsiString;
  begin
    SetString(Result, nil, Count);
    FillChar(Result[1], Count, B);
  end;

  procedure SetBuffersFromPointers(out DataBuf, MaskBuf: AnsiString; DefDataByte, DefMaskByte: Byte; Len: Integer; pThisData, pThisMask: PAnsiChar);
  begin
    if pThisMask = nil then begin
      if pThisData = nil then begin
        DataBuf := StringOfByte(DefDataByte, Len);
        MaskBuf := StringOfByte(DefMaskByte, Len);
      end
      else begin
        SetString(DataBuf, PAnsiChar(pThisData), Len);
        MaskBuf := StringOfByte($FF, Len);
      end;
    end
    else begin
      if pThisData = nil then begin
        DataBuf := StringOfByte(DefDataByte, Len);
        SetString(MaskBuf, PAnsiChar(pThisMask), Len);
      end
      else begin
        SetString(DataBuf, PAnsiChar(pThisData), Len);
        SetString(MaskBuf, PAnsiChar(pThisMask), Len);
      end;
    end;
  end;

var
  Data, DataMask, Outs, OutsMask, Tris, TrisMask,
  Content: AnsiString;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if (DIOBytes = 0) or (Tristates = 0) then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    SetBuffersFromPointers(Data, DataMask, $FF, $00, DIOBytes, pData, pDataMask);
    SetBuffersFromPointers(Outs, OutsMask, $00, $00, (DIOConfigBits + 7) div 8, pOuts, pOutsMask);
    SetBuffersFromPointers(Tris, TrisMask, $00, $FF, (Tristates + 7) div 8, pTristates, pTristatesMask);

    Move(Data[1], LastDIOData[0], DIOBytes);

    Content := OutsMask + Outs + DataMask + Data + TrisMask + Tris;
    Result := GenericVendorWrite(DeviceIndex, AUR_DIO_CONFIG_MASKED, 0, 0, Length(Content), PAnsiChar(Content));
  end;
end;

function DIO_Write1(DeviceIndex, BitIndex: LongWord; Data: ByteBool): LongWord; cdecl;
var
  ByteIndex: LongWord;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[DeviceIndex] do begin
      if DIOBytes = 0 then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      ByteIndex := BitIndex div 8;
      BitIndex := BitIndex and 7;

      if ByteIndex >= DIOBytes then begin
        Result := ERROR_INVALID_ADDRESS;
        Exit;
      end;

      if Data then
        LastDIOData[ByteIndex] := LastDIOData[ByteIndex] or (1 shl BitIndex)
      else
        LastDIOData[ByteIndex] := LastDIOData[ByteIndex] and not (1 shl BitIndex)
      ;

      Result := GenericVendorWrite(DeviceIndex, AUR_DIO_WRITE, 0, 0, DIOBytes, @LastDIOData[0]);
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function DIO_Write8(DeviceIndex, ByteIndex: LongWord; Data: Byte): LongWord; cdecl;
begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[DeviceIndex] do begin
      if DIOBytes = 0 then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      if ByteIndex >= DIOBytes then begin
        Result := ERROR_INVALID_ADDRESS;
        Exit;
      end;

      LastDIOData[ByteIndex] := Data;

      Result := GenericVendorWrite(DeviceIndex, AUR_DIO_WRITE, 0, 0, DIOBytes, @LastDIOData[0]);
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function DIO_WriteAll(DeviceIndex: LongWord; pData: Pointer): LongWord; cdecl;
begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[DeviceIndex] do begin
      if DIOBytes = 0 then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      Move(pData^, LastDIOData[0], DIOBytes);

      Result := GenericVendorWrite(DeviceIndex, AUR_DIO_WRITE, 0, 0, DIOBytes, @LastDIOData[0]);
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function DIO_Read1(DeviceIndex, BitIndex: LongWord; Buffer: PByte): LongWord; cdecl;
var
  Content: AnsiString;
  L: LongWord;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[DeviceIndex] do begin
      if DIOBytes = 0 then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      if BitIndex >= DIOBytes * 8 then begin
        Result := ERROR_INVALID_ADDRESS;
        Exit;
      end;

      L := DIOBytes;
      SetString(Content, nil, L);
      Result := GenericVendorRead(DeviceIndex, AUR_DIO_READ, 0, 0, L, PAnsiChar(Content));

      if Result = ERROR_SUCCESS then begin

        if (Byte(Content[(BitIndex div 8) + 1]) and (1 shl (BitIndex mod 8))) <> 0 then
          Buffer^ := 1
        else
          Buffer^ := 0
        ;
      end;
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function DIO_Read8(DeviceIndex, ByteIndex: LongWord; Buffer: PByte): LongWord; cdecl;
var
  Content: AnsiString;
  L: LongWord;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[DeviceIndex] do begin
      if DIOBytes = 0 then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      if ByteIndex >= DIOBytes then begin
        Result := ERROR_INVALID_ADDRESS;
        Exit;
      end;

      L := DIOBytes;
      SetString(Content, nil, L);
      Result := GenericVendorRead(DeviceIndex, AUR_DIO_READ, 0, 0, L, PAnsiChar(Content));

      if Result = ERROR_SUCCESS then begin
        PAnsiChar(Buffer)^ := Content[ByteIndex + 1];
      end;
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function DIO_ReadAll(DeviceIndex: LongWord; pData: Pointer): LongWord; cdecl;
var
  L: LongWord;

begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[DeviceIndex] do begin
      if DIOBytes = 0 then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      L := DIOBytes;
      Result := GenericVendorRead(DeviceIndex, AUR_DIO_READ, 0, 0, L, pData);
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

end.
