unit EEPROM;

interface

function CustomEEPROMRead(DeviceIndex: LongWord; StartAddress: LongWord; var DataSize: LongWord; Data: Pointer): LongWord; cdecl;
function CustomEEPROMWrite(DeviceIndex: LongWord; StartAddress: LongWord; DataSize: LongWord; Data: Pointer): LongWord; cdecl;

implementation

Uses
  Windows,
  AIOUSB_Constants,
  DeviceLevel; 

function CustomEEPROMRead(DeviceIndex: LongWord; StartAddress: LongWord; var DataSize: LongWord; Data: Pointer): LongWord; cdecl;
begin
  if StartAddress > $FF then begin
    Result := ERROR_SEEK;
    Exit;
  end;
  if DataSize > (LongWord($100) - StartAddress) then begin
    Result := ERROR_BAD_LENGTH;
    Exit;
  end;

  Result := GenericVendorRead(DeviceIndex, $A2, $1E00 + StartAddress, 0, DataSize, Data);
end;

function CustomEEPROMWrite(DeviceIndex: LongWord; StartAddress: LongWord; DataSize: LongWord; Data: Pointer): LongWord; cdecl;
const
  BlockSize = 32;

begin
  if StartAddress > $FF then begin
    Result := ERROR_SEEK;
    Exit;
  end;
  if DataSize > (LongWord($100) - StartAddress) then begin
    Result := ERROR_BAD_LENGTH;
    Exit;
  end;

  while DataSize > BlockSize do begin
    Result := GenericVendorWrite(DeviceIndex, $A2, $1E00 + StartAddress, 0, BlockSize, Data);
    if Result <> ERROR_SUCCESS then Exit;
    Inc(StartAddress, BlockSize);
    Dec(DataSize, BlockSize);
    Inc(PByte(Data), BlockSize);
  end;
  Result := GenericVendorWrite(DeviceIndex, $A2, $1E00 + StartAddress, 0, DataSize, Data);
end;

end.

