unit FIFO_Stream;
interface

function AIOUSB_ClearFIFO(DeviceIndex: LongWord; TimeMethod: LongWord): LongWord; cdecl;
function AIOUSB_GetStreamStatus(DeviceIndex: LongWord; var Status: LongWord): LongWord; cdecl;
function AIOUSB_SetStreamingBlockSize(DeviceIndex, BlockSize: LongWord): LongWord; cdecl;

implementation

uses
  Windows,
  AIOUSB_Constants,
  DeviceLevel,
  VRTable;

function AIOUSB_ClearFIFO(DeviceIndex: LongWord; TimeMethod: LongWord): LongWord; cdecl;
var
  Req: Byte;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if not bClearFIFO then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    case TimeMethod of
      0: Req := AUR_GEN_CLEAR_FIFO;
      1: Req := AUR_GEN_CLEAR_FIFO_NEXT;
      5: Req := AUR_GEN_ABORT_AND_CLEAR;
      86: Req := AUR_GEN_CLEAR_FIFO_WAIT;
      else begin
        Result := ERROR_INVALID_PARAMETER;
        Exit;
      end;
    end;
    Result := GenericVendorWrite(DeviceIndex, Req, 0, 0, 0, nil);
  end;
end;

function AIOUSB_GetStreamStatus(DeviceIndex: LongWord; var Status: LongWord): LongWord; cdecl;
var
  L: LongWord;
  ReadValue: LongWord;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if not bClearFIFO then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    L := SizeOf(ReadValue);
    Result := GenericVendorRead(DeviceIndex, AUR_GEN_STREAM_STATUS, 0, 0, L, @ReadValue);

    if Result = ERROR_SUCCESS then Status := ReadValue;
  end;
end;

function AIOUSB_SetStreamingBlockSize(DeviceIndex, BlockSize: LongWord): LongWord; cdecl;
begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if not (bDIOStream or bADCBulk) then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if (BlockSize = 0) or (BlockSize > 31*1024*1024) then begin
      Result := ERROR_INVALID_PARAMETER;
      Exit;
    end;

    if bADCBulk then begin
      if (BlockSize and $1FF) <> 0 then
        BlockSize := (BlockSize and $FFFFFE00) + $200
      else
        BlockSize := (BlockSize and $FFFFFE00)
      ;
    end;

    if bDIOStream then begin
      if (BlockSize and $FF) <> 0 then
        BlockSize := (BlockSize and $FFFFFF00) + $100
      else
        BlockSize := (BlockSize and $FFFFFF00)
      ;
    end;

    StreamingBlockSize := BlockSize;

    Result := ERROR_SUCCESS;
  end;
end;

end.
