unit FastDIO;
interface

uses
  Windows;

function DIO_StreamOpen(DeviceIndex: LongWord; bIsRead: LongBool): LongWord; cdecl;
function DIO_StreamClose(DeviceIndex: LongWord): LongWord; cdecl;
function DIO_StreamFrame(DeviceIndex, FramePoints: LongWord; pFrameData: PWord; var BytesTransferred: LongWord): LongWord; cdecl;
function DIO_StreamSetClocks(DeviceIndex: LongWord; var ReadClockHz, WriteClockHz: Double): LongWord; cdecl;

implementation

uses
  DeviceLevel,
  AIOUSB_Constants,
  VRTable;

function DIO_StreamOpen(DeviceIndex: LongWord; bIsRead: LongBool): LongWord; cdecl;
begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if not bDIOStream then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if bDIOOpen then begin
      Result := ERROR_TOO_MANY_OPEN_FILES;
      Exit;
    end;

    bDIOOpen := True;

    //Result := DoSetIFace(Handle, $100);
    //if Result <> ERROR_SUCCESS then Exit;
    
    if bIsRead then begin
      bDIORead := True;

      Result := GenericVendorWrite(DeviceIndex, $BC, 0, 0, 0, nil);
    end
    else begin
      bDIORead := False;

      Result := GenericVendorWrite(DeviceIndex, $BB, 0, 0, 0, nil);
    end;
  end;
end;

function DIO_StreamSetClocks(DeviceIndex: LongWord; var ReadClockHz, WriteClockHz: Double): LongWord; cdecl;
var
  Content: AnsiString;
  pDB: PByte;
  pDW: PWord absolute pDB;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if not bDIOStream then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    {
    if not bDIOOpen then begin
      Result := ERROR_FILE_NOT_FOUND;
      Exit;
    end;
    }

    SetString(Content, nil, 5);
    pDB := PByte(PAnsiChar(Content));
    pDB^ := $00;
    if WriteClockHz = 0 then Inc(pDB^, 1);
    if  ReadClockHz = 0 then Inc(pDB^, 2);
    Inc(pDB);
    pDW^ := OctDacFromFreq(WriteClockHz);
    Inc(pDW);
    pDW^ := OctDacFromFreq(ReadClockHz);
    {
    SetString(Content, nil, SizeOf(TVendorRequestData));
    pVRD.Request := AUR_DIO_SETCLOCKS;
    pVRD.Value := 0;
    pVRD.Index := 0;

    Content := Content + #$02 + WordAsStr(OctDacFromFreq(WriteClockHz)) + WordAsStr(OctDacFromFreq(WriteClockHz));
    }

    Result := GenericVendorWrite(DeviceIndex, AUR_DIO_SETCLOCKS, 0, 0, Length(Content), PAnsiChar(Content));
  end;
end;


function DIO_StreamClose(DeviceIndex: LongWord): LongWord; cdecl;
begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if not bDIOStream then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if not bDIOOpen then begin
      Result := ERROR_FILE_NOT_FOUND;
      Exit;
    end;

    bDIOOpen := False;
    Result := ERROR_SUCCESS;
  end;
end;

function DIO_StreamFrame(DeviceIndex, FramePoints: LongWord; pFrameData: PWord; var BytesTransferred: LongWord): LongWord; cdecl;
var
  TargetPipe: Word;
  L: LongWord;

begin
  try
    BytesTransferred := 0;

    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[DeviceIndex] do begin
      if not bDIOStream then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      if not bDIOOpen then begin
        Result := ERROR_FILE_NOT_FOUND;
        Exit;
      end;

      if FramePoints <= 0 then begin
        Result := ERROR_INVALID_PARAMETER;
        Exit;
      end;

      Result := ERROR_SUCCESS;
      if bDIORead then begin
        TargetPipe := $86;

        while FramePoints > StreamingBlockSize do begin
          L := 0;
          Result := _GenericBulkIn(DeviceIndex, TargetPipe, pFrameData, StreamingBlockSize * 2, L);
          if L < StreamingBlockSize * 2 then Result := ERROR_HANDLE_EOF;
          BytesTransferred := BytesTransferred + L;
          if Result <> ERROR_SUCCESS then Break;
          Inc(pFrameData, StreamingBlockSize);
          Dec(FramePoints, StreamingBlockSize);
        end;
        if (Result = ERROR_SUCCESS) and (FramePoints <> 0) then begin
          L := 0;
          Result := _GenericBulkIn(DeviceIndex, TargetPipe, pFrameData, FramePoints * 2, L);
          if Result = ERROR_SUCCESS then BytesTransferred := BytesTransferred + L;
        end;
      end
      else begin
        TargetPipe := $02;

        while FramePoints > StreamingBlockSize do begin
          L := 0;
          Result := AWU_GenericBulkOut(DeviceIndex, TargetPipe, pFrameData, StreamingBlockSize * 2, L);
          if L < StreamingBlockSize * 2 then Result := ERROR_HANDLE_EOF;
          BytesTransferred := BytesTransferred + L;
          if Result <> ERROR_SUCCESS then Break;
          Inc(pFrameData, StreamingBlockSize);
          Dec(FramePoints, StreamingBlockSize);
        end;
        if (Result = ERROR_SUCCESS) and (FramePoints <> 0) then begin
          L := 0;
          Result := AWU_GenericBulkOut(DeviceIndex, TargetPipe, pFrameData, FramePoints * 2, L);
          if Result = ERROR_SUCCESS then BytesTransferred := BytesTransferred + L;
        end;
      end;
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function DIO_16A_SetFunctionDirection(DeviceIndex, DIOStreaming, SPIBus, DIOA, DIOB, DIOCH: LongWord): LongWord; cdecl;
const
  FDD_Input                 = $00000001;
  FDD_Output                = $00000010;
  FDD_                      = $00000011;
  FDM_HandshakeAuto         = $00000100;
  FDM_HandshakeSlave        = $00001000;
  FDM_HandshakeForcedMaster = $00010000;
  FDM_                      = $00011100;

begin
  //`@
  Result := ERROR_NOT_SUPPORTED;
end;

end.
