unit Flash;

interface
function AIOUSB_FlashRead(DeviceIndex: LongWord; StartAddress: LongWord; var DataSize: LongWord; Data: Pointer): LongWord; cdecl;
function AIOUSB_FlashWrite(DeviceIndex: LongWord; StartAddress: LongWord; DataSize: LongWord; Data: Pointer): LongWord; cdecl;
function AIOUSB_FlashEraseSector(DeviceIndex: LongWord; Sector: Integer): LongWord; cdecl;

implementation
uses
  Windows,
  AIOUSB_Constants,
  DeviceLevel,
  VRTable;

function AIOUSB_FlashRead(DeviceIndex: LongWord; StartAddress: LongWord; var DataSize: LongWord; Data: Pointer): LongWord; cdecl;
begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if FlashSectors = 0 then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if StartAddress >= C_FLASH_SIZE then begin
      Result := ERROR_SEEK;
      Exit;
    end;
    if DataSize > (LongWord(C_FLASH_SIZE) - StartAddress) then begin
      Result := ERROR_BAD_LENGTH;
      Exit;
    end;
    if (DataSize and 1) <> 0 then begin
      DataSize := DataSize and $FFFFFFFE;
    end;

    Result := GenericVendorRead(DeviceIndex, AUR_FLASH_READWRITE, StartAddress shr 1, 0, DataSize, Data);
  end;
end;

function AIOUSB_FlashWrite(DeviceIndex: LongWord; StartAddress: LongWord; DataSize: LongWord; Data: Pointer): LongWord; cdecl;
begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if FlashSectors = 0 then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if StartAddress >= C_FLASH_SIZE then begin
      Result := ERROR_SEEK;
      Exit;
    end;
    if DataSize > (LongWord(C_FLASH_SIZE) - StartAddress) then begin
      Result := ERROR_BAD_LENGTH;
      Exit;
    end;
    if (DataSize and 1) <> 0 then begin
      DataSize := DataSize and $FFFFFFFE;
    end;

    while DataSize > C_FLASH_BLOCK_SIZE do begin
      Result := GenericVendorWrite(DeviceIndex, AUR_FLASH_READWRITE, StartAddress shr 1, StartAddress shr 17, C_FLASH_BLOCK_SIZE, Data);
      if Result <> ERROR_SUCCESS then Exit;
      Inc(StartAddress, C_FLASH_BLOCK_SIZE);
      Dec(DataSize, C_FLASH_BLOCK_SIZE);
      Inc(PByte(Data), C_FLASH_BLOCK_SIZE);
    end;
    Result := GenericVendorWrite(DeviceIndex, AUR_FLASH_READWRITE, StartAddress shr 1, StartAddress shr 17, DataSize, Data);
  end;
end;

function AIOUSB_FlashEraseSector(DeviceIndex: LongWord; Sector: Integer): LongWord; cdecl;
begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if FlashSectors = 0 then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if Sector >= FlashSectors then begin
      Result := ERROR_SEEK;
      Exit;
    end;

    Result := GenericVendorWrite(DeviceIndex, AUR_DEBUG_FLASH_ERASE, 0, Sector, 0, nil);
  end;
end;

end.
