unit u8254;

interface

uses
  Windows,
  theseUtils,
  ContBufThreads;

type

TCtrMonitorThread = class(TStartableThread)
public
  DI: LongWord;
  hListMutex: THandle;
  MeasuredPulseWidth: array of Word;
protected
  procedure Execute; override;
end;

function CTR_8254Mode(DeviceIndex, BlockIndex, CounterIndex, Mode: LongWord): LongWord; cdecl;
function CTR_8254Load(DeviceIndex, BlockIndex, CounterIndex: LongWord; LoadValue: Word): LongWord; cdecl;
function CTR_8254ModeLoad(DeviceIndex, BlockIndex, CounterIndex, Mode: LongWord; LoadValue: Word): LongWord; cdecl;
function CTR_8254ReadModeLoad(DeviceIndex, BlockIndex, CounterIndex, Mode: LongWord; LoadValue: Word; pReadValue: PWord): LongWord; cdecl;
function CTR_8254Read(DeviceIndex, BlockIndex, CounterIndex: LongWord; pReadValue: PWord): LongWord; cdecl;
function CTR_8254ReadStatus(DeviceIndex, BlockIndex, CounterIndex: LongWord; pReadValue: PWord; pStatus: PByte): LongWord; cdecl;
function CTR_8254ReadAll(DeviceIndex: LongWord; pData: PWord): LongWord; cdecl;
function CTR_StartOutputFreq(DeviceIndex, BlockIndex: LongWord; pHz: PDouble): LongWord; cdecl;
function CTR_8254ReadLatched(DeviceIndex: LongWord; pData: PWord): LongWord; cdecl;
function CTR_8254SelectGate(DeviceIndex, GateIndex: LongWord): LongWord; cdecl;

function CTR_SetWaitGates(DeviceIndex: LongWord; A, B: Byte): LongWord; cdecl;
function CTR_EndWaitGates(DeviceIndex: LongWord): LongWord; cdecl;
function CTR_WaitForGate(DeviceIndex: LongWord; GateIndex: Byte; var Content: Word): LongWord; cdecl;

function CTR_StartMeasuringPulseWidth(DeviceIndex: LongWord): LongWord; cdecl;
function CTR_StopMeasuringPulseWidth(DeviceIndex: LongWord): LongWord; cdecl;
function CTR_GetPulseWidthMeasurement(DeviceIndex, BlockIndex, CounterIndex: LongWord; pReadValue: PWord): LongWord; cdecl;

procedure CalcHzDivisors(pHz: PDouble; pDivisorA, pDivisorB: PWord; RootClock: Double);

implementation

uses
{$IFDEF VER130}  // Delphi 5
  SysUtils, Classes,
{$ELSE}         // anything other than Delphi 5, presumably XE2 or XE3
  System.SysUtils,
  System.Classes,
{$ENDIF}
  AIOUSB_Constants,
  DeviceLevel,
  VRTable;

procedure CalcHzDivisors(pHz: PDouble; pDivisorA, pDivisorB: PWord; RootClock: Double);
var
  L, DivisorA, DivisorB: LongWord;
  DivisorAB: Extended;
  Err, MinErr, Hz: Double;

begin
  Hz := pHz^;
  if Hz * 4 >= RootClock then begin
    DivisorA := 2;
    DivisorB := 2;
  end
  else begin
    DivisorAB := RootClock / Hz;

    L := Round(SqRt(DivisorAB));
    DivisorA := Round(DivisorAB / L);
    DivisorB := L;
    MinErr := Abs(Hz - (RootClock / (DivisorA * L)));
    for L := L downto 2 do begin
      DivisorA := Round(DivisorAB / L);
      if DivisorA > $FFFF then Break; //Limited to 16 bits, so this and all further L are invalid.
      Err := Abs(Hz - (RootClock / (DivisorA * L)));
      if Err = 0 then begin
        DivisorB := L;
        Break;
      end;
      if Err < MinErr then begin
        DivisorB := L;
        MinErr := Err;
      end;
    end;
    DivisorA := Round(DivisorAB / DivisorB);
  end;

  pDivisorA^ := DivisorA;
  pDivisorB^ := DivisorB;

  pHz^ := RootClock / (DivisorA * DivisorB);
end;

function MapCounterBlock(DeviceIndex: LongWord; var CounterIndex, BlockIndex: LongWord): LongWord;
begin
  Result := ERROR_SUCCESS;

  if BlockIndex = 0 then begin
    //Contiguous counter addressing
    BlockIndex := CounterIndex div 3;
    CounterIndex := CounterIndex mod 3;

    if BlockIndex >= Dev[DeviceIndex].Counters then
      Result := ERROR_INVALID_ADDRESS
    ;
  end
  else begin
    if (BlockIndex >= Dev[DeviceIndex].Counters)
    or (CounterIndex >= 3) then
      Result := ERROR_INVALID_ADDRESS
    ;
  end;
end;


function CTR_8254Mode(DeviceIndex, BlockIndex, CounterIndex, Mode: LongWord): LongWord; cdecl;
begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if Counters = 0 then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    Result := MapCounterBlock(DeviceIndex, CounterIndex, BlockIndex);
    if Result <> ERROR_SUCCESS then Exit;

    if Mode >= 6 then begin
      Result := ERROR_INVALID_ADDRESS;
      Exit;
    end;

    Mode := (CounterIndex shl 6) or (Mode shl 1) or $30;

    Result := GenericVendorWrite(DeviceIndex, AUR_CTR_MODE, BlockIndex or (Mode shl 8), 0, 0, nil);
  end;
end;

function CTR_8254Load(DeviceIndex, BlockIndex, CounterIndex: LongWord; LoadValue: Word): LongWord; cdecl;
var
  Mode: Byte;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if Counters = 0 then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    Result := MapCounterBlock(DeviceIndex, CounterIndex, BlockIndex);
    if Result <> ERROR_SUCCESS then Exit;

    Mode := (CounterIndex shl 6) //or (Mode shl 1) or $30
    ;

    Result := GenericVendorWrite(DeviceIndex, AUR_CTR_LOAD, BlockIndex or (Mode shl 8), LoadValue, 0, nil);
  end;
end;

function CTR_8254ModeLoad(DeviceIndex, BlockIndex, CounterIndex, Mode: LongWord; LoadValue: Word): LongWord; cdecl;
begin
  try
    Result := ValidateAndEnsureOpen(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[DeviceIndex] do begin
      if Counters = 0 then begin
        Result := ERROR_BAD_TOKEN_TYPE;
        Exit;
      end;

      Result := MapCounterBlock(DeviceIndex, CounterIndex, BlockIndex);
      if Result <> ERROR_SUCCESS then Exit;

      if Mode >= 6 then begin
        Result := ERROR_INVALID_ADDRESS;
        Exit;
      end;

      Mode := (CounterIndex shl 6) or (Mode shl 1) or $30;

      Result := GenericVendorWrite(DeviceIndex, AUR_CTR_MODELOAD, BlockIndex or (Mode shl 8), LoadValue, 0, nil);
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function CTR_8254ReadModeLoad(DeviceIndex, BlockIndex, CounterIndex, Mode: LongWord; LoadValue: Word; pReadValue: PWord): LongWord; cdecl;
var
  L: LongWord;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if Counters = 0 then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    Result := MapCounterBlock(DeviceIndex, CounterIndex, BlockIndex);
    if Result <> ERROR_SUCCESS then Exit;

    if Mode >= 6 then begin
      Result := ERROR_INVALID_ADDRESS;
      Exit;
    end;

    Mode := (CounterIndex shl 6) or (Mode shl 1) or $30;

    L := 2;
    Result := GenericVendorRead(DeviceIndex, AUR_CTR_MODELOAD, BlockIndex or (Mode shl 8), LoadValue, L, pReadValue);
  end;
end;

function CTR_8254Read(DeviceIndex, BlockIndex, CounterIndex: LongWord; pReadValue: PWord): LongWord; cdecl;
var
  L: LongWord;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if Counters = 0 then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    Result := MapCounterBlock(DeviceIndex, CounterIndex, BlockIndex);
    if Result <> ERROR_SUCCESS then Exit;

    L := 2;
    Result := GenericVendorRead(DeviceIndex, AUR_CTR_READ, BlockIndex or (CounterIndex shl 8), 0, L, pReadValue);
  end;
end;

function CTR_8254ReadStatus(DeviceIndex, BlockIndex, CounterIndex: LongWord; pReadValue: PWord; pStatus: PByte): LongWord; cdecl;
var
  Content: AnsiString;
  L: LongWord;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if Counters = 0 then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    Result := MapCounterBlock(DeviceIndex, CounterIndex, BlockIndex);
    if Result <> ERROR_SUCCESS then Exit;

    L := 3;
    SetString(Content, nil, L);
    Result := GenericVendorRead(DeviceIndex, AUR_CTR_READ, BlockIndex or (CounterIndex shl 8), 0, L, PAnsiChar(Content));
    pReadValue^ := PWord(@Content[1])^;
    pStatus^    := PByte(@Content[3])^;
  end;
end;

function CTR_8254ReadAll(DeviceIndex: LongWord; pData: PWord): LongWord; cdecl;
var
  L: LongWord;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if Counters = 0 then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    L := Counters * 3 * 2;
    Result := GenericVendorRead(DeviceIndex, AUR_CTR_READALL, 0, 0, L, pData);
  end;
end;

function CTR_8254ReadLatched(DeviceIndex: LongWord; pData: PWord): LongWord; cdecl;
var
  L: LongWord;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if Counters = 0 then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    L := Counters * 3 * 2 + 1;
    Result := GenericVendorRead(DeviceIndex, AUR_CTR_READLATCHED, 0, 0, L, pData);
  end;
end;

function CTR_8254SelectGate(DeviceIndex, GateIndex: LongWord): LongWord; cdecl;
begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if Counters = 0 then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if not bGateSelectable then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if GateIndex >= Counters * 3 then begin
      Result := ERROR_INVALID_ADDRESS;
      Exit;
    end;

    Result := GenericVendorWrite(DeviceIndex, AUR_CTR_SELGATE, GateIndex, 0, 0, nil);
  end;
end;

function CTR_StartOutputFreq(DeviceIndex, BlockIndex: LongWord; pHz: PDouble): LongWord; cdecl;
var
  L, DivisorA, DivisorB: LongWord;
  DivisorAB: Extended;
  Err, MinErr, Hz: Double;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if Counters = 0 then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if BlockIndex >= Counters then begin
      Result := ERROR_INVALID_ADDRESS;
      Exit;
    end;

    if pHz = nil then begin
      Result := ERROR_INVALID_PARAMETER;
      Exit;
    end;

    Hz := pHz^;
    if Hz <= 0 then begin
      CTR_8254Mode(DeviceIndex, BlockIndex, 1, 2);
      Result := CTR_8254Mode(DeviceIndex, BlockIndex, 2, 3);
    end
    else begin
      try
        if Hz * 4 >= RootClock then begin
          DivisorA := 2;
          DivisorB := 2;
        end
        else begin
          DivisorAB := RootClock / Hz;

          L := Round(SqRt(DivisorAB));
          DivisorA := Round(DivisorAB / L);
          DivisorB := L;
          MinErr := Abs(Hz - (RootClock / (DivisorA * L)));
          for L := L downto 2 do begin
            DivisorA := Round(DivisorAB / L);
            if DivisorA > $FFFF then Break; //Limited to 16 bits, so this and all further L are invalid.
            Err := Abs(Hz - (RootClock / (DivisorA * L)));
            if Err = 0 then begin
              DivisorB := L;
              Break;
            end;
            if Err < MinErr then begin
              DivisorB := L;
              MinErr := Err;
            end;
          end;
          DivisorA := Round(DivisorAB / DivisorB);
        end;
      except
        Result := ERROR_INVALID_DATA;
        Exit;
      end;

      //LastPurpose := AUR_CTR_PUR_OFRQ;

      pHz^ := RootClock / (DivisorA * DivisorB);

      CTR_8254ModeLoad(DeviceIndex, BlockIndex, 1, 2, DivisorA);
      Result := CTR_8254ModeLoad(DeviceIndex, BlockIndex, 2, 3, DivisorB);
    end;
  end;
end;

function CTR_SetWaitGates(DeviceIndex: LongWord; A, B: Byte): LongWord; cdecl;
begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if Counters = 0 then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if not bGateSelectable then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if A >= Counters * 3 then begin
      Result := ERROR_INVALID_ADDRESS;
      Exit;
    end;

    if B >= Counters * 3 then begin
      Result := ERROR_INVALID_ADDRESS;
      Exit;
    end;

    //DoSetIFace(Handle, $100);

    Result := GenericVendorWrite(DeviceIndex, AUR_CTR_COS_BULK_GATE2, 0, A or (B shl 8), 0, nil);
  end;
end;

function CTR_EndWaitGates(DeviceIndex: LongWord): LongWord; cdecl;
begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if Counters = 0 then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if not bGateSelectable then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    Result := GenericVendorWrite(DeviceIndex, AUR_CTR_COS_BULK_ABORT, 0, 0, 0, nil);
  end;
end;

function CTR_WaitForGate(DeviceIndex: LongWord; GateIndex: Byte; var Content: Word): LongWord; cdecl;
var
  TargetPipe: LongWord;
  L: LongWord;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if Counters = 0 then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if not bGateSelectable then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    case GateIndex of
      0: TargetPipe := $82;
      1: TargetPipe := $86;
      else begin
        Result := ERROR_INVALID_ADDRESS;
        Exit;
      end;
    end;

    L := 0;
    Result := AWU_GenericBulkIn(DeviceIndex, TargetPipe, @Content, SizeOf(Content), L);
  end;
end;

function CTR_StartMeasuringPulseWidth(DeviceIndex: LongWord): LongWord; cdecl;
var
  I: Integer;

begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if Counters = 0 then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if CtrMonitorThread <> nil then CTR_StopMeasuringPulseWidth(DeviceIndex);
    CtrMonitorThread := nil;
    try
      CtrMonitorThread := TCtrMonitorThread.Create(True);
      with TCtrMonitorThread(CtrMonitorThread) do begin
        DI := DeviceIndex;
        hListMutex := CreateMutex(nil, False, nil);
        SetLength(MeasuredPulseWidth, Counters * 3);
        for I := 0 to High(MeasuredPulseWidth) do MeasuredPulseWidth[I] := 0;
        Start;
      end;
      Result := ERROR_SUCCESS;
    except
      Result := ERROR_INTERNAL_ERROR;
      FreeAndNil(CtrMonitorThread);
    end;
  end;
end;

function CTR_StopMeasuringPulseWidth(DeviceIndex: LongWord): LongWord; cdecl;
begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if Counters = 0 then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    if CtrMonitorThread = nil then begin
      Result := ERROR_FILE_NOT_FOUND;
      Exit;
    end;

    FreeAndNil(CtrMonitorThread);
    Result := ERROR_SUCCESS;
  end;
end;

function CTR_GetPulseWidthMeasurement(DeviceIndex, BlockIndex, CounterIndex: LongWord; pReadValue: PWord): LongWord; cdecl;
begin
  Result := ValidateAndEnsureOpen(DeviceIndex);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[DeviceIndex] do begin
    if Counters = 0 then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    Result := MapCounterBlock(DeviceIndex, CounterIndex, BlockIndex);
    if Result <> ERROR_SUCCESS then Exit;

    if CtrMonitorThread = nil then begin
      Result := ERROR_FILE_NOT_FOUND;
      Exit;
    end;

    try
      pReadValue^ := 0;
      try
        WaitForSingleObject(TCtrMonitorThread(CtrMonitorThread).hListMutex, INFINITE);
        pReadValue^ := TCtrMonitorThread(CtrMonitorThread).MeasuredPulseWidth[BlockIndex * 3 + CounterIndex];
        TCtrMonitorThread(CtrMonitorThread).MeasuredPulseWidth[BlockIndex * 3 + CounterIndex] := 0;
        ReleaseMutex(TCtrMonitorThread(CtrMonitorThread).hListMutex);
        if pReadValue^ = 0 then
          Result := ERROR_NO_DATA_DETECTED
        else
          Result := ERROR_SUCCESS
        ;
      except
        Result := ERROR_INTERNAL_ERROR;
      end;
    except
      Result := ERROR_NOACCESS;
    end;
  end;
end;

procedure TCtrMonitorThread.Execute;
var
  ReadingsA, ReadingsB: array of Word;
  I, L: Integer;
  bA: Boolean;

begin
  L := Length(MeasuredPulseWidth);
  SetLength(ReadingsA, L);
  SetLength(ReadingsB, L);
  CTR_8254ReadAll(DI, @ReadingsB[0]);
  bA := True;
  repeat
    if bA then
      CTR_8254ReadAll(DI, @ReadingsA[0])
    else
      CTR_8254ReadAll(DI, @ReadingsB[0])
    ;
    bA := not bA;

    WaitForSingleObject(hListMutex, INFINITE);
    for I := 0 to L - 1 do
      if ReadingsA[I] = ReadingsB[I] then
        MeasuredPulseWidth[I] := ReadingsA[I]
    ;
    ReleaseMutex(hListMutex);

  until Terminated;
  CloseHandle(hListMutex);
end;
end.
